/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#include "EmulatorCommon.h"
#include "Bank_Dummy.h"

#include "CPU_REG.h"			// Software::BusError
#include "RAM_ROM.h"			// gValidateMemory
#include "Bank_SRAM.h"			// SRAMBank::GetMemoryStart


// ===========================================================================
//		 Dummy Bank Accessors
// ===========================================================================
// Dummy banks are non-existent blocks of memory.  Dummy bank accessors do
// not do anything.

AddressBank	DummyBank::fgAddressBank =
{
	DummyBank::GetLong,
	DummyBank::GetWord,
	DummyBank::GetByte,
	DummyBank::SetLong,
	DummyBank::SetWord,
	DummyBank::SetByte,
	DummyBank::GetRealAddress,
	DummyBank::ValidAddress
};

struct MapRange
{
	uaecptr	addr;
	uae_u32	size;
};

// Let us map in at least eight segments of emulator RAM into the "virtual
// RAM" of the emulated ROM.  Each time we call a ROM function via the
// ATrap class, two entries are used up.  By setting kNumPhysicalBanks,
// we allow for four of those to go on at the same time.
//
// I'm now bumping this up to something way bigger.  The idea behind keeping
// the number small was to catch places where ranges were allocated but not
// de-allocated.  There doesn't seem to be much problem with that, so this
// debugging aid is not really needed anymore.  On the other hand, we now
// map in strings from getenv.  These mappings stay around indefinitely, so
// we need to allow for such unbounded growth.

const int		kNumPhysicalBanks	= 100;
MapRange		gPhysicalBanks [kNumPhysicalBanks];

inline Bool HackForHwrGetRAMSize (uaecptr address)
{
//	if ((address & 0xFF000000) == SRAMBank::GetMemoryStart ())
	if (address == SRAMBank::GetMemoryStart () + gRAMBank_Size)
		return true;

	return false;
}


// ---------------------------------------------------------------------------
//		 DummyBank::Initialize
// ---------------------------------------------------------------------------

void DummyBank::Initialize (void)
{
	// Clear the memory banks.

	Memory::InitializeBanks (fgAddressBank, 0, 0xFFFF);

	memset (gPhysicalBanks, 0, sizeof (gPhysicalBanks));
}


// ---------------------------------------------------------------------------
//		 DummyBank::Reset
// ---------------------------------------------------------------------------

void DummyBank::Reset (void)
{
}


// ---------------------------------------------------------------------------
//		 DummyBank::GetLong
// ---------------------------------------------------------------------------

uae_u32 DummyBank::GetLong (uaecptr iAddress)
{
	// Hack to keep HwrGetRAMSize working: it runs off
	// the end of RAM while testing it.

	if (HackForHwrGetRAMSize (iAddress))
		return 0;

#if (VALIDATE_DUMMY_GET)
	if (gMemAccessFlags.fValidate_DummyGet && !ValidAddress (iAddress, sizeof (long)))
	{
		InvalidAccess (iAddress, sizeof (long), true);
	}
#endif

#if HAS_PROFILING
	if (gProfilingEnabled && gProfilingCounted)
	{
		::ProfileIncrementRead (2, WAITSTATES_DUMMYBANK);
	}
#endif

	if (!ValidAddress (iAddress, sizeof (long)))
	{
		return ~0;
	}

	uae_u8*	p = (uae_u8*) iAddress;
	return (((uae_u32) p[0]) << 24) | (((uae_u32) p[1]) << 16) | (((uae_u32) p[2]) << 8) | p[3];
}


// ---------------------------------------------------------------------------
//		 DummyBank::GetWord
// ---------------------------------------------------------------------------

uae_u32 DummyBank::GetWord (uaecptr iAddress)
{
	// Hack to keep HwrGetRAMSize working: it runs off
	// the end of RAM while testing it.

	if (HackForHwrGetRAMSize (iAddress))
		return 0;

#if (VALIDATE_DUMMY_GET)
	if (gMemAccessFlags.fValidate_DummyGet && !ValidAddress (iAddress, sizeof (short)))
	{
		InvalidAccess (iAddress, sizeof (short), true);
	}
#endif

#if HAS_PROFILING
	if (gProfilingEnabled && gProfilingCounted)
	{
		::ProfileIncrementRead (1, WAITSTATES_DUMMYBANK);
	}
#endif

	if (!ValidAddress (iAddress, sizeof (short)))
	{
		return ~0;
	}

	uae_u8*	p = (uae_u8*) iAddress;
	return (((uae_u32) p[0]) << 8) | p[1];
}


// ---------------------------------------------------------------------------
//		 DummyBank::GetByte
// ---------------------------------------------------------------------------

uae_u32 DummyBank::GetByte (uaecptr iAddress)
{
	// Hack to keep HwrGetRAMSize working: it runs off
	// the end of RAM while testing it.

	if (HackForHwrGetRAMSize (iAddress))
		return 0;

#if (VALIDATE_DUMMY_GET)
	if (gMemAccessFlags.fValidate_DummyGet && !ValidAddress (iAddress, sizeof (char)))
	{
		InvalidAccess (iAddress, sizeof (char), true);
	}
#endif

#if HAS_PROFILING
	if (gProfilingEnabled && gProfilingCounted)
	{
		::ProfileIncrementRead (1, WAITSTATES_DUMMYBANK);
	}
#endif

	if (!ValidAddress (iAddress, sizeof (char)))
	{
		return ~0;
	}

	uae_u8*	p = (uae_u8*) iAddress;
	return p[0];
}


// ---------------------------------------------------------------------------
//		 DummyBank::SetLong
// ---------------------------------------------------------------------------

void DummyBank::SetLong (uaecptr iAddress, uae_u32 iLongValue)
{
	// Hack to keep HwrGetRAMSize working: it runs off
	// the end of RAM while testing it.

	if (HackForHwrGetRAMSize (iAddress))
		return;

#if (VALIDATE_DUMMY_SET)
	if (gMemAccessFlags.fValidate_DummySet && !ValidAddress (iAddress, sizeof (long)))
	{
		InvalidAccess (iAddress, sizeof (long), false);
	}
#endif

#if HAS_PROFILING
	if (gProfilingEnabled && gProfilingCounted)
	{
		::ProfileIncrementWrite (2, WAITSTATES_DUMMYBANK);
	}
#endif

	if (!ValidAddress (iAddress, sizeof (long)))
	{
		return;
	}

	uae_u8*	p = (uae_u8*) iAddress;
	p[0] = (uae_u8) (iLongValue >> 24);
	p[1] = (uae_u8) (iLongValue >> 16);
	p[2] = (uae_u8) (iLongValue >> 8);
	p[3] = (uae_u8) (iLongValue >> 0);
}


// ---------------------------------------------------------------------------
//		 DummyBank::SetWord
// ---------------------------------------------------------------------------

void DummyBank::SetWord (uaecptr iAddress, uae_u32 iWordValue)
{
	// Hack to keep HwrGetRAMSize working: it runs off
	// the end of RAM while testing it.

	if (HackForHwrGetRAMSize (iAddress))
		return;

#if (VALIDATE_DUMMY_SET)
	if (gMemAccessFlags.fValidate_DummySet && !ValidAddress (iAddress, sizeof (short)))
	{
		InvalidAccess (iAddress, sizeof (short), false);
	}
#endif

#if HAS_PROFILING
	if (gProfilingEnabled && gProfilingCounted)
	{
		::ProfileIncrementWrite (1, WAITSTATES_DUMMYBANK);
	}
#endif

	if (!ValidAddress (iAddress, sizeof (short)))
	{
		return;
	}

	uae_u8*	p = (uae_u8*) iAddress;
	p[0] = (uae_u8) (iWordValue >> 8);
	p[1] = (uae_u8) (iWordValue >> 0);
}


// ---------------------------------------------------------------------------
//		 DummyBank::SetByte
// ---------------------------------------------------------------------------

void DummyBank::SetByte (uaecptr iAddress, uae_u32 iByteValue)
{
	// Hack to keep HwrGetRAMSize working: it runs off
	// the end of RAM while testing it.

	if (HackForHwrGetRAMSize (iAddress))
		return;

#if (VALIDATE_DUMMY_SET)
	if (gMemAccessFlags.fValidate_DummySet && !ValidAddress (iAddress, sizeof (char)))
	{
		InvalidAccess (iAddress, sizeof (char), false);
	}
#endif

#if HAS_PROFILING
	if (gProfilingEnabled && gProfilingCounted)
	{
		::ProfileIncrementWrite (1, WAITSTATES_DUMMYBANK);
	}
#endif

	if (!ValidAddress (iAddress, sizeof (char)))
	{
		return;
	}

	uae_u8*	p = (uae_u8*) iAddress;
	p[0] = (uae_u8) (iByteValue >> 0);
}


// ---------------------------------------------------------------------------
//		 DummyBank::ValidAddress
// ---------------------------------------------------------------------------

int DummyBank::ValidAddress (uaecptr iAddress, uae_u32 iSize)
{
	int	result = false;

	// Allow access to the physical range of memory that we've
	// mapped into Palm OS-emulated space.

	for (int ii = 0; ii < kNumPhysicalBanks; ++ii)
	{
		if (gPhysicalBanks[ii].addr)
		{
			uaecptr	start = gPhysicalBanks[ii].addr;
			uaecptr	end = start + gPhysicalBanks[ii].size;

			if ((iAddress >= start) && (iAddress + iSize <= end))
			{
				result = true;
				break;
			}
		}
	}

	assert (result);

	return result;
}


// ---------------------------------------------------------------------------
//		 DummyBank::GetRealAddress
// ---------------------------------------------------------------------------

uae_u8* DummyBank::GetRealAddress (uaecptr iAddress)
{
	return (uae_u8*) iAddress;
}


// ---------------------------------------------------------------------------
//		 DummyBank::InvalidAccess
// ---------------------------------------------------------------------------

void DummyBank::InvalidAccess (uaecptr iAddress, long size, Bool forRead)
{
	Software::BusError ();
}


// ---------------------------------------------------------------------------
//		 DummyBank::MapPhysicalMemory
// ---------------------------------------------------------------------------
// Maps a range of physical memory to appear at the same location of the
// emulated Palm OS's virtual memory.

void DummyBank::MapPhysicalMemory (const void* addr, uae_u32 size)
{
	// Start mapping from longword boundaries.  It appears that MWDebug
	// rounds down to those boundaries on occassion.

	uaecptr	roundedAddress = ((uaecptr) addr) & ~0x03;
	size += ((uaecptr) addr) - roundedAddress;

	
	// Make sure that the passed in memory range doesn't overlap with
	// memory ranges already in use.  The following loop gets the
	// memory handlers for the range of memore we're mapping in, and
	// makes sure that they are DummyBank memory handlers.

	uae_s32	startIndex = bankindex(roundedAddress);
	uae_s32	endIndex = bankindex(roundedAddress + size);

	for (uae_s32 ii = startIndex; ii <= endIndex; ++ii)
	{
		assert (mem_banks[ii] == &fgAddressBank);
	}

	// We're cool...map in the range.

	int	jj;
	for (jj = 0; jj < kNumPhysicalBanks; ++jj)
	{
		// If we've already mapped this one in, just return.

		if (gPhysicalBanks[jj].addr == roundedAddress)
		{
			break;
		}

		// Found an empty record; use it.

		if (gPhysicalBanks[jj].addr == UAE_NULL)
		{
			gPhysicalBanks[jj].addr = roundedAddress;
			gPhysicalBanks[jj].size = size;
			break;
		}
	}

	if (jj == kNumPhysicalBanks)
	{
		assert (!"Too many physical memory mappings.");
		abort ();
	}
}


// ---------------------------------------------------------------------------
//		 DummyBank::UnmapPhysicalMemory
// ---------------------------------------------------------------------------
// Unmaps a range of physical memory from appearing at the same location of
// the emulated Palm OS's virtual memory.

void DummyBank::UnmapPhysicalMemory (const void* addr)
{
	// Start mapping from longword boundaries.  It appears that MWDebug
	// rounds down to those boundaries on occassion.

	uaecptr	roundedAddress = ((uaecptr) addr) & ~0x03;

	for (int ii = 0; ii < kNumPhysicalBanks; ++ii)
	{
		if (gPhysicalBanks[ii].addr == roundedAddress)
		{
			gPhysicalBanks[ii].addr = UAE_NULL;
			gPhysicalBanks[ii].size = 0;
		}
	}
}


// ---------------------------------------------------------------------------
//		 DummyBank::GetMappingInfo
// ---------------------------------------------------------------------------

void DummyBank::GetMappingInfo (const void* addr, void** start, uae_u32* len)
{
	for (int ii = 0; ii < kNumPhysicalBanks; ++ii)
	{
		if ((uaecptr) addr >= gPhysicalBanks[ii].addr &&
			(uaecptr) addr < gPhysicalBanks[ii].addr + gPhysicalBanks[ii].size)
		{
			*start = (void*) gPhysicalBanks[ii].addr;
			*len = gPhysicalBanks[ii].size;

			return;
		}
	}

	*start = NULL;
	*len = 0;
}


