/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#include "EmulatorCommon.h"
#include "CPU_REG.h"

#include "AppPreferences.h"		// gPrefs->FillStack
#include "ATraps.h"				// DoingCall
#include "Bank_MC68328.h"		// MC68328Bank::Reset, GetInterruptLevel, Cycle, etc.
#include "Bank_MC68EZ328.h"		// MC68EZ328Bank::Reset, GetInterruptLevel, Cycle, etc.


#include "Bank_ROM.h"			// ROMBank::GetMemoryStart
#include "Byteswapping.h"		// Canonical
#include "DebugMgr.h"			// Debug::Initialize
#include "ErrorHandling.h"		// kError_ stuff
#include "Gremlins.h"			// Gremlins_Initialize, Gremlins_IsOn
#include "Logging.h"			// Logging::gSystemCalls, AppendMsg
#include "MetaMemory.h"			// MetaMemory
#include "Miscellaneous.h"		// GetFunctionAddress
#include "PalmHeap.h"			// PalmHeap
#include "Platform.h"			// Initialize, Reset, DisconnectFromDebugger, GetMilliseconds, etc.
#include "RAM_ROM.h"			// Memory::Initialize, Memory::Reset, CEnableFullAccess
#include "StringData.h"			// kExceptionNames
#include "Strings.r.h"			// kStr_ values
#include "TrapPatches.h"		// Patches::Initialize, Patches::Reset, Patches::HandleTrap15
#include "UAE_Utils.h"			// uae_memset
#include "UART.h"				// UART::Initialize, Reset, Dispose

#if HAS_PROFILING
#include "Profiling.h"
#endif


// Types

#pragma mark Types


// Definitions of the stack frames used in Software::ProcessException.

/*
	Exception stack frame:

			15                       0
	sp ->	+------------------------+ <-+						 Higher Address
			| Function Code          |   |								|
			+------------------------+   |								|
			| Access Address (High)  |   | Only for Address and			|
			+------------------------+   | Bus errors.					|
			| Access Address (Low)   |   |								|
			+------------------------+   |								|
			| Instruction Register   |   |								|
			+------------------------+ <-+								|
			| Status Register        |									|
			+------------------------+									|
			| Program Counter (High) |									|
			+------------------------+									|
			| Program Counter (Low)  |									|
			+------------------------+									v

	Function code:
		Bits		Function
		0-2			Function Code
		3			I/N = Instruction = 0, Not = 1
		4			R/W = Write = 0, Read = 1
*/

#include "PalmPack.h"

struct ExceptionStackFrame1
{
	uae_u16	statusRegister;
	uae_u32	programCounter;
};

struct ExceptionStackFrame2
{
	uae_u16	functionCode;
	uae_u32	accessAddress;
	uae_u16	instructionRegister;
	uae_u16	statusRegister;
	uae_u32	programCounter;
};

#include "PalmPackPop.h"


// Globals.

#pragma mark Globals

uae_u32			gBreakReason;
uae_u32			gBreakReasonMask;
Bool			gBreakOnException[kException_LastException];
TByteQueue		gKeyQueue;
uaecptr			gLastTraceAddress;
uaecptr			gInstructionStart;

int areg_byteinc[] = { 1,1,1,1,1,1,1,2 };	// (normally in newcpu.c)
int imm8_table[] = { 8,1,2,3,4,5,6,7 };		// (normally in newcpu.c)

int movem_index1[256];						// (normally in newcpu.c)
int movem_index2[256];						// (normally in newcpu.c)
int movem_next[256];						// (normally in newcpu.c)

int fpp_movem_index1[256];					// (normally in newcpu.c)
int fpp_movem_index2[256];					// (normally in newcpu.c)
int fpp_movem_next[256];					// (normally in newcpu.c)

cpuop_func *cpufunctbl[65536];				// (normally in newcpu.c)

#if HAS_PROFILING
perfRec perftbl[65536];
#endif
											// (normally in newcpu.c)
uae_u16 last_op_for_exception_3;			/* Opcode of faulting instruction */
uaecptr last_addr_for_exception_3;			/* PC at fault time */
uaecptr last_fault_for_exception_3;			/* Address that generated the exception */

struct regstruct regs;						// (normally in newcpu.c)
struct flag_struct regflags;				// (normally in support.c)


const uae_u16	kOpcode_JMP_Abs32 = 0x4EF9;


uae_u32*	gConvert1To8;	// Used to convert a 4-bit nybble consisting of 4 1-bit pixels
							// to a 32-bit value containing 4 8-bit pixels.

uae_u32*	gConvert2To8;	// Used to convert an 8-bit byte consisting of 4 2-bit pixels
							// to a 32-bit value containing 4 8-bit pixels.

uae_u16*	gConvert4To8;	// Used to convert an 8-bit byte consisting of 2 4-bit pixels
							// to a 16-bit value containing 4 8-bit pixels.
							

#if REGISTER_HISTORY

	#define reg_history_size	512
	long	reg_history_index;
	struct regstruct	reg_history[reg_history_size];

#endif

#if EXCEPTION_HISTORY

	struct exception_history_rec
	{
		const char*		name;
		uae_u32			pc;
		uae_u32			sp;
	};
	
	#define exception_history_size	512
	long	exception_history_index;
	exception_history_rec	exception_history[exception_history_size];

#endif


#if TIME_STARTUP
	uae_u32	gStart;
#endif


#if COUNT_TRAPS
	long*	gTrapCounts;
	long	gTotalTrapCount;
#endif




#pragma mark -


// ===========================================================================
//		 Emulator
// ===========================================================================

uae_u16			Emulator::fgNextOpcode;
Bool			Emulator::fgNextOpcodeValid;
uae_u32			Emulator::fgHardwareDevice;
unsigned long	Emulator::fgInstructionCount;
unsigned long	Emulator::fgCycleCount;


/***********************************************************************
 *
 * FUNCTION:	Emulator::Initialize
 *
 * DESCRIPTION:	Initializes the emulator hardware, software, and memory.
 *				This function is called just once after a new ROM is
 *				loaded.
 *
 * PARAMETERS:	hardwareDevice - value indicating what hardware to
 *					emulate; could be "unspecified".
 *				ramFileRef - file containing saved RAM contents; could
 *					be "unspecified", in which case no RAM contents are
 *					restored.
 *				romFileRef - file containing ROM image; could be just a
 *					"big" ROM or a "big + small" ROM.
 *				ramSize - size of RAM buffer to create. Used when
 *					ramFileRef is not specified.
 *
 * RETURNED:	nothing
 *
 * REVISION HISTORY:
 *				Name	Date		Description
 *				----	----		-----------
 *
 ***********************************************************************/

void Emulator::Initialize (	uae_u32			hardwareDevice,
							FileReference&	ramFileRef,
							FileReference&	romFileRef,
							uae_s32			ramSize)
{
	// Set the hardware device here. It will get validated
	// in Memory::InitializeROM when we check the CardHeader flags.

	fgHardwareDevice = hardwareDevice;

	Memory::Initialize (ramFileRef, romFileRef, ramSize);
	Hardware::Initialize ();
	UART::Initialize (EZMode () ? UART::kUART_DragonballEZ : UART::kUART_Dragonball);
	Gremlins_Initialize ();
	Patches::Initialize ();
	Platform::Initialize ();
	Debug::Initialize ();
	Screen::Initialize ();
	PalmHeap::Initialize ();
	Software::Initialize ();

	Emulator::Reset ();
}


// ---------------------------------------------------------------------------
//		 Emulator::Reset
// ---------------------------------------------------------------------------
// Reset the emulator.  This effects a "warm restart" of the ROM.  This
// function can be called any number of times after a ROM is loaded.  It is
// also called as part of the Initialize sequence.

void Emulator::Reset (void)
{
	LowMem::fgSysDispatchTableSize = 0;
	fgNextOpcodeValid = false;

	fgInstructionCount = 0;
	fgCycleCount = fgCycleCount;

	Memory::Reset ();
	Hardware::Reset ();
	UART::Reset ();
	Gremlins_Reset ();
	Patches::Reset ();
	Platform::Reset ();
	Debug::Reset ();
	Screen::Reset ();
	PalmHeap::Reset ();
	Software::Reset ();

	Emulator::ClearBreakReason (kBreak_All);
	Emulator::SetBreakReasonMask (kBreak_All);

#if REGISTER_HISTORY
	reg_history_index = 0;
#endif

#if EXCEPTION_HISTORY
	exception_history_index = 0;
#endif

#if COUNT_TRAPS
	Platform::DisposeMemory (gTrapCounts);
#endif

	// If the appropriate modifier key is down, install a temporary breakpoint
	// at the start of the Big ROM.

	if (Platform::StopOnResetKeyDown ())
	{
		uaecptr	romStart = ROMBank::GetMemoryStart ();
		uaecptr	headerVersion = get_long (romStart + offsetof (CardHeaderType, hdrVersion));
		long	bigROMOffset = 0x03000;
		if (headerVersion > 1)
		{
			bigROMOffset = get_long (romStart + offsetof (CardHeaderType, bigROMOffset));
			bigROMOffset &= 0x000FFFFF;		// Allows for 1 Meg offset.
		}

		uaecptr	resetVector = get_long (romStart + bigROMOffset + offsetof (CardHeaderType, resetVector));

		Debug::InstallBreakpoint (resetVector);
	}
}


// ---------------------------------------------------------------------------
//		 Emulator::Dispose
// ---------------------------------------------------------------------------
// Shuts down the emulator and cleans up.

void Emulator::Dispose (void)
{
	Platform::DisconnectFromDebugger ();
	Memory::Dispose ();
	UART::Dispose ();
	PalmHeap::Dispose ();
	Software::Dispose ();
}


// ---------------------------------------------------------------------------
//		 Emulator::Execute
// ---------------------------------------------------------------------------
// Processes instructions until an event occurs.

/*
		Comments on interrupt handling:

		- SPCFLAG_INT is set at the end of MakeFromSR.  MakeFromSR is
		  called at the end of any opcode that changes it.

		- SPCFLAG_INT is also set at the end of Registers::UpdateInterrupts,
		  which is called when we want to emulate an interrupt occuring.

		- SPCFLAG_INT is set at the end of Software::ProcessInterrupt.
		  Software::ProcessInterrupt is called from the main CPU loop if
		  the requested interrupt is greater than the interrupt level.

		- In the main CPU loop, if SPCFLAG_INT or SPCFLAG_DOINT are set
		  clear them.  If the interrupt is greater than the interrupt
		  level, call Software::ProcessInterrupt, set regs.stopped to zero,
		  and clear the SPCFLAG_STOP flag.

		- Later in the main CPU loop if SPCFLAG_DOINT is set, clear it.
		  If the interrupt is greater than the interrupt level, call
		  Software::ProcessInterrupt and set regs.stopped to zero. Note
		  that SPCFLAG_STOP is NOT cleared (I don't know why).

		- After that, if SPCFLAG_INT is set, clear it and set SPCFLAG_DOINT.


		Comments on trace handling:

		- In MakeFromSR, if t0 or t1 are set, the SPCFLAG_TRACE is
		  set. Otherwise, SPCFLAG_TRACE and SPCFLAG_DOTRACE are cleared.
		  MakeFromSR is called at the end of any opcode that changes it.

		- In main CPU loop after an opcode has been executed, if
		  SPCFLAG_DOTRACE is set, call Software::ProcessException (kException_Trace)

		- At the end of Software::ProcessException(), SPCFLAG_TRACE and
		  SPCFLAG_DOTRACE are cleared.

		- Near end of main CPU loop, if SPCFLAG_TRACE is set, SPCFLAG_TRACE
		  is cleared and SPCFLAG_DOTRACE is set.
*/

long Emulator::Execute (void)
{
	// Make sure that we're entering this function with a clean slate.
	// Still allow kBreak_StopRequest, as it's possible that someone
	// started up the CPU thread and then immediately asked it to stop.

	assert ((gBreakReason & (gBreakReasonMask & ~kBreak_StopRequest)) == 0);

	// Declare this here (instead of where it's actually used), so that VC++ doesn't
	// give us any lip about jumping over it with our goto statement.

	uae_u32 opcode;

#if TIME_STARTUP
	static calledOnce;

	if (!calledOnce)
	{
		calledOnce = true;

		gStart = Platform::GetMilliseconds ();
	}
#endif

#if defined (macintosh) && defined (_DEBUG)
	// Put in a little dead-man's switch. If this function doesn't
	// exit for a long time, let us get into the debugger.
	uae_u32	deadManStart;
	if ((gBreakReasonMask & kBreak_CheckForEvents) == 0)
	{
		deadManStart = Platform::GetMilliseconds ();
	}
#endif

	// Check for the stopped flag before entering the "execute an opcode"
	// section.  It could be that we last exited the loop while still in
	// stop mode, and we need to wind our way back down to that spot.

	if ((regs.spcflags & SPCFLAG_STOP) != 0)
		goto StoppedLoop;

	while ((gBreakReason & gBreakReasonMask) == 0)
	{
#if REGISTER_HISTORY
//		if (!ATrap::DoingCall ())
		{
			// Save the registers for the post-mortem, but don't record the
			// instructions we generate when calling the ROM as a subroutine.
			// We want those to be as transparent as possible.  In particular,
			// we don't want any functions that we call as part of figuring
			// out why a problem occured to knock the problem-causing registers
			// off of our array.

			++reg_history_index;
			reg_history[reg_history_index & (reg_history_size - 1)] = regs;
		}
#endif

#if HAS_PROFILING
#if HAS_PROFILING_DEBUG
		UInt64	readCyclesSaved = 0;
		UInt64	writeCyclesSaved = 0;
		UInt64	clockCyclesSaved = 0;

		if (gProfilingEnabled)
		{
			readCyclesSaved = gReadCycles;
			writeCyclesSaved = gWriteCycles;
			clockCyclesSaved = gClockCycles;
		}
#endif

		// Turn gProfilingCounted on here so the get_iword() fetch is counted.

		gProfilingCounted = true;
#endif

#if defined (macintosh) && defined (_DEBUG)
		// Put in a little dead-man's switch. If this function doesn't
		// exit for a long time, let us get into the debugger.
		if ((gBreakReasonMask & kBreak_CheckForEvents) == 0)
		{
			uae_u32	deadManNow = Platform::GetMilliseconds ();
			if ((deadManNow - deadManStart) > 5000)
			{
				Platform::Debugger ();
			}
		}
#endif

		// Remember the starting point of this opcode.  Later, if an
		// exception occurs, we may want to reset the PC to this point.

		gInstructionStart = m68k_getpc ();

		// Get the next instruction.  Sometimes the next opcode comes
		// from fgNextOpcode (which holds the opcode in case its normal
		// location in memory is occupied by some breakpoint opcode),
		// but normally we'll fetch it with the GET_OPCODE macro.
		//
		// !!! Bug here...see comments in Debug::ConditionalBreak.

		opcode = GET_OPCODE;

		if (fgNextOpcodeValid)
		{
			opcode = fgNextOpcode;
			fgNextOpcodeValid = false;
		}

		fgInstructionCount++;

		// Execute the instruction.  Catch any exceptions thrown because of
		// address/bus errors.  In CodeWarrior, the overhead of this try/catch
		// block is minimal: the SP is saved, and there's a branch around the
		// catch block.  According to Greg, the overhead in VC++ is similar.

		try
		{
			fgCycleCount += (*cpufunctbl[opcode]) (opcode);

#if HAS_PROFILING
			if (gProfilingEnabled)
			{
#if HAS_PROFILING_DEBUG
				// Validity check on RAM_ROM stuff.

				Boolean	tryAgain = false;

				if (perftbl[opcode].readCycles != 0xFF && gReadCycles - readCyclesSaved != perftbl[opcode].readCycles)
				{
					gReadMismatch += gReadCycles - readCyclesSaved - perftbl[opcode].readCycles;
				}

				if (perftbl[opcode].writeCycles != 0xFF && gWriteCycles - writeCyclesSaved != perftbl[opcode].writeCycles)
				{
					gWriteMismatch += gWriteCycles - writeCyclesSaved - perftbl[opcode].writeCycles;
				}

				if (tryAgain)
				{
					(*cpufunctbl[opcode]) (opcode);
				}
#endif
				// Add in the extra time taken to execute the instruction.

				ProfileIncrementClock (perftbl[opcode].extraCycles);
			}
#endif

		}
		catch (ErrCode errCode)
		{
			// By the time we reach here, the break code that caused the
			// exception to be thrown should be recorded in gBreakReason.
			// Any error should also have already been reported.

			assert (errCode == kError_NoError);

			// Actually, this is not a good assert. We don't always want to
			// break if we throw an exception. An example is when we reset
			// the device, and want to abort anything else that's going on
			// after we get the state into the right state.
//			assert (gBreakReason != 0);
		}

		// Perform periodic tasks.

		Hardware::Cycle (false);

StoppedLoop:

		// Handle special conditions.  NB: the code reached by calling
		// Emulator::ExecuteSpecial used to be inline in this function.  Moving it
		// out (thus simplifying both Emulator::Execute and Emulator::ExecuteSpecial)
		// sped up the CPU loop by 9%!

		if (regs.spcflags)
		{
			Emulator::ExecuteSpecial ();
		}

#if HAS_PROFILING
		gProfilingCounted = false;
#endif
	}	// while (...)

	// Return the reason for our breaking, and return that reason to the caller.

	uae_u32	breakReason = gBreakReason;

	gBreakReason &= ~gBreakReasonMask;

	return breakReason & gBreakReasonMask;
}


// ---------------------------------------------------------------------------
//		 Emulator::ExecuteSpecial
// ---------------------------------------------------------------------------

void Emulator::ExecuteSpecial (void)
{
	// If we're tracing, execute the trace vector.

	if ((regs.spcflags & SPCFLAG_DOTRACE) && !(regs.spcflags & SPCFLAG_STOP))
	{
		Software::ProcessException (kException_Trace, gLastTraceAddress, m68k_getpc ());
	}

	if (regs.spcflags & SPCFLAG_STOP)
	{
		Emulator::ExecuteStoppedLoop ();

		// If we're still stopped, exit now; we have to handle events.

		if (regs.spcflags & SPCFLAG_STOP)
			return;
	}

	// Do trace-mode stuff (do_trace from newcpu.c does more,
	// but it's only needed for CPU_LEVEL > 0)

	if (regs.spcflags & SPCFLAG_TRACE)
	{
	    gLastTraceAddress = m68k_getpc ();
		regs.spcflags &= ~SPCFLAG_TRACE;
		regs.spcflags |= SPCFLAG_DOTRACE;
	}

	if (regs.spcflags & SPCFLAG_DOINT)
	{
		uae_s32 interruptLevel = HWRegisters::GetInterruptLevel ();
		regs.spcflags &= ~SPCFLAG_DOINT;	// was ~(SPCFLAG_INT | SPCFLAG_DOINT) in Greg and Craig, but the latest UAE has this
		if ((interruptLevel != -1) && (interruptLevel > regs.intmask))
		{
			Software::ProcessInterrupt (interruptLevel);
			regs.stopped = 0;
		}
	}

	if (regs.spcflags & SPCFLAG_INT)
	{
		regs.spcflags &= ~SPCFLAG_INT;
		regs.spcflags |= SPCFLAG_DOINT;
	}

	if (regs.spcflags & SPCFLAG_BRK)
	{
		regs.spcflags &= ~SPCFLAG_BRK;
		Emulator::SetBreakReason (kException_Trace);
	}
}


// ---------------------------------------------------------------------------
//		 Emulator::ExecuteStoppedLoop
// ---------------------------------------------------------------------------

void Emulator::ExecuteStoppedLoop (void)
{
	// While the CPU is stopped (because a STOP instruction was
	// executed) do some idle tasks.

	do {
		// Slow down processing so that the timer used
		// to increment the tickcount doesn't run too quickly.

		Platform::Delay ();

		// Perform periodic tasks.

		Hardware::Cycle (true);

		// Process an interrupt (see if it's time to wake up).

		if (regs.spcflags & (SPCFLAG_INT | SPCFLAG_DOINT))
		{
			uae_s32	interruptLevel = HWRegisters::GetInterruptLevel ();

			regs.spcflags &= ~(SPCFLAG_INT | SPCFLAG_DOINT);

			if ((interruptLevel != -1) && (interruptLevel > regs.intmask))
			{
				Software::ProcessInterrupt (interruptLevel);
				regs.stopped = 0;
				regs.spcflags &= ~SPCFLAG_STOP;
			}
		}
	} while ((regs.spcflags & SPCFLAG_STOP) && ((gBreakReason & gBreakReasonMask) == 0));
}


// ---------------------------------------------------------------------------
//		 Emulator::ExecuteUntilATrap
// ---------------------------------------------------------------------------

long Emulator::ExecuteUntilATrap (void)
{
	// Tell the CPU emulator to exit on the next "ATrap".

	CBreakOnException	breaker (kException_Trap15);

	// Run the processor until we break.

	uae_u32	oldBreakMask = Emulator::ClearBreakReasonMask (kBreak_StopRequest | kBreak_CheckForEvents);
	long	reason = Emulator::Execute ();
	Emulator::SetBreakReasonMask (oldBreakMask);

	// If a request to handle updates came in while we were running, ignore
	// it by clearing it.  We can handle it when we get back to the main
	// CPU loop.  We _don't_ want this break reason to remain registered,
	// as we're not in a position to handle it right now (we're probably
	// setting up to make a ROM subroutine call) and leaving it registered
	// will just cause the CPU loop to exit again.

	Emulator::ClearBreakReason (kBreak_CheckForEvents);

	return reason;
}


// ---------------------------------------------------------------------------
//		 Emulator::ExecuteUntilBreak
// ---------------------------------------------------------------------------
// Processes instructions until a break occurs.

long Emulator::ExecuteUntilBreak (void)
{
	// ExecuteUntilBreak is called when we're calling ROM functions as
	// subroutines.  In that context, we DON'T want to break if another
	// thread wants to break in by telling us to break on an "ATrap".

	Bool	oldBreakOn15 = SetBreakOnException (kException_Trap15, false);

	// Similarly, we don't want to break to handle events or because another
	// thread asks us to.  Our call into the ROM at this point should be
	// pretty much atomic (modulo real exceptions like bus errors).

	uae_u32	oldBreakMask = Emulator::ClearBreakReasonMask (kBreak_StopRequest | kBreak_CheckForEvents);

	// Run the processor until we break.

	long	reason;

	do {
		reason = Emulator::Execute ();

		// See if we broke on an ATrap anyway.  If we did, then it's because
		// another thread called SetBreakOnException (kException_Trap15, true)
		// after we set it to false.  Remember the fact that the other thread
		// wants to break on the next ATrap (by setting oldBreakOn15 to true),
		// and set that break flag back to false.

		if ((reason & kBreak_Exception) == kException_Trap15)
		{
			reason &= ~kBreak_Exception;
			oldBreakOn15 = true;
			SetBreakOnException (kException_Trap15, false);
		}
	}
	while (reason == 0);

	// Restore the old settings.

	Emulator::SetBreakReasonMask (oldBreakMask);
	(void) SetBreakOnException (kException_Trap15, oldBreakOn15);

	// If a request to handle updates came in while we were running, ignore
	// it by clearing it.  We can handle it when we get back to the main
	// CPU loop.  We _don't_ want this break reason to remain registered,
	// as we're not in a position to handle it right now (we're probably
	// setting up to make a ROM subroutine call) and leaving it registered
	// will just cause the CPU loop to exit again.

	Emulator::ClearBreakReason (kBreak_CheckForEvents);

	return reason;
}


// ---------------------------------------------------------------------------
//		 Emulator::SetNextOpcode
// ---------------------------------------------------------------------------

void Emulator::SetNextOpcode (uae_u16 opcode)
{
	assert (fgNextOpcodeValid == false);

	fgNextOpcode = opcode;
	fgNextOpcodeValid = true;
}


// ---------------------------------------------------------------------------
//		 Emulator::SetBreakReason
// ---------------------------------------------------------------------------

uae_u32 Emulator::SetBreakReason (uae_u32 setBits)
{
	uae_u32	oldReason = gBreakReason;

	gBreakReason |= setBits;

	return oldReason;
}


// ---------------------------------------------------------------------------
//		 Emulator::SetBreakReasonMask
// ---------------------------------------------------------------------------

uae_u32 Emulator::SetBreakReasonMask (uae_u32 setBits)
{
	uae_u32	oldReasonMask = gBreakReasonMask;

	gBreakReasonMask |= setBits;

	return oldReasonMask;
}


// ---------------------------------------------------------------------------
//		 Emulator::ClearBreakReason
// ---------------------------------------------------------------------------

uae_u32 Emulator::ClearBreakReason (uae_u32 clearBits)
{
	uae_u32	oldReason = gBreakReason;

	gBreakReason &= ~clearBits;

	return oldReason;
}


// ---------------------------------------------------------------------------
//		 Emulator::ClearBreakReasonMask
// ---------------------------------------------------------------------------

uae_u32 Emulator::ClearBreakReasonMask (uae_u32 clearBits)
{
	uae_u32	oldReasonMask = gBreakReasonMask;

	gBreakReasonMask &= ~clearBits;

	return oldReasonMask;
}


// ---------------------------------------------------------------------------
//		 Emulator::SetBreakOnException
// ---------------------------------------------------------------------------
// Specify whether or not the CPU loop should exit when the given exception
// is encountered.  If so, then the CPU loop exits with the PC pointing to
// the instruction that caused the loop to exit (and not 2, 4, or 6 bytes
// beyond it).

Bool Emulator::SetBreakOnException (uae_s32 exceptionNumber, Bool doBreak)
{
	Bool	oldBreak = gBreakOnException[exceptionNumber];
	gBreakOnException[exceptionNumber] = doBreak;
	return oldBreak;
}


// ---------------------------------------------------------------------------
//		 Emulator::SetHardwareDevice
// ---------------------------------------------------------------------------
//	Called during Emulator::Initialize to validate the hardware device
//	against the ROM.  When the user creates a new session, they specify
//	that a certain kind of hardware be emulated and that a certain ROM
//	be loaded from disk.  This function makes sure that the ROM can
//	run on the specified hardware.

void Emulator::SetHardwareDevice (uae_u32 hardwareDevice)
{
	switch (fgHardwareDevice)
	{
		// If there is no hardware type specified, yet, accept
		// the value passed in to this function.

		case kUnspecified:
			fgHardwareDevice = hardwareDevice;
			break;

		// If we were told to emulate a Dragonball device, reject
		// a ROM that runs on DragonballEZ.

		case kPilot1000:
		case kPilot5000:
		case kPalmPilotPersonal:
		case kPalmPilotProfessional:
		case kPalmPilotWithUpgrade:
		case kPalmIII:
		case kPalmVII:
		case kQualComm:
			if (hardwareDevice == kPalmIIIx ||
				hardwareDevice == kPalmV)
			{
				Errors::Throw (kError_WrongROMForType);
			}
			break;

		// If we were told to emulate a DragonballEZ device, reject
		// a ROM that runs on Dragonball.

		case kPalmIIIx:
		case kPalmV:
			if (hardwareDevice == kPilot1000 ||
				hardwareDevice == kPilot5000 ||
				hardwareDevice == kPalmPilotPersonal ||
				hardwareDevice == kPalmPilotProfessional ||
				hardwareDevice == kPalmPilotWithUpgrade ||
				hardwareDevice == kPalmIII ||
				hardwareDevice == kPalmVII)
			{
				Errors::Throw (kError_WrongROMForType);
			}
			break;
	}
}


// ---------------------------------------------------------------------------
//		 Emulator::HandleDlgButton
// ---------------------------------------------------------------------------
//	Common code to handle the result of the Continue/Debug/Reset dialog.
//
//	If the user clicked on Debug or Reset, the appropriate actions are taken
//	(PC set, variables re-initialized, etc.) and THEN AND EXCEPTION IS THROWN
//	to get us out of here.  Only if the user clicked on Continue is a normal
//	return taken.
//
//	"Emulator" isn't really the right place for this function, but I'm not
//	sure where the right place is...

void Emulator::HandleDlgButton (int button, uaecptr oldpc)
{
	// If the user clicked on Debug, simulate a bus error in order to
	// get us into the debugger.

	if (button == Errors::kDebug)
	{
//		Debug::SendMessage ("Control transferred to the debugger.");
		Debug::EnterDebugger (kException_Trap0 + sysDbgBreakpointTrapNum, oldpc);
		Emulator::SetBreakReason (kException_Trap0 + sysDbgBreakpointTrapNum);
	}
	else if (button == Errors::kReset)
	{
		Emulator::Reset();
		Errors::Scram ();
	}
	else	// ...Errors::kContinue...
	{
	}
}


#pragma mark -

// ===========================================================================
//		 Hardware
// ===========================================================================

Bool	Hardware::fgHavePenEvent;
Bool	Hardware::fgPenIsDown;
Point	Hardware::fgPenLocation;


// ---------------------------------------------------------------------------
//		 Hardware::Initialize
// ---------------------------------------------------------------------------
// One-time initialization of emulator. Called from Emulator::Initialize.

void Hardware::Initialize (void)
{
	// Initialize some CPU-related tables

	// (This initialization code is taken from init_m68k in newcpu.c)

	int		i, j;

	for (i = 0 ; i < 256 ; i++)
	{
		for (j = 0 ; j < 8 ; j++)
		{
			if (i & (1 << j))
			{
				break;
			}
		}

		movem_index1[i] = j;
		movem_index2[i] = 7-j;
		movem_next[i] = i & (~(1 << j));
	}

	for (i = 0 ; i < 256 ; i++)
	{
		for (j = 7 ; j >= 0 ; j--)
		{
			if (i & (1 << j))
			{
				break;
			}
		}

		fpp_movem_index1[i] = j;
		fpp_movem_index2[i] = 7-j;
		fpp_movem_next[i] = i & (~(1 << j));
	}

	// read_table68k is going to try to allocate a bunch of memory.
	// But it assumes that that attempt will succeed, and try using
	// that memory without checking for NULL.  This can crash you horribly
	// on a Mac, so preflight that allocation here.

    table68k = (struct instr*) Platform::AllocateMemory (65536 * sizeof (struct instr));

	Platform::DisposeMemory (table68k);

	read_table68k ();
	do_merges ();

	// The rest of this code is based on build_cpufunctbl in newcpu.c.

    unsigned long	opcode;
    struct cputbl*	tbl = op_smalltbl_3;

	for (opcode = 0; opcode < 65536; opcode++)
	{
		cpufunctbl[opcode] = op_illg;
	}

	for (i = 0; tbl[i].handler != NULL; i++)
	{
		if (!tbl[i].specific)
		{
			cpufunctbl[tbl[i].opcode] = tbl[i].handler;
#if HAS_PROFILING
			perftbl[tbl[i].opcode] = tbl[i].perf;
#endif
		}
	}

	for (opcode = 0; opcode < 65536; opcode++)
	{
		cpuop_func*	f;

		if (table68k[opcode].mnemo == i_ILLG || table68k[opcode].clev > 0)
		{
			continue;
		}

		if (table68k[opcode].handler != -1)
		{
			f = cpufunctbl[table68k[opcode].handler];
			if (f == op_illg)
			{
				abort ();
			}

			cpufunctbl[opcode] = f;
#if HAS_PROFILING
			perftbl[opcode] = perftbl[table68k[opcode].handler];
#endif

		}
	}	

	for (i = 0; tbl[i].handler != NULL; i++)
	{
		if (tbl[i].specific)
		{
			cpufunctbl[tbl[i].opcode] = tbl[i].handler;
#if HAS_PROFILING
			perftbl[tbl[i].opcode] = tbl[i].perf;
#endif
		}
	}

	// (hey readcpu doesn't free this guy!)

	Platform::DisposeMemory (table68k);
}


// ---------------------------------------------------------------------------
//		 Hardware::Reset
// ---------------------------------------------------------------------------
// Sets the registers to default values.  This occurs when the emulator boots
// for the first time or is reset. Also called directly from the RESET opcode
// handler (which would seem to bypass m68k_reset in UAE, but is not skipped
// here, since we merged m68K_reset and customreset into Hardware::Reset).

void Hardware::Reset (void)
{
	// Re-initialize some variables.

	fgHavePenEvent = false;
	fgPenIsDown = false;

	// (taken from m68k_reset in newcpu.c)

	uaecptr	romStart = ROMBank::GetMemoryStart ();
	m68k_areg (regs, 7) = get_long (romStart);
	m68k_setpc (get_long (romStart + 4));

	// Note, on the 68K, the t0 and m flags must always be zero.

	regs.kick_mask	= 0xF80000;			// (not a 68K register)
	regs.s			= 1;				// supervisor/user state
	regs.m			= 0;				// master/interrupt state
	regs.stopped	= 0;				// (not a 68K register)
	regs.t1			= 0;				// Trace enable: 00 = no trace, 10 = trace on any instruction
	regs.t0			= 0;				//		01 = trace on change of flow, 11 = undefined
	ZFLG = CFLG = NFLG = VFLG = 0;
	regs.spcflags	= 0;				// (not a 68K register)
	regs.intmask	= 7;				// disable all interrupts
	regs.vbr = regs.sfc = regs.dfc = 0;
	regs.fpcr = regs.fpsr = regs.fpiar = 0;

	// UAE doesn't do this next step, but it seems like a good idea.

	Registers::UpdateSRFromRegisters ();

	// Initialize the registers.

	HWRegisters::Reset ();
	
	
}


// ---------------------------------------------------------------------------
//		 Hardware::Cycle
// ---------------------------------------------------------------------------
// Handles periodic events that need to occur when the processor cycles (like
// updating timer registers).  This function is called in two places from
// Emulator::Execute.  Interestingly, the loop runs 3% FASTER if this function
// is in its own separate function instead of being inline.

void Hardware::Cycle (Bool sleeping)
{
	// Don't do anything if we're in the middle of an ATrap call. We don't
	// need interrupts firing or tmr counters incrementing.

	if (ATrap::DoingCall ())
	{
		return;
	}

	// Perform CPU-specific idling.

	HWRegisters::Cycle (sleeping);

	// If we're not sleeping, and there's data in the queue,
	// make sure our UART registers reflect that.

	if (!sleeping)
	{
		if (UART::GetRxQueue().GetUsed () > 0)
		{
			HWRegisters::UpdateUARTRegisters (false);
		}
	}

	// Note, the way we do this is very critical!  Previously, I'd updated
	// the screen and checked for events separately by calling two different
	// functions.  Each of these functions would check to see if an
	// appropropriate amount of time had elapsed, and perform their function
	// if it had.
	//
	// Making 4 function calls (one to the "update screen" function, one to
	// the "handle events" function, and two to LMGetTicks()) was extremely
	// expensive.  Changing this to a single call to LMTicks (in the inline
	// Platform_TimeForCycle function) speeds up the main CPU loop by 20%!

	if (Platform::TimeForCycle (sleeping))
	{
		HWRegisters::CycleSlowly (sleeping);

		// Do some platform-specific stuff.

		Platform::Cycle ();
	}
}


// ---------------------------------------------------------------------------
//		 Hardware::PenEvent
// ---------------------------------------------------------------------------
// Handles a touch-screen pen event by updating the appropriate registers.

void Hardware::PenEvent (Bool	penIsDown,
						uae_s32	penX,
						uae_s32	penY)
{
	// If Gremlins is running, do nothing. If we tried to post anything
	// now, we might be interrupting a call to EvtEnqueuePenPoint.

	if (Gremlins_IsOn ())
		return;

	if (!Patches::fgUIInitialized)
		return;

	Point	pen = { penY, penX };

	// If this is a pen-down event...

	if (penIsDown)
	{
		// If this pen-down event is the same as the last pen-down
		// event, do nothing.

		if (fgPenIsDown && pen.h == fgPenLocation.h && pen.v == fgPenLocation.v)
		{
			return;
		}

		fgPenLocation = pen;
	}

	// Remember pen up/down state for next time.

	fgPenIsDown = penIsDown;
	fgHavePenEvent = true;

	// Make sure the app's awake.  Normally, we post events on a patch to
	// SysEvGroupWait.  However, if the Palm device is already waiting,
	// then that trap will never get called.  By calling EvtWakeup now,
	// we'll wake up the Palm device from its nap.

	// Only try waking up if the interrupt level is low-enough.  When the
	// kernel is mucking with sensitive data, it sets the level to 6.  If
	// we were to call EvtWakeup at this time, we could be munging data
	// that's already in use.

//	if (regs.intmask < 6)
	{
		Errors::ReportIfPalmError (kStr_EnterPen, EvtWakeup ());
	}
}


// ---------------------------------------------------------------------------
//		 Hardware::ButtonEvent
// ---------------------------------------------------------------------------
// Handles a Palm device button event by updating the appropriate registers.

void Hardware::ButtonEvent (Bool	iButton_IsDown,
							uae_s32	iButton)
{
	// If Gremlins is running, do nothing. If we tried to post anything
	// now, we might be interrupting a call to KeyHandleInterrupt.

	if (Gremlins_IsOn ())
		return;

//	if (!Patches::fgUIInitialized)
//		return;

	HWRegisters::ButtonEvent (iButton_IsDown, iButton);
}


// ---------------------------------------------------------------------------
//		 Hardware::HotSyncEvent
// ---------------------------------------------------------------------------
// Handles a HotSync button event by updating the appropriate registers.

void Hardware::HotSyncEvent (Bool iButton_IsDown)
{
	// If Gremlins is running, do nothing. If we tried to post anything
	// now, we might be interrupting a call to KeyHandleInterrupt.

	if (Gremlins_IsOn ())
		return;

//	if (!Patches::fgUIInitialized)
//		return;

	HWRegisters::HotSyncEvent (iButton_IsDown);
}


// ---------------------------------------------------------------------------
//		 Hardware::KeyboardEvent
// ---------------------------------------------------------------------------
// Handles a key down event on the desktop keyboard by adding the key to the
// key queue.

void Hardware::KeyboardEvent (uae_u8 iKey)
{
	// If Gremlins is running, do nothing. If we tried to post anything
	// now, we might be interrupting a call to KeyHandleInterrupt.

	if (Gremlins_IsOn ())
		return;

	if (!Patches::fgUIInitialized)
		return;

	// Add the key to the queue if it isn't full.

	if (gKeyQueue.GetFree () > 0)
	{
		gKeyQueue.Put (iKey);
	}

	// Make sure the app's awake.  Normally, we post events on a patch to
	// SysEvGroupWait.  However, if the Palm device is already waiting,
	// then that trap will never get called.  By calling EvtWakeup now,
	// we'll wake up the Palm device from its nap.

	// Only try waking up if the interrupt level is low-enough.  When the
	// kernel is mucking with sensitive data, it sets the level to 6.  If
	// we were to call EvtWakeup at this time, we could be munging data
	// that's already in use.

//	if (regs.intmask < 6)
	{
		Errors::ReportIfPalmError (kStr_EnterKey, EvtWakeup ());
	}
}


#pragma mark -

// ===========================================================================
//		 Screen
// ===========================================================================

uaecptr		Screen::fgScreenLow;
uaecptr		Screen::fgScreenHigh;

uaecptr		Screen::fgScreenBegin;
uaecptr		Screen::fgScreenEnd;

Screen::BufferInfo	Screen::fgBufferInfo;

const int	kBitsPerByte = 8;		// Once we figure out how many bits are on a
									// scanline, divide by this to get the number
									// of bytes.
const int	kLCDHeight = 160;
const int	kLCDWidth = 160;


// ---------------------------------------------------------------------------
//		 Screen::Initialize
// ---------------------------------------------------------------------------

void Screen::Initialize (void)
{
	fgScreenBegin	= UAE_NULL;
	fgScreenEnd		= UAE_NULL;
}


// ---------------------------------------------------------------------------
//		 Screen::Reset
// ---------------------------------------------------------------------------

void Screen::Reset (void)
{
}


// ---------------------------------------------------------------------------
//		 Screen::InvalidateAll
// ---------------------------------------------------------------------------

void Screen::InvalidateAll (void)
{
	fgScreenLow = 0;
	fgScreenHigh = 0xFFFFFFFF;

	MetaMemory::UnmarkScreen (fgScreenBegin, fgScreenEnd);
	fgScreenBegin	= MetaMemory::GetScreenBegin ();
	fgScreenEnd		= MetaMemory::GetScreenEnd ();
	MetaMemory::MarkScreen (fgScreenBegin, fgScreenEnd);
}


// ---------------------------------------------------------------------------
//		 Screen::GetBits
// ---------------------------------------------------------------------------

#define ONE_BIT_CONVERSION_DEPTH	8
#define TWO_BIT_CONVERSION_DEPTH	8
#define FOUR_BIT_CONVERSION_DEPTH	8

Bool Screen::GetBits (BufferInfo& info, Bool incremental)
{
	// Get the current state of things.

	// Fill out info on the LCD screen.

	info.lcdBuffer		= HWRegisters::GetLCDStartAddr ();
	info.lcdDepth		= HWRegisters::GetLCDDepth ();
	info.lcdRowBytes	= HWRegisters::GetLCDRowBytes ();

	// Fill out info on the buffer we'll be returning.

	info.myBuffer		= NULL;
	info.myCLUT			= NULL;
	info.myDepth		= info.lcdDepth == 1 ? ONE_BIT_CONVERSION_DEPTH :
						  info.lcdDepth == 2 ? TWO_BIT_CONVERSION_DEPTH : FOUR_BIT_CONVERSION_DEPTH;
	info.myRowBytes	= info.lcdRowBytes * info.myDepth / info.lcdDepth;

	if (fgBufferInfo.myCLUT == NULL)
	{
		Platform::DisposeMemory (fgBufferInfo.myCLUT);
		fgBufferInfo.myCLUT = (RGBType*) Platform::AllocateMemory (sizeof(RGBType) * 256);
	}

	HWRegisters::GetLCDPalette (fgBufferInfo.myCLUT);
	
	info.myCLUT = fgBufferInfo.myCLUT;
	
	// Fill out some misc (common) info.

	info.firstLine		= 0;
	info.lastLine		= kLCDHeight;
	info.height			= HWRegisters::GetLCDHeight ();
	info.backlightOn	= HWRegisters::BacklightIsOn ();
	info.lcdOn			= HWRegisters::LCDIsOn ();

	info.visibleBounds.left		= 0 + HWRegisters::GetLCDOffset ();
	info.visibleBounds.top		= 0;
	info.visibleBounds.right	= HWRegisters::GetLCDWidth () + HWRegisters::GetLCDOffset ();
	info.visibleBounds.bottom	= HWRegisters::GetLCDHeight ();

	// If the screen is off, we can just return now.  If the LCD on/off
	// state changed since the last time, return true so that the
	// screen is updated.

	if (!info.lcdOn)
	{
		Bool somethingChanged = info.lcdOn != fgBufferInfo.lcdOn;
		fgBufferInfo.lcdOn = info.lcdOn;
		return somethingChanged;
	}

	// Get the first and last scanlines to update.  This is a *big* win when
	// running Gremlins.  Without this check, a typical test ran 240 seconds.
	// With the check, the test ran 170 seconds.

	// Get the screen begin and end.  We'll be clipping against this range (just in
	// case screenLow and/or screenHigh got out of whack, which might happen in
	// a multi-threaded system and we aren't using synchronization objects).

	uaecptr	screenBegin	= MetaMemory::GetScreenBegin ();
	uaecptr	screenEnd	= MetaMemory::GetScreenEnd ();

	// Get the range of bytes affected.  From this information, we can determine
	// the first and last affected scanlines.  After we have those values, reset
	// fgScreenLow/High with sentinel values so that they can be munged again
	// by DRAMBank::SetFoo methods.

	uaecptr	screenLow	= Screen::fgScreenLow;	Screen::fgScreenLow = screenEnd;
	uaecptr	screenHigh	= Screen::fgScreenHigh;	Screen::fgScreenHigh = screenBegin;

	// Clip screenLow and screenHigh to the range of the screen.

	if (screenLow < screenBegin)
		screenLow = screenBegin;

	if (screenHigh > screenEnd)
		screenHigh = screenEnd;

	// If anything cataclysmic has changed, set screenLow and screenHigh
	// to cover the entire screen.

	if (info.lcdBuffer				!= fgBufferInfo.lcdBuffer				||
		info.lcdDepth				!= fgBufferInfo.lcdDepth				||
		info.lcdRowBytes			!= fgBufferInfo.lcdRowBytes				||
		info.height					!= fgBufferInfo.height					||
		info.backlightOn			!= fgBufferInfo.backlightOn				||
		info.lcdOn					!= fgBufferInfo.lcdOn					||
		info.visibleBounds.left		!= fgBufferInfo.visibleBounds.left		||
		info.visibleBounds.top		!= fgBufferInfo.visibleBounds.top		||
		info.visibleBounds.right	!= fgBufferInfo.visibleBounds.right		||
		info.visibleBounds.bottom	!= fgBufferInfo.visibleBounds.bottom	||
		!incremental)
	{
		screenLow = screenBegin;
		screenHigh = screenEnd;
	}

	// If no lines need to be updated, we can return now.

	if (screenLow >= screenHigh)
	{
		return false;
	}

	// Determine the first and last changed scanlines.

	info.firstLine	= (screenLow - info.lcdBuffer) / info.lcdRowBytes;
	info.lastLine	= (screenHigh - info.lcdBuffer - 1) / info.lcdRowBytes + 1;

	// Get a buffer, in case we don't have one already, or
	// in case we need a bigger one.

	long	oldArea = fgBufferInfo.height * fgBufferInfo.myRowBytes;
	long	newArea = info.height * info.myRowBytes;

	if (fgBufferInfo.myBuffer == NULL || newArea > oldArea)
	{
		Platform::DisposeMemory (fgBufferInfo.myBuffer);
		fgBufferInfo.myBuffer = (char*) Platform::AllocateMemory (newArea);
	}
	
	info.myBuffer = fgBufferInfo.myBuffer;

	if (!gConvert1To8)
	{
		const int	kTableSize = 16;
		gConvert1To8 = (uae_u32*) Platform::AllocateMemory (sizeof (*gConvert1To8) * kTableSize);

		for (int ii = 0; ii < kTableSize; ++ii)
		{
			uae_u32	mask = 0x08;
			uae_u32	pixValue = 0x01000000;
			uae_u32	pixels = 0;

			while (mask)
			{
				if ((ii & mask) != 0)
				{
					pixels |= pixValue;
				}

				mask >>= 1;
				pixValue >>= 8;
			}

			Canonical (pixels);
			gConvert1To8[ii] = pixels;
		}
	}

	if (!gConvert2To8)
	{
		const int	kTableSize = 256;
		gConvert2To8 = (uae_u32*) Platform::AllocateMemory (sizeof (*gConvert2To8) * kTableSize);

		for (int ii = 0; ii < kTableSize; ++ii)
		{
			uae_u32	mask = 0xC0;
			uae_u32	shift = 24 - 6;
			uae_u32	pixels = 0;

			while (mask)
			{
				uae_u32	pixel = (ii & mask);
				pixel <<= shift;
				pixels |= pixel;

				mask >>= 2;
				shift -= 8 - 2;
			}

			Canonical (pixels);
			gConvert2To8[ii] = pixels;
		}
	}

	if (!gConvert4To8)
	{
		const int	kTableSize = 256;
		uae_u16 		pixels;
		
		gConvert4To8 = (uae_u16*) Platform::AllocateMemory (sizeof (*gConvert4To8) * kTableSize);

		for (int ii = 0; ii < kTableSize; ++ii)
		{
			pixels = ((ii & 0xf0) << 4) | (ii & 0x0f);
			Canonical(pixels);
			gConvert4To8[ii] = pixels;
		}
	}

	// Get the changed scanlines, converting to 8 bpp along the way.

	uae_u8*		realBaseAddress = get_real_address (info.lcdBuffer);
	uae_u32*	srcP = (uae_u32*) (realBaseAddress + info.firstLine * info.lcdRowBytes);
	uae_u32*	destP = (uae_u32*) (info.myBuffer + info.firstLine * info.myRowBytes);
	long		srcBytesToCopy = info.lcdRowBytes * (info.lastLine - info.firstLine);

	if (info.lcdDepth == 1)
	{
		if (info.myDepth == 1)
		{
			uae_u8*	srcP = (uae_u8*) (realBaseAddress + info.firstLine * info.lcdRowBytes);
			uae_u8*	destP = (uae_u8*) (info.myBuffer + info.firstLine * info.myRowBytes);

			while (srcBytesToCopy)
			{
				*destP++ = do_get_mem_byte (srcP++);

				srcBytesToCopy -= sizeof (*srcP);
			}
		}
		else if (info.myDepth == 8)
		{
			uae_u8*		srcP = (uae_u8*) (realBaseAddress + info.firstLine * info.lcdRowBytes);
			uae_u32*	destP = (uae_u32*) (info.myBuffer + info.firstLine * info.myRowBytes);

			while (srcBytesToCopy)
			{
				// Get 8 pixels.

				uae_u8	bits = do_get_mem_byte (srcP++);

				// Convert each nybble (which contains 4 1-bit pixels) to a 32-bit value
				// containing 4 8-bit pixels.

				*destP++ = gConvert1To8[(bits >>  4) & 0x0F];
				*destP++ = gConvert1To8[(bits >>  0) & 0x0F];

				srcBytesToCopy -= sizeof (*srcP);
			}
		}
		else
		{
			assert (!"Unsupported bit depth");
		}
	}
	else if (info.lcdDepth == 2)
	{
		if (info.myDepth == 2)
		{
			uae_u8*	srcP = (uae_u8*) (realBaseAddress + info.firstLine * info.lcdRowBytes);
			uae_u8*	destP = (uae_u8*) (info.myBuffer + info.firstLine * info.myRowBytes);

			while (srcBytesToCopy)
			{
				*destP++ = do_get_mem_byte (srcP++);

				srcBytesToCopy -= sizeof (*srcP);
			}
		}
		else if (info.myDepth == 8)
		{
			uae_u8*		srcP = (uae_u8*) (realBaseAddress + info.firstLine * info.lcdRowBytes);
			uae_u32*	destP = (uae_u32*) (info.myBuffer + info.firstLine * info.myRowBytes);

			while (srcBytesToCopy)
			{
				// Get 4 pixels.

				uae_u8	bits = do_get_mem_byte (srcP++);

				// Convert each byte (which contains 4 2-bit pixels) to a 32-bit value
				// containing 4 8-bit pixels.

				*destP++ = gConvert2To8[bits];

				srcBytesToCopy -= sizeof (*srcP);
			}
		}
		else
		{
			assert (!"Unsupported bit depth");
		}
	}


	else
	{
		assert (!"Unsupported bit depth");
	}

	fgBufferInfo = info;

	return true;
}


#pragma mark -

// ===========================================================================
//		 Software
// ===========================================================================

ExceptionHandlerProc	Software::fgExceptionHandlers[kException_LastException];
StackList				Software::fgStackList;


// ---------------------------------------------------------------------------
//		 Software::Initialize
// ---------------------------------------------------------------------------

void Software::Initialize (void)
{
	InstallExceptionHandler (kException_Trap0 + kATrapTailpatchTrapNum,
							Software::ProcessTrap13);
	InstallExceptionHandler (kException_Trap0 + sysDispatchTrapNum,
							Software::ProcessTrap15);
}


// ---------------------------------------------------------------------------
//		 Software::Reset
// ---------------------------------------------------------------------------

void Software::Reset (void)
{
	fgStackList.clear ();
}


// ---------------------------------------------------------------------------
//		 Software::Dispose
// ---------------------------------------------------------------------------

void Software::Dispose (void)
{
}


// ---------------------------------------------------------------------------
//		 Software::BusError
// ---------------------------------------------------------------------------
// Called by the FooBank_XXX routines when an illegal reference to memory
// is made.

void	Software::BusError(void)
{
	Software::ProcessException (kException_BusErr, gInstructionStart, m68k_getpc ());

	// Get outta here fast.  We don't try to execute the rest of the
	// instruction.

	Errors::Scram ();
}


// ---------------------------------------------------------------------------
//		 Software::AddressError
// ---------------------------------------------------------------------------
// Called by the FooBank_XXX routines when an illegal reference to memory
// is made.

void	Software::AddressError(void)
{
	Software::ProcessException (kException_AddressErr, gInstructionStart, m68k_getpc ());

	// Get outta here fast.  We don't try to execute the rest of the
	// instruction.

	Errors::Scram ();
}


// ---------------------------------------------------------------------------
//		 Software::ProcessException
// ---------------------------------------------------------------------------
// Handles an exception by updating the registers and the processor state.
//
// Note that the original PC may or may not have been bumped by now, so use
// "oldPC" to get address of the instruction that caused the exception.

void Software::ProcessException (uae_s32 iException, uaecptr oldpc, uaecptr curpc)
{
#if EXCEPTION_HISTORY
	// Save the exception for the post-mortem

	exception_history_index++;

	long	index = exception_history_index & (exception_history_size - 1);
	exception_history[index].name = kExceptionNames[iException];

	exception_history[index].pc = oldpc;
	exception_history[index].sp = m68k_areg (regs, 7);
#endif

	// See if we've been asked to break on this exception.

	if (gBreakOnException[iException])
	{
		// Back up the PC to point to where this whole mess started.

		m68k_setpc (oldpc);

		// Tell the CPU loop to stop.

		Emulator::SetBreakReason (iException);

		return;
	}

#if HAS_PROFILING
	// Don't count cycles spent in exception handlers against functions.

	if (gProfilingEnabled)
	{
		ProfileInterruptEnter (iException, curpc);
	}
#endif

	// Let any custom exception handler have a go at it.  If it returns
	// true, it handled it completely, and we don't have anything else to do.

	if (fgExceptionHandlers[iException] &&
		fgExceptionHandlers[iException] (iException, oldpc))
	{
#if HAS_PROFILING
		if (gProfilingEnabled)
		{
			ProfileInterruptExit (curpc);
		}
#endif
		return;
	}

	// The following is vaguely modelled after Exception() from newcpu.c

	// Make sure the Status Register is up-to-date.

	Registers::UpdateSRFromRegisters ();

	// If not in supervisor mode, set the usp, restore A7 from the isp,
	// and transfer to supervisor mode.

	if (!regs.s)
	{
		regs.usp = m68k_areg (regs, 7);
		m68k_areg (regs, 7) = regs.isp;
		regs.s = 1;
	}

	// Set up the stack frame.

	if (iException == kException_BusErr || iException == kException_AddressErr)
	{
		assert (sizeof (ExceptionStackFrame2) == 14);
		m68k_areg (regs, 7) -= sizeof (ExceptionStackFrame2);

		uaecptr	frame = m68k_areg (regs, 7);

		put_word (frame + offsetof (ExceptionStackFrame2, functionCode), 0);	// Eh...Palm OS doesn't use these 3 anyway...
		put_long (frame + offsetof (ExceptionStackFrame2, accessAddress), 0);
		put_word (frame + offsetof (ExceptionStackFrame2, instructionRegister), 0);

		put_word (frame + offsetof (ExceptionStackFrame2, statusRegister), regs.sr);
		put_long (frame + offsetof (ExceptionStackFrame2, programCounter), curpc);
	}
	else
	{
		assert (sizeof (ExceptionStackFrame1) == 6);
		m68k_areg (regs, 7) -= sizeof (ExceptionStackFrame1);

		uaecptr	frame = m68k_areg (regs, 7);

		put_word (frame + offsetof (ExceptionStackFrame1, statusRegister), regs.sr);
		put_long (frame + offsetof (ExceptionStackFrame1, programCounter), curpc);
	}

	{
		CEnableFullAccess	munge;	// Remove blocks on memory access.

		// Get the exception handler address and jam it in the PC.
		m68k_setpc (get_long (regs.vbr + 4 * iException));

	}

	// Turn tracing off.

	regs.t1 = regs.t0 = regs.m = 0;
	regs.spcflags &= ~(SPCFLAG_TRACE | SPCFLAG_DOTRACE);
}


// ---------------------------------------------------------------------------
//		 Software::ProcessInterrupt
// ---------------------------------------------------------------------------
// Handles an interrupt by updating the registers and the processor state.

void Software::ProcessInterrupt (uae_s32 iInterrupt)
{
	Software::ProcessException (iInterrupt + HWRegisters::GetInterruptBase (),
								m68k_getpc (), m68k_getpc ());
	regs.intmask = iInterrupt;
	regs.spcflags |= SPCFLAG_INT;

	Registers::UpdateSRFromRegisters ();	// This isn't in UAE, but it seems like
											// a good idea to do after changing intmask.
}


// ---------------------------------------------------------------------------
//		 Software::ProcessIllegalInstruction
// ---------------------------------------------------------------------------
// Handles an illegal opcode by updating the registers and the processor state.

void Software::ProcessIllegalInstruction (uae_u32 iOpcode)
{
	// This function is loosely based on op_illg in newcpu.c

    uaecptr oldpc = m68k_getpc ();
    uaecptr curpc = oldpc + 2;

	// Process an FTrap.

	if ((iOpcode & 0xF000) == 0xF000)
	{
		Software::ProcessException (kException_FTrap, oldpc, curpc);
		return;
	}

	// Process an ATrap.

	if ((iOpcode & 0xF000) == 0xA000)
	{
		Software::ProcessException (kException_ATrap, oldpc, curpc);
		return;
	}

	Software::ProcessException (kException_IllegalInstr, oldpc, curpc);
}


// ---------------------------------------------------------------------------
//		 Software::InstallExceptionHandler
// ---------------------------------------------------------------------------
// Specified a function to be called when the given exception occurs.

void Software::InstallExceptionHandler (uae_s32 exceptionNumber,
										ExceptionHandlerProc fn)
{
	fgExceptionHandlers[exceptionNumber] = fn;
}


// ---------------------------------------------------------------------------
//		 Software::RemoveExceptionHandler
// ---------------------------------------------------------------------------

void Software::RemoveExceptionHandler (uae_s32 exceptionNumber)
{
	fgExceptionHandlers[exceptionNumber] = NULL;
}


// ---------------------------------------------------------------------------
//		 Software::ProcessJSR
// ---------------------------------------------------------------------------

Bool Software::ProcessJSR (uaecptr oldpc, uaecptr dest)
{
#if HAS_PROFILING
	if (gProfilingEnabled)
	{
		StDisableMemoryCycleCounting	stopper;

		// In multi-segmented application generated by CodeWarrior,
		// inter-segment function calls are implemented by JSR'ing
		// to a jump table entry.  This jump table entry contains
		// a JMP to the final location.  Detect this case and emit
		// a function entry record for the final location, not the
		// jump table entry.
		uae_u8*	realMem = get_real_address (dest);
		if (do_get_mem_word (realMem) == kOpcode_JMP_Abs32)
		{
			dest = do_get_mem_long (realMem + 2);
		}

		ProfileFnEnter (dest, oldpc);	// "oldpc" contains return address
	}
#endif

	return false;	// We didn't completely handle it; do default handling.
}


// ---------------------------------------------------------------------------
//		 Software::ProcessJSR_Ind
// ---------------------------------------------------------------------------
//	Check for SYS_TRAP_FAST calls.

Bool Software::ProcessJSR_Ind (uaecptr oldpc)
{
	Bool	handledIt = false;	// Default to calling ROM.

	//	inline asm SysTrapFast(Int trapNum)
	//	{
	//			MOVE.L	struct(LowMemType.fixed.globals.sysDispatchTableP), A1
	//			MOVE.L	((trapNum-sysTrapBase)*4)(A1), A1
	//			JSR		(A1)																// call it
	//	}
	//
	//	#define SYS_TRAP_FAST(trapNum) \
	//		FIVEWORD_INLINE( \
	//			0x2278, 0x0122, \
	//			0x2269, (trapNum-sysTrapBase)*4, \
	//			0x4e91)

	uae_u8*	realMem = get_real_address (oldpc);
	if (do_get_mem_word (realMem) == 0x4e91 &&
		do_get_mem_word (realMem - 4) == 0x2269 &&
		do_get_mem_word (realMem - 8) == 0x2278)
	{
		handledIt = Software::ProcessSystemCall (oldpc);
	}

	return handledIt;
}


// ---------------------------------------------------------------------------
//		 Software::ProcessRTS
// ---------------------------------------------------------------------------

Bool Software::ProcessRTS (void)
{
#if HAS_PROFILING
	if (gProfilingEnabled)
	{
		StDisableMemoryCycleCounting	stopper;
		ProfileFnExit (get_long (m68k_areg (regs, 7)), m68k_getpc () + 2);
	}
#endif

	return false;	// We didn't completely handle it; do default handling.
}


// ---------------------------------------------------------------------------
//		 Software::ProcessRTE
// ---------------------------------------------------------------------------

Bool Software::ProcessRTE (uaecptr newpc)
{
#if HAS_PROFILING
	if (gProfilingEnabled)
	{
		ProfileInterruptExit (newpc);
	}
#endif

	return false;	// We didn't completely handle it; do default handling.
}


// ---------------------------------------------------------------------------
//		 Software::ProcessLINK
// ---------------------------------------------------------------------------

Bool Software::ProcessLINK (uae_s32 linkSize)
{
	if (gPrefs->FillStack () && linkSize < 0)
	{
		uae_memset (m68k_areg (regs, 7), 0xD5, -linkSize);
	}

	return false;	// We didn't completely handle it; do default handling.
}


// ---------------------------------------------------------------------------
//		 Software::ProcessTrap13	
// ---------------------------------------------------------------------------
// Handle a tail patch. Traps that we want to tail patch have a TRAP D stored
// right after the 4-byte trap sequence used to invoke them.
// Patches::HandleTrap13 will restore the opcode removed when we put the
// TRAP D there, back up the PC so that we can re-execute that opcode, and
// call the tail patch handler.

Bool Software::ProcessTrap13 (uae_s32 exceptionNumber, uaecptr oldpc)
{
	Patches::HandleTrap13 (oldpc);

	// Backup the PC so that we can re-execute the opcode at this
	// location.

	m68k_setpc (oldpc);

	return true;
}


// ---------------------------------------------------------------------------
//		 Software::ProcessTrap15
// ---------------------------------------------------------------------------
// Handle a trap. Traps are of the format TRAP $F / $Axxx.

Bool Software::ProcessTrap15 (uae_s32 exceptionNumber, uaecptr oldpc)
{
	return Software::ProcessSystemCall (oldpc);
}


/***********************************************************************
 *
 * FUNCTION:	Software::ProcessSystemCall
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	
 *
 * RETURNED:	
 *
 ***********************************************************************/

Bool Software::ProcessSystemCall (uaecptr startPC)
{
	// ======================================================================
	//	Determine what ROM function is about to be called, and determine
	//	the method by which it is being called.
	// ======================================================================

	// Determine how the system function is being called.  There are two ways:
	//
	//		* Via SYS_TRAP macro:
	//
	//			TRAP	$F
	//			DC.W	$Axxx
	//
	//		* Via SYS_TRAP_FAST macro:
	//
	//			MOVE.L	struct(LowMemType.fixed.globals.sysDispatchTableP), A1
	//			MOVE.L	((trapNum-sysTrapBase)*4)(A1), A1
	//			JSR		(A1)
	//
	// The PC is current pointing to either the TRAP $F or the JSR (A1),
	// so we can look at the opcode to determine how we got here.

	uae_u8*	realMem = get_real_address (startPC);
	Bool	viaTrap = do_get_mem_word (realMem) == (m68kTrapInstr + sysDispatchTrapNum);


	// Get the trap word (the $Axxx value) and turn it into a zero-based index.

	uae_u16	trapWord;
	uaecptr	returnAddress;
	
	if (viaTrap)
	{
		trapWord = do_get_mem_word (realMem + 2);
		returnAddress = startPC + 4;
	}
	else
	{
		trapWord = (do_get_mem_word (realMem - 2) / 4) | sysTrapBase;
		returnAddress = startPC + 2;
	}

	uae_u16	trapIndex = SysTrapIndex (trapWord);


	// Not all development systems generate the correct dispatch
	// numbers; some leave off the preceding "A".  Make sure it's
	// set so that we can recognize it as a trap dispatch number.
	// (This code is here specifically so that the profiling routines
	//  will work, which check for trap numbers masquerading as function
	//  addresses by checking to see if they are in the sysTrapBase range.)

	if (trapWord == trapIndex)
		trapWord |= sysTrapBase;

	assert (trapWord >= sysTrapBase && trapWord < sysTrapBase + 0x1000);

	// ======================================================================
	//	Log the fact that this ROM function was being called.  This is
	//	internal debugging information only.
	// ======================================================================

#if COUNT_TRAPS
	// Record this trap call, allocating the table used to hold that
	// information if necessary.

	if (LowMem::TrapExists (trapIndex))
	{
		if (gTrapCounts == NULL)
		{
			gTrapCounts = new long[LowMem::fgSysDispatchTableSize];
			memset (gTrapCounts, 0, sizeof(long) * LowMem::fgSysDispatchTableSize);
		}

		gTrapCounts[trapIndex]++;
		gTotalTrapCount++;
	}
#endif

	// ======================================================================
	//	Validate the dispatch number for the ROM function we're about to call.
	// ======================================================================

	uaecptr	address = ValidateAndGetFunctionAddress (trapWord, startPC);
	if (address == UAE_NULL)
	{
		// *I* don't know what to do with this trap...maybe the ROM
		// does.  But I don't expect we'll ever get here anyway.

		return kExecuteROM;
	}

	// ======================================================================
	//	Record what function we're calling.
	// ======================================================================

	if (LogSystemCalls () && !ATrap::DoingCall ())
	{
		char	name [sysPktMaxNameLen];
		uae_u32	extra;

		if (IsSystemTrap (trapWord))
		{
			extra = m68k_dreg (regs, 2);
		}
		else
		{
			extra = get_word (m68k_areg (regs, 7));
		}

		strcpy (name, ::GetTrapName (trapWord, extra, true));

		AppendMsg ("--- System Call 0x%04X: %s.", (long) trapWord, name);
	}

	// ======================================================================
	// Let the debugger have a crack at it.  It may want to do a "break
	// on ATrap" thingy.  If so, it will return true.  Otherwise,
	// if no action is necessary, it will return false.
	// ======================================================================

	if (Debug::HandleSystemCall (trapWord, startPC))
	{
		return kSkipROM;
	}


	// ======================================================================
	//	Tell the profiler that we're entering a function.
	// ======================================================================
#if HAS_PROFILING
	// This ProfileFnEnter will show the function call within the trap handler
	// exception, which gives us a nice breakdown of which traps are being
	// called and frequency.

	if (gProfilingEnabled)
	{
		StDisableMemoryCycleCounting	stopper;
		ProfileFnEnter (trapWord, returnAddress);
	}
#endif

	// ======================================================================
	// If this trap is patched, let the patch handler handle the patch.
	// ======================================================================

	CallROMType	result = Patches::HandleSystemCall (trapWord, startPC);


	// ======================================================================
	//	If we completely handled the function in head and tail patches, tell
	//	the profiler that we exited the function and get out of here.
	// ======================================================================
	
	if (result == kSkipROM)
	{
#if HAS_PROFILING
		if (gProfilingEnabled)
		{
			StDisableMemoryCycleCounting	stopper;
			ProfileFnExit (returnAddress, trapWord);
		}
#endif

		// Set the PC to point past the instructions that go us here.

		m68k_setpc (returnAddress);

		// Return true to say that everything has been handled.

		return true;
	}


	// ======================================================================
	//	If we're profiling, don't dispatch to the ROM function outselves.
	//	We want the ROM to do it so that we get accurate dispatch times.
	// ======================================================================

#if HAS_PROFILING
	// Don't do native dispatching if the profiler is on!

	if (gProfilingEnabled)
	{
		// Return false to do default exception handler processing.

		return false;
	}
#endif

	// ======================================================================
	// Otherwise, let's run the trap dispatcher native.  This gives
	// us about a 10-12% speed-up.
	// ======================================================================

	// Push the return address onto the stack. Subtract 4 from the stack, and then
	// store the appropriate return address.

	uaecptr oldA7 = m68k_areg(regs, 7);

	m68k_areg (regs, 7) -= 4;
	put_long (m68k_areg (regs, 7), returnAddress);

	// Change to the new PC.

	assert (address != UAE_NULL);
	m68k_setpc (address);

	Software_CheckStackPointer (0, oldA7);

	return true;	// Return true to say that everything has been handled.
}


/***********************************************************************
 *
 * FUNCTION:	Software::ValidateAndGetFunctionAddress
 *
 * DESCRIPTION:	Checks to see if "trapWord" represents an implemented
 *				function or not.
 *
 * PARAMETERS:	trapWord - dispatch number to check
 *
 * RETURNED:	The address of the function if it is implemented, NULL
 *				if it isn't.  NOTE: this function may not return at
 *				all; it may just throw an exception.
 *
 ***********************************************************************/

uaecptr Software::ValidateAndGetFunctionAddress (uae_u16 trapWord, uaecptr oldpc)
{
	uaecptr	result;
	uae_u32	errorBase;

	// ======================================================================
	//	Get the function's address.
	// ======================================================================

	// See if it's a non-library call (that is, a call in the A000-A7FF range).

	if (IsSystemTrap (trapWord))
	{
		result = ::GetFunctionAddress (trapWord, m68k_dreg (regs, 2), true);
		errorBase = kStr_SysTrapBase;

	}

	// It's a library call.

	else
	{
		Word	refNum = get_word (m68k_areg (regs, 7));
		result = ::GetFunctionAddress (trapWord, refNum, true);
		errorBase = gLibErrorBase;
	}

	// ======================================================================
	//	Report an error if we can't get the address.
	// ======================================================================

	if (result == UAE_NULL)
	{
		int button = Errors::ReportUnimplementedTrap (trapWord, errorBase);

		// (Exception is thrown if button == kDebug or kReset).

		Emulator::HandleDlgButton (button, oldpc);

		return NULL;	// Never reached.  HandleDlgButton will throw
						// an exception if the user selects Debug or
						// Reset, the only two buttons we offer them.
	}

	return result;
}


// ---------------------------------------------------------------------------
//		 Software::CheckStackPointer
// ---------------------------------------------------------------------------
//	!!! This operation slows down Gremlins by about 10%; it should be sped up
//	or made an option.

void Software::CheckStackPointer (int opcodeType, uaecptr oldA7)
{
	// See if the stack pointer changed. If not, just leave.

	if (oldA7 == m68k_areg (regs, 7))
		return;

	uaecptr	curA7 = m68k_areg (regs, 7);

	// See what kind of operation just changed the stack pointer.

	if (opcodeType == kMoveAOpcode || opcodeType == kEXGOpcode || opcodeType == kLEAOpcode)
	{
		// If it was a "drastic" change, assume that we're *setting* the
		// stack pointer to a new stack.  Scarf up information about that
		// block of memory and treat that block as a stack.

		if (!MetaMemory::HeapsInitialized ())	// If heaps not set up yet, just leave.
			return;

		// See if we already know about this stack.

		StackList::iterator	iter = FindStackChunk (curA7);

		// If not, get some information about it and save it off

		if (iter == fgStackList.end ())
		{
			// !!! Should ensure that the memory semaphore is not acquired;
			// the heap may be in flux at this point...

			PalmHeap::ChunkInfo	chunkInfo = PalmHeap::GetChunkContaining (curA7);

			// Couldn't find the chunk? Just leave...

			if (chunkInfo.version == 0)
				return;

			// The chunk is free? Just leave...assume this is the SysAppExit bug.
			// If it isn't, we'll catch it soon enough when something tries
			// to use the stack.

			if (chunkInfo.free)
				return;

			RememberStackChunk (chunkInfo);
		}
	}
	else
	{
		// If it was an "incremental" change, make sure that A7 is not
		// now outside the stack it used to be in (as indicated by oldA7).

		// If we haven't yet recorded any memory chunks as stack, then assume
		// that we're still booting up and using a range of memory that was
		// "allocated" before the memory manager was started up.  Under that
		// circumstance, we don't really know the stack range, so allow
		// any sort of access.

		if (fgStackList.empty ())
			return;

		// Find the stack indicated by oldA7

		StackList::iterator	iter = FindStackChunk (oldA7);

		// If we couldn't find it, just exit (what else to do?)

		if (iter == fgStackList.end ())
			return;

		// Make sure that the current A7 is in the stack, too.

		const int	kStackSlush = 100;	// Warn when we get within this many bytes
										// of the bottom of the stack.

		// Test for "close to the end" (in which case we display a warning), and
		// "past the end" (in which case we print a fatal error).

		Bool	warning = curA7 <= (*iter).fBottom + kStackSlush;
		Bool	fatal = curA7 <= (*iter).fBottom;

		if (warning)
		{
			int	button = Errors::ReportStackOverflow (fatal);
			Emulator::HandleDlgButton (button, m68k_getpc ());
		}
	}
}


// ---------------------------------------------------------------------------
//		 Software::RememberStackChunk
// ---------------------------------------------------------------------------

Bool Software::RememberStackChunk (const PalmHeap::ChunkInfo& chunkInfo)
{
	// Get the range of the chunk.  (!!! Remember just the body,
	// or the header and any trailer, too?)

	uaecptr	chunkStart = chunkInfo.chunkHdrStart;	// + chunkInfo.chunkHdrSize;
	uaecptr	chunkSize = chunkInfo.size;	// - chunkInfo.chunkHdrSize - chunkInfo.sizeAdj;

	// See if this range is already in the list.

	StackList::iterator	iter = FindStackChunk (chunkStart);

	// If so, do nothing and return true.

	if (iter != fgStackList.end ())
		return true;	// Already there

	// Create a range object for the chunk.

	StackRange	range (chunkStart, chunkStart + chunkSize);

	// Add the range to the list and return false.

	fgStackList.insert (iter, range);

	return false;	// Chunk wasn't already there.
}


// ---------------------------------------------------------------------------
//		 Software::ForgetStackChunk
// ---------------------------------------------------------------------------

void Software::ForgetStackChunk (uaecptr chunkStart)
{
	// See if this chunk is already in the list.

	StackList::iterator	iter = FindStackChunk (chunkStart);

	// If so, remove it.

	if (iter != fgStackList.end ())
	{
		fgStackList.erase (iter);
	}
}


// ---------------------------------------------------------------------------
//		 Software::FindStackChunk
// ---------------------------------------------------------------------------

StackList::iterator Software::FindStackChunk (uaecptr testAddress)
{
	StackList::iterator	iter = fgStackList.begin ();

	for (; iter != fgStackList.end (); ++iter)
	{
		StackRange	range = *iter;
		if ((testAddress >= range.fBottom) && (testAddress < range.fTop))
		{
			break;
		}
	}

	return iter;
}


#pragma mark -

// ===========================================================================
//		 LowMem
// ===========================================================================

uae_u16		LowMem::fgSysDispatchTableSize;


// ---------------------------------------------------------------------------
//		 LowMem::GetEvtMgrIdle
// ---------------------------------------------------------------------------

uae_u8 LowMem::GetEvtMgrIdle (void)
{
	CEnableFullAccess	munge;	// Remove blocks on memory access.

	// First, get sysEvtMgrGlobalsP, which is a pointer in low-memory.

	uae_u32	sysEvtMgrGlobalsP	= LowMem_GetGlobal (sysEvtMgrGlobalsP);

	/*
	**	Next look at the "idle" boolean in the SysEvtMgrGlobalsPtr
	**	contained in sysEvtMgrGlobalsP:
	**
	**	struct SysEvtMgrGlobalsType
	**	{
	**		...
	**		Boolean			idle;
	**		...
	**	};
	*/

	size_t	idleOffset = offsetof (SysEvtMgrGlobalsType, idle);

	// Under Palm OS 1.0, there was an extra 4-byte value at the
	// location currently holding "gremlinsFlag", which is before
	// the "idle" field.  Under that version of the OS, we therefore
	// have to add 4 to get the right offset.

	if (Patches::OSMajorVersion () == 1)
	{
		idleOffset += sizeof (Long);
	}

	uae_u8	idle = get_byte (sysEvtMgrGlobalsP + idleOffset);

	return idle;
}


// ---------------------------------------------------------------------------
//		 LowMem::TrapExists
// ---------------------------------------------------------------------------

Bool LowMem::TrapExists (uae_u16 iATrap)
{
	// If we haven't already, get the number of traps in this system.

	if (fgSysDispatchTableSize == 0)
	{
		CEnableFullAccess	munge;	// Remove blocks on memory access.

		fgSysDispatchTableSize = LowMem_GetGlobal (sysDispatchTableSize);
	}

	return (SysTrapIndex (iATrap) == SysTrapIndex (sysTrapHostControl)) ||
			SysTrapIndex (iATrap) < fgSysDispatchTableSize;
}


// ---------------------------------------------------------------------------
//		 LowMem::GetTrapAddress
// ---------------------------------------------------------------------------

uaecptr LowMem::GetTrapAddress (uae_u16 iATrap)
{
	uaecptr	address = UAE_NULL;

	CEnableFullAccess	munge;	// Remove blocks on memory access.

	// If the trap number is not in range, just return NULL to signal
	// that normal process of the TRAP $F should occur.

	if (TrapExists (iATrap))
	{
		// Get sysDispatchTableP, which is a pointer in low-memory.

		uae_u32	sysDispatchTableP	= LowMem_GetGlobal (sysDispatchTableP);

		// If it's NULL, return NULL.

		if (sysDispatchTableP != UAE_NULL)
		{
			// Next, get the right guy from its contents.

			address = get_long (sysDispatchTableP + SysTrapIndex (iATrap) * 4);
		}
	}

	return address;
}


#pragma mark -

// ===========================================================================
//		 Registers
// ===========================================================================

// ---------------------------------------------------------------------------
//		 Registers::GetRegisters
// ---------------------------------------------------------------------------

void Registers::GetRegisters (regstruct& oRegisters)
{
	Registers::UpdateSRFromRegisters ();

	oRegisters = regs;
	oRegisters.pc = m68k_getpc ();
}


// ---------------------------------------------------------------------------
//		 Registers::SetRegisters
// ---------------------------------------------------------------------------

void Registers::SetRegisters (regstruct& iRegisters)
{
	regs = iRegisters;
	Registers::UpdateRegistersFromSR ();
	m68k_setpc (iRegisters.pc);
}


// ---------------------------------------------------------------------------
//		 Registers::UpdateSRFromRegisters
// ---------------------------------------------------------------------------
// Updates the status register to reflect the state of the other registers.

void Registers::UpdateSRFromRegisters (void)
{
	// (taken from MakeSR in newcpu.c)

	regs.sr = ((regs.t1 << 15) | (regs.t0 << 14)
			| (regs.s << 13) | (regs.m << 12) | (regs.intmask << 8)
			| (XFLG << 4) | (NFLG << 3) | (ZFLG << 2) | (VFLG << 1) 
			|  CFLG);
}


// ---------------------------------------------------------------------------
//		 Registers::UpdateRegistersFromSR
// ---------------------------------------------------------------------------
// Updates the registers to reflect the state of the status register.

void Registers::UpdateRegistersFromSR (void)
{
	// (taken from MakeFromSR in newcpu.c)
	
	int oldm = regs.m;
	int olds = regs.s;
	
	regs.t1 = (regs.sr >> 15) & 1;
	regs.t0 = (regs.sr >> 14) & 1;
	regs.s = (regs.sr >> 13) & 1;
	regs.m = (regs.sr >> 12) & 1;
	regs.intmask = (regs.sr >> 8) & 7;

	XFLG = (regs.sr >> 4) & 1;
	NFLG = (regs.sr >> 3) & 1;
	ZFLG = (regs.sr >> 2) & 1;
	VFLG = (regs.sr >> 1) & 1;
	CFLG = regs.sr & 1;

	if (olds != regs.s)
	{
		if (olds)
		{
			regs.isp = m68k_areg(regs, 7);
			m68k_areg(regs, 7) = regs.usp;
		}
		else
		{
			regs.usp = m68k_areg(regs, 7);
			m68k_areg(regs, 7) = regs.isp;
		}
	}
	
	regs.spcflags |= SPCFLAG_INT;

	if (regs.t1 || regs.t0)
	{
		regs.spcflags |= SPCFLAG_TRACE;
	}
	else
	{
		regs.spcflags &= ~(SPCFLAG_TRACE | SPCFLAG_DOTRACE);
	}
}


#pragma mark -

// ===========================================================================
//		 HWRegisters
// ===========================================================================

uae_u32	HWRegisters::fgCurrentMilliseconds;


// ---------------------------------------------------------------------------
//		 HWRegisters::Reset
// ---------------------------------------------------------------------------

void HWRegisters::Reset (void)
{
	if (Emulator::EZMode ())
		MC68EZ328Bank::Reset ();
	else
		MC68328Bank::Reset ();
}


// ---------------------------------------------------------------------------
//		 HWRegisters::Cycle
// ---------------------------------------------------------------------------

void HWRegisters::Cycle (Bool sleeping)
{
	if (Emulator::EZMode ())
		MC68EZ328Bank::Cycle (sleeping);
	else
		MC68328Bank::Cycle (sleeping);
}


// ---------------------------------------------------------------------------
//		 HWRegisters::CycleSlowly
// ---------------------------------------------------------------------------

void HWRegisters::CycleSlowly (Bool sleeping)
{
	if (Emulator::EZMode ())
		MC68EZ328Bank::CycleSlowly (sleeping);
	else
		MC68328Bank::CycleSlowly (sleeping);
}


// ---------------------------------------------------------------------------
//		 HWRegisters::GetInterruptLevel
// ---------------------------------------------------------------------------

uae_s32 HWRegisters::GetInterruptLevel (void)
{
	if (Emulator::EZMode ())
		return MC68EZ328Bank::GetInterruptLevel ();

	return MC68328Bank::GetInterruptLevel ();
}


// ---------------------------------------------------------------------------
//		 HWRegisters::GetInterruptBase
// ---------------------------------------------------------------------------

uae_s32 HWRegisters::GetInterruptBase (void)
{
	if (Emulator::EZMode ())
		return MC68EZ328Bank::GetInterruptBase ();

	return MC68328Bank::GetInterruptBase ();
}


// ---------------------------------------------------------------------------
//		 HWRegisters::ButtonEvent
// ---------------------------------------------------------------------------

void HWRegisters::ButtonEvent (Bool iButton_IsDown, uae_s32 iButton)
{
	if (Emulator::EZMode ())
		MC68EZ328Bank::ButtonEvent (iButton_IsDown, iButton);
	else
		MC68328Bank::ButtonEvent (iButton_IsDown, iButton);
}


// ---------------------------------------------------------------------------
//		 HWRegisters::HotSyncEvent
// ---------------------------------------------------------------------------

void HWRegisters::HotSyncEvent (Bool iButton_IsDown)
{
	if (Emulator::EZMode ())
		MC68EZ328Bank::HotSyncEvent (iButton_IsDown);
	else
		MC68328Bank::HotSyncEvent (iButton_IsDown);
}


// ---------------------------------------------------------------------------
//		 HWRegisters::GetLCDDepth
// ---------------------------------------------------------------------------

int HWRegisters::GetLCDDepth (void)
{

	if (Emulator::EZMode ())
		return MC68EZ328Bank::GetLCDDepth ();

	return MC68328Bank::GetLCDDepth ();
}


// ---------------------------------------------------------------------------
//		 HWRegisters::GetLCDRowBytes
// ---------------------------------------------------------------------------

int HWRegisters::GetLCDRowBytes (void)
{

	if (Emulator::EZMode ())
		return MC68EZ328Bank::GetLCDRowBytes ();

	return MC68328Bank::GetLCDRowBytes ();
}


// ---------------------------------------------------------------------------
//		 HWRegisters::GetLCDWidth
// ---------------------------------------------------------------------------

int HWRegisters::GetLCDWidth (void)
{

	if (Emulator::EZMode ())
		return MC68EZ328Bank::GetLCDWidth ();

	return MC68328Bank::GetLCDWidth ();
}


// ---------------------------------------------------------------------------
//		 HWRegisters::GetLCDHeight
// ---------------------------------------------------------------------------

int HWRegisters::GetLCDHeight (void)
{

	if (Emulator::EZMode ())
		return MC68EZ328Bank::GetLCDHeight ();

	return MC68328Bank::GetLCDHeight ();
}


// ---------------------------------------------------------------------------
//		 HWRegisters::GetLCDStartAddr
// ---------------------------------------------------------------------------

uaecptr HWRegisters::GetLCDStartAddr (void)
{

	if (Emulator::EZMode ())
		return MC68EZ328Bank::GetLCDStartAddr ();

	return MC68328Bank::GetLCDStartAddr ();
}


// ---------------------------------------------------------------------------
//		 HWRegisters::GetLCDPalette
// ---------------------------------------------------------------------------

void HWRegisters::GetLCDPalette (Screen::RGBType* thePalette)
{

#if 1
	const Screen::RGBType	kRGB_ScreenGreen		= { 0x7B, 0x8C, 0x5A };
	const Screen::RGBType	kRGB_ScreenLightGreen	= { 0x64, 0xF0, 0xDC };
	const Screen::RGBType	kRGB_ScreenWhite		= { 0xFF, 0xFF, 0xFF };
	const Screen::RGBType	kRGB_ScreenBlack		= { 0x00, 0x00, 0x00 };

	int	numColors;
	if (GetLCDDepth () == 1)
		numColors = 2;
	else
		numColors = 4;

	Screen::RGBType	background, foreground = kRGB_ScreenBlack;
	if (gPrefs->WhiteBackground ())
		background = kRGB_ScreenWhite;
	else if (BacklightIsOn ())
		background = kRGB_ScreenLightGreen;
	else
		background = kRGB_ScreenGreen;

	long	br = ((long) background.r);
	long	bg = ((long) background.g);
	long	bb = ((long) background.b);

	long	dr = ((long) foreground.r) - ((long) background.r);
	long	dg = ((long) foreground.g) - ((long) background.g);
	long	db = ((long) foreground.b) - ((long) background.b);

	for (int color = 0; color < numColors; ++color)
	{
		thePalette[color].r	= (Byte) (br + dr * color / (numColors - 1));
		thePalette[color].g	= (Byte) (bg + dg * color / (numColors - 1));
		thePalette[color].b	= (Byte) (bb + db * color / (numColors - 1));
	}
#else
	if (Emulator::EZMode ())
	{
		MC68EZ328Bank::GetLCDPalette (thePalette);
		return;
	}

	MC68328Bank::GetLCDPalette (thePalette);
#endif
}


// ---------------------------------------------------------------------------
//		 HWRegisters::GetLCDOffset
// ---------------------------------------------------------------------------

int HWRegisters::GetLCDOffset (void)
{

	if (Emulator::EZMode ())
		return MC68EZ328Bank::GetLCDOffset ();

	return MC68328Bank::GetLCDOffset ();
}


// ---------------------------------------------------------------------------
//		 HWRegisters::LCDIsOn
// ---------------------------------------------------------------------------

Bool HWRegisters::LCDIsOn (void)
{

	if (Emulator::EZMode ())
		return MC68EZ328Bank::LCDIsOn ();

	return MC68328Bank::LCDIsOn ();
}


// ---------------------------------------------------------------------------
//		 HWRegisters::BacklightIsOn
// ---------------------------------------------------------------------------

Bool HWRegisters::BacklightIsOn (void)
{

	if (Emulator::EZMode ())
		return MC68EZ328Bank::BacklightIsOn ();

	return MC68328Bank::BacklightIsOn ();
}


// ---------------------------------------------------------------------------
//		 HWRegisters::TurnSoundOff
// ---------------------------------------------------------------------------

void HWRegisters::TurnSoundOff (void)
{
	if (Emulator::EZMode ())
		MC68EZ328Bank::TurnSoundOff ();
	else
		MC68328Bank::TurnSoundOff ();
}


// ---------------------------------------------------------------------------
//		 HWRegisters::UpdateUARTRegisters
// ---------------------------------------------------------------------------

void HWRegisters::UpdateUARTRegisters (Bool refreshRxData)
{
	if (Emulator::EZMode ())
		MC68EZ328Bank::UpdateUARTState (refreshRxData);
	else
		MC68328Bank::UpdateUARTState (refreshRxData);
}


#pragma mark -

// ---------------------------------------------------------------------------
// Glue functions to bridge from UAE-generated code to the implemenations
// we define.
// ---------------------------------------------------------------------------

void customreset (void)
{
	Emulator::Reset ();
}


void Exception (int nr, uaecptr oldpc)
{
	uaecptr	curpc = m68k_getpc ();

	// Some opcode handlers (like op_4e40 for the TRAP #xx functions)
	// call Exception with oldpc == 0. We really need the PC so that
	// we can look up dispatch numbers, so provide that here.

	if (oldpc == 0)
	{
		if (kException_Trap0 >= 32 && nr <= kException_Trap15)
		{
			// pc was bumped by 2 already in op_4e40 handler.

			oldpc = curpc - 2;
		}
		else
		{
			assert (false);
		}
	}

	Software::ProcessException (nr, oldpc, curpc);
}


unsigned long op_illg (uae_u32 iOpcode)
{
	Software::ProcessIllegalInstruction (iOpcode);
	return 0;
}


void Software_ProcessLINK (uae_s32 linkSize)
{
	Software::ProcessLINK (linkSize);
}


int Software_ProcessRTS (void)
{
	return Software::ProcessRTS ();
}


int Software_ProcessRTE (uae_s32 newpc)
{
	return Software::ProcessRTE (newpc);
}


int Software_ProcessJSR (uaecptr oldpc, uaecptr dest)
{
	return Software::ProcessJSR (oldpc, dest);
}


int Software_ProcessJSR_Ind (uaecptr oldpc)
{
	return Software::ProcessJSR_Ind (oldpc);
}


void Software_CheckStackPointer (int opcodeType, uaecptr oldA7)
{
	Software::CheckStackPointer (opcodeType, oldA7);
}


void MakeSR (void)
{
	Registers::UpdateSRFromRegisters ();
}


void MakeFromSR (void)
{
	Registers::UpdateRegistersFromSR ();
}


void* xmalloc (size_t size)
{
	return Platform::AllocateMemory (size);
}


// Dummy functions just so that we can link under VC++.

void m68k_move2c (int, uae_u32 *)
{
}


void m68k_movec2 (int, uae_u32 *)
{
}


void m68k_divl (uae_u32, uae_u32, uae_u16, uaecptr)
{
}


void m68k_mull (uae_u32, uae_u32, uae_u16)
{
}


void fpp_opp (uae_u32, uae_u16)
{
}


void fdbcc_opp (uae_u32, uae_u16)
{
}


void fscc_opp (uae_u32, uae_u16)
{
}


void ftrapcc_opp (uae_u32,uaecptr)
{
}


void fbcc_opp (uae_u32, uaecptr, uae_u32)
{
}


void fsave_opp (uae_u32)
{
}


void frestore_opp (uae_u32)
{
}

