/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#ifndef _DEBUGMGR_H_
#define _DEBUGMGR_H_

#include "CPU_REG.h"		// kException_Trace

// Types

#pragma mark Types

typedef DWord (*registerFun)(int num);
typedef Boolean (*compareFun)(DWord a, DWord b);

typedef struct BreakpointCondition
// For now, breakpoint conditions must be of the form <reg> <cond> <value>, where
// <reg> is a 68000 register, <cond> is a binary comparison operator and <value> is
// a 32-bit integer.  All comparisons are unsigned!
{
	registerFun regType;
	int regNum;
	compareFun condition;
	DWord value;

	// The source text.  We keep this around so that, for example, a user can specify
	// a condition using either hex or decimal notation and they will see the same
	// notation the next time they edit the breakpoint.
	char *source;	
} BreakpointCondition;


struct DebugGlobalsType
{
	// Mode settings

	Bool			inDebugger;						// true if in debug mode
	Bool			ignoreDbgBreaks;				// if true, ignore DbgBreak's
	Bool			reEntered;						// true if we re-entered
	Bool			firstEntrance;					// true first time we enter debugger
	Bool			stepSpy;						// true if step spying.
	Bool			breakingOnATrap;				// true if there are A-Traps to check
	Bool			continuing;						// if true, restarting the CPU
													// after some sort of break, so
													// don't do A-Trap or BP stuff.

	// Breakpoints and saved opcodes behind each one

	BreakpointType	bp[dbgTotalBreakpoints];
	uae_u16			bpOpcode[dbgTotalBreakpoints];
	BreakpointCondition *bpCondition[dbgTotalBreakpoints];	// condition, or NULL if none

	// Current trap breaks

	Word			trapBreak[dbgTotalTrapBreaks];
	Word			trapParam[dbgTotalTrapBreaks];

	// Step spy support

	uaecptr			ssAddr;							// address to step spy on 
	DWord			ssValue;						// saved value

	// Exception type

	uae_s32			excType;						// why we entered debugger

	// (adam) Data breakpoint support.  This is similar to the step spy capability, but can monitor
	// an arbitrary range of addresses for writes, and doesn't require a saved value.

	bool			watchEnabled;
	uaecptr			watchAddr;		// address to watch, or 0 if none
	DWord			watchBytes;		// number of bytes to watch
};


// For reasons of "convenience", the official type declarations for these
// packets omits their bodies.  For the emulator, we'd really like to have
// the bodies, so we declare these parallel types.

struct SysPktReadRegsRspType2
{
	_sysPktBodyCommon;								// Common Body header
	M68KRegsType	reg;							// <- return registers
};

struct SysPktGetBreakpointsRspType2
{
	_sysPktBodyCommon;								// Common Body header
	BreakpointType	bp[dbgTotalBreakpoints];
};

struct SysPktSetBreakpointsCmdType2
{
	_sysPktBodyCommon;								// Common Body header
	BreakpointType	bp[dbgTotalBreakpoints];
};

struct SysPktGetTrapBreaksRspType2
{
	_sysPktBodyCommon;								// Common Body header
	Word		trapBP[dbgTotalTrapBreaks];
};

struct SysPktSetTrapBreaksCmdType2
{
	_sysPktBodyCommon;								// Common Body header
	Word		trapBP[dbgTotalTrapBreaks];
};

struct SysPktGetTrapConditionsRspType2
{
	_sysPktBodyCommon;								// Common Body header
	Word		trapParam[dbgTotalTrapBreaks];
};

struct SysPktSetTrapConditionsCmdType2
{
	_sysPktBodyCommon;								// Common Body header
	Word		trapParam[dbgTotalTrapBreaks];
};

// class Debug

#pragma mark class Debug

class Debug
{
	public:
		static void		Initialize			(void);
		static void		Reset				(void);
		static ErrCode	HandleNewPacket		(const SlkPktHeaderType& header,
											 const SysPktBodyType& pkt);
		static ErrCode	SendPacket			(const void* pkt, long size);
		static ErrCode	SendResponse		(int rsp);
		static ErrCode	SendMessage			(const char*);
		static Bool		BreakIfConnected	(uae_s32 exceptionNumber, uaecptr oldpc);
		static Bool		HandleTrap8			(uae_s32 exceptionNumber, uaecptr oldpc);
		static Bool		HandleSystemCall	(uae_u16 trapWord, uaecptr oldpc);
		static ErrCode	EnterDebugger		(uae_s32 reason, uaecptr oldpc);
		static ErrCode	ExitDebugger		(void);
		static Bool		InDebugger			(void);
		static void		CheckStepSpy		(uaecptr writeAddress, int writeBytes);

		static void		InstallBreakpoint	(uaecptr, int index = -1);
		static void		RemoveBreakpoint	(uaecptr);

		static void		InstallBreakpoints	(void);
		static void		RemoveBreakpoints	(void);

		static BreakpointCondition*
						NewBreakpointCondition	(const char* sourceString);
		static void		SetBreakpoint		(int index, uaecptr addr, BreakpointCondition* c);
		static void		ClearBreakpoint		(int index);

	private:
		static ErrCode	SendState			(void);
		static ErrCode	ReadMem				(SysPktReadMemCmdType&);
		static ErrCode	WriteMem			(SysPktWriteMemCmdType&);
		static ErrCode	SendRoutineName		(SysPktRtnNameCmdType&);
		static ErrCode	ReadRegs			(SysPktReadRegsCmdType&);
		static ErrCode	WriteRegs			(SysPktWriteRegsCmdType&);
		static ErrCode	Continue			(SysPktContinueCmdType&);
		static ErrCode	RPC					(SysPktRPCType&);
		static ErrCode	GetBreakpoints		(SysPktGetBreakpointsCmdType&);
		static ErrCode	SetBreakpoints		(SysPktSetBreakpointsCmdType2&);
		static ErrCode	ToggleBreak			(SysPktDbgBreakToggleCmdType&);
		static ErrCode	GetTrapBreaks		(SysPktGetTrapBreaksCmdType&);
		static ErrCode	SetTrapBreaks		(SysPktSetTrapBreaksCmdType2&);
		static ErrCode	Find				(SysPktFindCmdType&);
		static ErrCode	GetTrapConditions	(SysPktGetTrapConditionsCmdType&);
		static ErrCode	SetTrapConditions	(SysPktSetTrapConditionsCmdType2&);
		static void		GetRegs				(M68KRegsType&);
		static void		SetRegs				(const M68KRegsType&);
		static Bool		ConditionalBreak	(uae_s32 exceptionNumber, uaecptr oldpc);

	public:	// !!! Make this private again.
		static DebugGlobalsType		fgDebuggerGlobals;
	private:
		static SlkPktHeaderType*	fgLatestHeader;
		static long					fgLastTrapIndex;
		static char**				fgTrapNameTable;
};


// This function is called from the RAM setter functions. Make it inline so
// that the common case where stepSpy == FALSE executes quickly.  Yes, this
// _does_ make a difference.

inline void Debug::CheckStepSpy (uaecptr writeAddress, int writeBytes)
{
	if (fgDebuggerGlobals.stepSpy)
	{
		if (get_long (fgDebuggerGlobals.ssAddr) != fgDebuggerGlobals.ssValue)
		{
			// !!! Look into adam's comments below.
			Emulator::SetBreakReason (kException_Trace);
		}
	}
	if (fgDebuggerGlobals.watchEnabled
		&& writeAddress < fgDebuggerGlobals.watchAddr + fgDebuggerGlobals.watchBytes
		&& writeAddress + writeBytes > fgDebuggerGlobals.watchAddr)
			// (adam) The step spy code above apparently thinks it can break into
			// the debugger by calling Emulator::SetBreakReason, but that won't work
			// (I tried it).  Software::ProcessException() will do everything we need
			// here; it calls Debug::BreakIfConnected(), which in turn calls
			// Debug::EnterDebugger().
			// We must first check whether we are already in the debugger,
			// because it's possible that a single instruction will trigger the
			// data breakpoint several times.  This can happen, for example, when we hit
			// a trap #15 calling the PalmOS system function MemMove.  In this case, for
			// efficiency, the debugger emulates the entire MemMove call - this happens
			// in the function Headpatch::MemMove.  A single MemMove can trigger the
			// data breakpoint many times, but we must call Debug::EnterDebugger() only once
			// or very bad things will happen.
		if (!fgDebuggerGlobals.inDebugger)
			Software::ProcessException (kException_Trace, m68k_getpc (), m68k_getpc ());
}


#endif /* _DEBUGMGR_H_ */
  
  
