/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#include "EmulatorCommon.h"
#include "ErrorHandling.h"

#include "AppPreferences.h"		// gPrefs
#include "CPU_REG.h"			// SysTrapIndex
#include "Gremlins.h"			// Gremlins_IsOn
#include "Logging.h"			// LogWrite
#include "MetaMemory.h"			// MetaMemory
#include "Platform.h"			// GetString, CommonDialog
#include "Strings.r.h"			// kStr_ values
#include "TrapPatches.h"		// Patches::fgCurAppInfo


// ===========================================================================
//		 Errors
// ===========================================================================

ParamList		Errors::fgUserParameters;
unsigned long	Errors::fgWarningFlags;


// ---------------------------------------------------------------------------
//		 Errors::Initialize
// ---------------------------------------------------------------------------

void Errors::Initialize (void)
{
}


// ---------------------------------------------------------------------------
//		 Errors::Reset
// ---------------------------------------------------------------------------

void Errors::Reset (void)
{
	ClearWarningFlags ();
}


// ---------------------------------------------------------------------------
//		 Errors::Dispose
// ---------------------------------------------------------------------------

void Errors::Dispose (void)
{
}


// ---------------------------------------------------------------------------
//		 Errors::ReportIfError
// ---------------------------------------------------------------------------
// Checks for the indicated error condition.  If there is an error, it
// displays a "Could not foo because bar." message in a dialog with an OK
// button.  If an optional recovery string is provided, the message is
// "Could no foo because bar.  Do this." message.  If "throwAfter" is true,
// Errors::Scram is called after the dialog is dismissed.
//
// "operation" and "recovery" are "kStr_Foo" values.
// "error" is Mac or Windows error number, or a kError_Foo value.

void Errors::ReportIfError (StrCode operation, ErrCode error, StrCode recovery, Bool throwAfter)
{
	if (error)
	{
		Errors::SetStandardParameters ();

		if (Errors::SetErrorParameters (operation, error, recovery))
		{
			Errors::DoDialog (kStr_OpErrorRecover, kOK);
		}
		else
		{
			Errors::DoDialog (kStr_OpError, kOK);
		}

		if (throwAfter)
		{
			Errors::Scram ();
		}
	}
}


// ---------------------------------------------------------------------------
//		 Errors::ReportIfPalmError
// ---------------------------------------------------------------------------
// Checks for the indicated error condition.  If there is an error, it
// displays a "Could not foo because bar." message in a dialog with an OK
// button.  If an optional recovery string is provided, the message is
// "Could no foo because bar.  Do this." message.  If "throwAfter" is true,
// Errors::Scram is called after the dialog is dismissed.
//
// "operation" and "recovery" are "kStr_Foo" values.
// "err" is a Palm OS error number.

void Errors::ReportIfPalmError (StrCode operation, Err err, StrCode recovery, Bool throwAfter)
{
	if (err)
	{
		Errors::ReportIfError (operation, ::ConvertFromPalmError (err), recovery);
	}
}


// ---------------------------------------------------------------------------
//		 Errors::ReportIfNULL
// ---------------------------------------------------------------------------
// Checks for the indicated error condition.  If there is an error, it
// displays a "Could not foo because bar." message in a dialog with an OK
// button.  If an optional recovery string is provided, the message is
// "Could no foo because bar.  Do this." message.  If "throwAfter" is true,
// Errors::Scram is called after the dialog is dismissed.
//
// "operation" and "recovery" are "kStr_Foo" values.
// "p" is a pointer to be tested.

void Errors::ReportIfNULL (StrCode operation, void* p, StrCode recovery, Bool throwAfter)
{
	if (p == NULL)
	{
		Errors::ReportIfError (operation, kError_OutOfMemory, recovery);
	}
}


// ---------------------------------------------------------------------------
//		 Errors::ReportPreventedAccess
// ---------------------------------------------------------------------------
// Some process just accessed some memory that it shouldn't have. "kind"
// holds a value indicating what part of memory was accessed (low memory,
// system globals, screen buffer, or hardware registers), and "forRead"
// tells if whether it was a read or write operation.
//
// !!! Eventually, we'll probably want to include the memory address so that
// we can report it.

int Errors::ReportPreventedAccess (uaecptr address, long size, Bool forRead,
							  EAccessType kind)
{
	// See if we've already issued this warning for this application.
	// If so, just leave.

	if (AlreadyWarned (kind))
	{
		return kContinue;
	}

	// Generate the warning message.

	// Get the resource ID for the right error message template.

	int	strIndex;

	switch (kind)
	{
		case kRegisterAccess:
			if (!gPrefs->ReportHardwareRegisterAccess ())
			{
				return kContinue;
			}

			strIndex = kStr_RegisterAccess;
			break;

		case kLowMemAccess:
			if (!gPrefs->ReportLowMemoryAccess ())
			{
				return kContinue;
			}

			strIndex = kStr_LowMemAccess;
			break;

		case kGlobalVarAccess:
			if (!gPrefs->ReportSystemGlobalAccess ())
			{
				return kContinue;
			}

			strIndex = kStr_GlobalsAccess;
			break;

		case kScreenAccess:
			if (!gPrefs->ReportScreenAccess ())
			{
				return kContinue;
			}

			strIndex = kStr_ScreenAccess;
			break;

		case kMemMgrAccess:
			if (!gPrefs->ReportMemMgrDataAccess ())
			{
				return kContinue;
			}

			strIndex = kStr_MemMgrAccess;
			break;

		case kUnlockedChunkAccess:
			if (!gPrefs->ReportUnlockedChunkAccess ())
			{
				return kContinue;
			}

			strIndex = kStr_UnlockedChunkAccess;
			break;

		case kLowStackAccess:
			if (!gPrefs->ReportLowStackAccess ())
			{
				return kContinue;
			}

			strIndex = kStr_LowStackAccess;
			break;

		case kUninitializedStackAccess:
			if (!gPrefs->ReportUninitializedStackAccess ())
			{
				return kContinue;
			}

			strIndex = kStr_UninitializedStackAccess;
			break;

		case kFreeChunkAccess:
			if (!gPrefs->ReportFreeChunkAccess ())
			{
				return kContinue;
			}

			strIndex = kStr_FreeChunkAccess;
			break;

		case kUninitializedChunkAccess:
			if (!gPrefs->ReportUninitializedChunkAccess ())
			{
				return kContinue;
			}

			strIndex = kStr_UninitializedChunkAccess;
			break;

		case kStorageHeapAccess:
			if (!gPrefs->ReportStorageHeapAccess ())
			{
				return kContinue;
			}

			strIndex = kStr_StorageHeapAccess;
			break;

		default: assert (false); break;
	}

	// Get the application name and the operation text that need to
	// be inserted into the template.

	Errors::SetStandardParameters ();
	Errors::SetParameter ("%operation", Platform::GetString (forRead ? kStr_ReadFrom : kStr_WrittenTo));

	// Display the message with Continue and Debug buttons.

	int	button = Errors::DoDialog (strIndex, Errors::kContinueDebugReset);

	return button;
}


// ---------------------------------------------------------------------------
//		 Errors::ReportUnimplementedTrap
// ---------------------------------------------------------------------------

int Errors::ReportUnimplementedTrap (uae_u16 trapWord, int resourceBase)
{
	if (!gPrefs->ReportUnimplementedTrap ())
	{
		return kContinue;
	}

	Errors::SetStandardParameters ();

	char	trapNumber[20];
	sprintf (trapNumber, "%04X", (long) trapWord);

	string	trapName (Platform::GetString (resourceBase + (trapWord & 0x7FF)));
	if (trapName[0] == '<')	// Start of "<Missing string...>"
	{
		trapName = Platform::GetString (kStr_UnknownTrapNumber);
	}

	Errors::SetParameter ("%trapNum", trapNumber);
	Errors::SetParameter ("%trapName", trapName);

	int	button = Errors::DoDialog (kStr_UnimplementedTrap, kDebugReset);

	return button;
}


// ---------------------------------------------------------------------------
//		 Errors::ReportUnhandledException
// ---------------------------------------------------------------------------

int Errors::ReportUnhandledException (uae_s32 exceptionNumber)
{
	if (!gPrefs->ReportUnhandledException ())
	{
		return kContinue;
	}

	Errors::SetStandardParameters ();
	Errors::SetParameter ("%operation", Platform::GetString (kStr_ExceptionBase + exceptionNumber));

	int	button = Errors::DoDialog (kStr_UnhandledException, kDebugReset);

	return button;
}


// ---------------------------------------------------------------------------
//		 Errors::ReportCorruptedHeap
// ---------------------------------------------------------------------------

int Errors::ReportCorruptedHeap (int corruptionType)
{
	if (!gPrefs->ReportCorruptedHeap ())
	{
		return kContinue;
	}

	Errors::SetStandardParameters ();

	int	strID;

	switch ((PalmHeap::EPalmHeapResult) corruptionType)
	{
		case PalmHeap::kCorruptedHeap_ChunkNotInHeap:
			strID = kStr_ChunkNotInHeap;
			break;

		case PalmHeap::kCorruptedHeap_ChunkTooLarge:
			strID = kStr_ChunkTooLarge;
			break;

		case PalmHeap::kCorruptedHeap_InvalidFlags:
			strID = kStr_InvalidFlags;
			break;

		case PalmHeap::kCorruptedHeap_HOffsetNotInMPT:
			strID = kStr_HOffsetNotInMPT;
			break;

		case PalmHeap::kCorruptedHeap_HOffsetNotBackPointing:
			strID = kStr_HOffsetNotBackPointing;
			break;

		case PalmHeap::kCorruptedHeap_InvalidLockCount:
			strID = kStr_InvalidLockCount;
			break;

		default:
			assert (false);
	}

	Errors::SetParameter ("%corruption_type", Platform::GetString (strID));

	int	button = Errors::DoDialog (kStr_CorruptedHeap, kDebugReset);

	return button;
}


// ---------------------------------------------------------------------------
//		 Errors::ReportStackOverflow
// ---------------------------------------------------------------------------

int Errors::ReportStackOverflow (Bool fatal)
{
	if (fatal)
	{
		if (!gPrefs->ReportStackOverflow ())
		{
			return kContinue;
		}
	}
	else
	{
		if (!gPrefs->ReportStackAlmostOverflow ())
		{
			return kContinue;
		}
	}

	// If this is just a warning, warn only once.

	if (!fatal && AlreadyWarned (kStackOverflowWarning))
	{
		return kContinue;
	}

	Errors::SetStandardParameters ();

	int	strID = fatal ? kStr_StackOverflowFatal : kStr_StackOverflowWarning;
	int	dlgID = fatal ? kDebugReset : kContinueDebugReset;

	int	button = Errors::DoDialog (strID, dlgID);

	return button;
}


// ---------------------------------------------------------------------------
//		 Errors::ReportLongMemSemaphore
// ---------------------------------------------------------------------------

int Errors::ReportLongMemSemaphore (unsigned long elapsedMSecs)
{
	if (!gPrefs->ReportMemMgrSemaphore ())
	{
		return kContinue;
	}

	if (AlreadyWarned (kLongMemSemaphore))
	{
		return kContinue;
	}

	Errors::SetStandardParameters ();

	char	elapsed[20];
	sprintf (elapsed, "%ld", elapsedMSecs);
	Errors::SetParameter ("%elapsed", elapsed);

	int	button = Errors::DoDialog (kStr_LongMemSemaphore, kContinueDebugReset);

	return button;
}


// ---------------------------------------------------------------------------
//		 Errors::ReportSizelessFormObject
// ---------------------------------------------------------------------------

int Errors::ReportSizelessFormObject (Word id, const RectangleType& r)
{
//	if (!gPrefs->ReportMemMgrSemaphore ())
//	{
//		return kContinue;
//	}

	if (AlreadyWarned (kSizelessFormObject))
	{
		return kContinue;
	}

	Errors::SetStandardParameters ();

	char	temp[20];

	sprintf (temp, "%ld", id);
	Errors::SetParameter ("%objID", temp);

	sprintf (temp, "%ld", r.topLeft.x);
	Errors::SetParameter ("%left", temp);

	sprintf (temp, "%ld", r.topLeft.y);
	Errors::SetParameter ("%top", temp);

	sprintf (temp, "%ld", r.topLeft.x + r.extent.x);
	Errors::SetParameter ("%right", temp);

	sprintf (temp, "%ld", r.topLeft.y + r.extent.y);
	Errors::SetParameter ("%bottom", temp);

	int	button = Errors::DoDialog (kStr_SizelessFormObject, kContinue);

	return button;
}


// ---------------------------------------------------------------------------
//		 Errors::ReportLongMemSemaphore
// ---------------------------------------------------------------------------

int Errors::ReportOffscreenFormObject (Word id, const RectangleType& r)
{
//	if (!gPrefs->ReportMemMgrSemaphore ())
//	{
//		return kContinue;
//	}

	if (AlreadyWarned (kOffscreenFormObject))
	{
		return kContinue;
	}

	Errors::SetStandardParameters ();

	char	temp[20];

	sprintf (temp, "%ld", id);
	Errors::SetParameter ("%objID", temp);

	sprintf (temp, "%ld", r.topLeft.x);
	Errors::SetParameter ("%left", temp);

	sprintf (temp, "%ld", r.topLeft.y);
	Errors::SetParameter ("%top", temp);

	sprintf (temp, "%ld", r.topLeft.x + r.extent.x);
	Errors::SetParameter ("%right", temp);

	sprintf (temp, "%ld", r.topLeft.y + r.extent.y);
	Errors::SetParameter ("%bottom", temp);

	int	button = Errors::DoDialog (kStr_OffscreenFormObject, kContinue);

	return button;
}


// ---------------------------------------------------------------------------
//		 Errors:SysFatalAlert
// ---------------------------------------------------------------------------

int Errors::SysFatalAlert (const char* appMsg)
{
	if (!gPrefs->ReportSysFatalAlert ())
	{
		return kContinue;
	}

	Errors::SetStandardParameters ();

	// Format the entire message.  We use different messages depending on
	// whether or not Gremlins is running.

#if 0
	if (Gremlins_IsOn ())
	{
		msg	= Platform::GetString (kStringErrDisplayWithGremlins);

		assert (msg != NULL);

		msg = Errors::ReplaceParameter (msg, "%message", appMsg);
		msg = Errors::ReplaceParameter (msg, "#gremlin_number", Gremlins_GremlinNumber ());
		msg = Errors::ReplaceParameter (msg, "#gremlin_event", Gremlins_EventNumber ());
	}
	else
#endif
	{
		Errors::SetParameter ("%message", appMsg);
	}

	int	button = Errors::DoDialog (kStr_ErrDisplay, Errors::kContinueDebugReset);

	return button;
}


// ---------------------------------------------------------------------------
//		 Errors::SetParameter
// ---------------------------------------------------------------------------
//	Add a user-supplied substitution rule. Parameters are deleted after they
//	are used (by making a call to ReportMessage), so they need to be
//	re-established after every call to that function.

void Errors::SetParameter (const string& key, const string& value)
{
	fgUserParameters.push_back (key);
	fgUserParameters.push_back (value);
}


// ---------------------------------------------------------------------------
//		 Errors::SetParameter
// ---------------------------------------------------------------------------
//	Add a user-supplied substitution rule. Parameters are deleted after they
//	are used (by making a call to ReportMessage), so they need to be
//	re-established after every call to that function.
//
//	This version of the function is a convenience for use on the Mac side,
//	where Pascal strings are used a lot.  (Actually, a string constructor
//	that took Pascal strings would be nice...).

void Errors::SetParameter (const string& key, const unsigned char* value)
{
	fgUserParameters.push_back (key);
	fgUserParameters.push_back (string ((char*) &value[1], value[0]));
}


// ---------------------------------------------------------------------------
//		 Errors::SetErrorParameters
// ---------------------------------------------------------------------------

Bool Errors::SetErrorParameters (StrCode operation, ErrCode error, StrCode recovery)
{
	// Add %operation, %reason, and %recovery entries to our parameter
	// list in such a way that they appear at the start of the list
	// in that order. That way, if any of the replacement strings
	// contain parameters to be replaced, they can be replaced by
	// user-defined values.

	string s;

	// Add the error number as a possible substitution value.

	ErrCode	errCode = error;
	if (::IsPalmError (errCode))
	{
		errCode = ::ConvertToPalmError (errCode);
	}

	char	errCodeString[20];
	sprintf (errCodeString, "%ld", errCode);
	Errors::SetParameter ("%error", errCodeString);

	// If the caller didn't provide a recovery string ID, let's try to get
	// appropriate for this error code.

	if (recovery == 0)
	{
		recovery = Errors::GetIDForRecovery (error);
	}

	// If we (now) have a recovery string ID, use it to get the string
	// and add it to our parameter list.

	if (recovery)
	{
		s = Platform::GetString (recovery);
		fgUserParameters.push_front (s);
		fgUserParameters.push_front ("%recovery");
	}

	// Get a string for the error code provided.  If we don't have a canned
	// string, create a generic string that includes the error number.

	StrCode	errID = Errors::GetIDForError (error);
	s = Platform::GetString (errID);
	fgUserParameters.push_front (s);
	fgUserParameters.push_front ("%reason");

	// Finally, add the operation string to the parameter list.

	s = Platform::GetString (operation);
	fgUserParameters.push_front (s);
	fgUserParameters.push_front ("%operation");

	return recovery != 0;	// Return whether or not there's a recovery string in
							// the parameter list.  The caller will want to know
							// this so that it can provide the right template for it.
}


// ---------------------------------------------------------------------------
//		 Errors::SetStandardParameters
// ---------------------------------------------------------------------------
//	Set the following three parameters for text substitution:
//
//		%Application = current application name, or "The current application"
//		%application = current application name, or "the current application"
//		%version = current application version, or "(unknown version)"

void Errors::SetStandardParameters (void)
{
	string	appNameUC;
	string	appNameLC;
	string	version;

	Errors::GetAppName (appNameUC, appNameLC);
	Errors::GetAppVersion (version);

	Errors::SetParameter ("%Application", appNameUC);
	Errors::SetParameter ("%application", appNameLC);
	Errors::SetParameter ("%version", version);
}


// ---------------------------------------------------------------------------
//		 Errors::DoDialog
// ---------------------------------------------------------------------------
//	Displays a dialog box with the given message and according to
//	the given flags.  Returns which button was clicked.

int Errors::DoDialog (StrCode messageID, int flags)
{
	string	msg (Platform::GetString (messageID));
	return Errors::DoDialog (NULL, msg.c_str (), flags);
}


// ---------------------------------------------------------------------------
//		 Errors::DoDialog
// ---------------------------------------------------------------------------
//	Displays a dialog box with the given message and according to
//	the given flags.  Returns which button was clicked.

int Errors::DoDialog (const char* title, const char* msg, int flags)
{
	string	titleStr;

	if (title)
	{
		titleStr = title;
	}
	else
	{
		titleStr = Platform::GetString (kStr_AppName);
	}

	string	msgStr (msg);
	msgStr = Errors::ReplaceParameters (msgStr, fgUserParameters);

	if ((flags & kContinue) != 0)
	{
		if (LogWarningMessages ())
		{
			unsigned long	currentStep = Gremlins_EventCounter ();
			AppendMsg ("--- %lu:\n%s", currentStep - 1, msgStr.c_str ());
		}

		AppPreferences::GremlinInfo&	info = gPrefs->GetGremlinInfo ();
		if (info.fContinuePastWarnings)
		{
			return kContinue;
		}
	}

	DumpLog ();

	return Platform::CommonDialog (titleStr.c_str (), msgStr.c_str (), flags);
}


// ---------------------------------------------------------------------------
//		 Errors::ReplaceParameters
// ---------------------------------------------------------------------------
//	Take a string template (that is, a string containing text interspersed
//	with "parameters" (of the form "%parameterName") that need to be
//	replaced) and replace the parameters with their final values.

string Errors::ReplaceParameters (StrCode templateID)
{
	return Errors::ReplaceParameters (templateID, fgUserParameters);
}


// ---------------------------------------------------------------------------
//		 Errors::ReplaceParameters
// ---------------------------------------------------------------------------
//	Take a string template (that is, a string containing text interspersed
//	with "parameters" (of the form "%parameterName") that need to be
//	replaced) and replace the parameters with their final values.

string Errors::ReplaceParameters (StrCode templateID, ParamList& params)
{
	string	templ = Platform::GetString (templateID);
	return Errors::ReplaceParameters (templ, params);
}


// ---------------------------------------------------------------------------
//		 Errors::ReplaceParameters
// ---------------------------------------------------------------------------
//	Take a string template (that is, a string containing text interspersed
//	with "parameters" (of the form "%parameterName") that need to be
//	replaced) and replace the parameters with their final values.
//
//	Parameters are replaced one at a time, according to their position in
//	the ParamList.  The first parameter is first fully replace in the
//	string template, in such a manner than recursion does not occur (that
//	is, if a parameter value itself include a parameter with the same
//	name, that embedded parameter is not also replace).
//
//	The next parameter is then pulled off of ParamList and treate the
//	same way. This time, any new parameters introduced during the previous
//	substitution pass can be replaced.
//
//	This approach allows for sequences like the following:
//
//	string template "Could not %operation because %reason."
//
//	parameter list:		"%operation"	"save the file %extra"
//						"%reason"		"the disk is full"
//						"%extra"		"FooBlitzky"
//
//	result: "Could not save the file FooBlitzky" because the disk is full."
//
//	If the "%extra" in "save the file %extra" had been "%operation", it
//	would have stayed that way and never been replaced (unless "%operation"
//	appeared in the parameter list again).

string Errors::ReplaceParameters (const string& templ, ParamList& params)
{
	string	result (templ);

	ParamList::iterator	paramIter;

	for (paramIter = params.begin (); paramIter != params.end (); )
	{
		string	key		= *paramIter++;
		string	value	= *paramIter++;

		string::size_type	pos = 0;

		for (;;)
		{
			pos = result.find (key, pos);
			if (pos == string::npos)
				break;
			result.replace (pos, key.size (), value);
			pos += value.size ();
		}
	}

	fgUserParameters.clear ();

	return result;
}


// ---------------------------------------------------------------------------
//		 Errors::Throw
// ---------------------------------------------------------------------------

void Errors::Throw (ErrCode error)
{
	throw error;	// This type matches PowerPlant's ExceptionCode,
					// which is what's thrown by Throw_ and caught
					// by Catch_
}


// ---------------------------------------------------------------------------
//		 Errors::ThrowIfError
// ---------------------------------------------------------------------------

void Errors::ThrowIfError (ErrCode error)
{
	if (error)
	{
		Errors::Throw (error);
	}
}


// ---------------------------------------------------------------------------
//		 Errors::ThrowIfPalmError
// ---------------------------------------------------------------------------

void Errors::ThrowIfPalmError (Err error)
{
	if (error)
	{
		Errors::Throw (::ConvertFromPalmError (error));
	}
}


// ---------------------------------------------------------------------------
//		 Errors::ThrowIfNULL
// ---------------------------------------------------------------------------

void Errors::ThrowIfNULL (void* p)
{
	if (!p)
	{
		Errors::Throw (kError_OutOfMemory);
	}
}


// ---------------------------------------------------------------------------
//		 Errors::Scram
// ---------------------------------------------------------------------------

void Errors::Scram (void)
{
	Errors::Throw (kError_NoError);
}


// ---------------------------------------------------------------------------
//		 Errors::ClearWarningFlags
// ---------------------------------------------------------------------------

void Errors::ClearWarningFlags (void)
{
	fgWarningFlags = 0;
}


// ---------------------------------------------------------------------------
//		 Errors::GetIDForError
// ---------------------------------------------------------------------------

int Errors::GetIDForError (ErrCode error)
{
	if (::IsPalmError (error))
	{
		switch (::ConvertToPalmError (error))
		{
			case dmErrDatabaseOpen:			return kStr_DmErrDatabaseOpen;
			case memErrNotEnoughSpace:		return kStr_MemErrNotEnoughSpace;
			default:						return kStr_GenericPalmError;
		}
	}
	else if (::IsEmuError (error))
	{
		switch (error)
		{
			case kError_OutOfMemory:		return kStr_MemFull;
			case kError_BadROM:				return kStr_BadROM;
			case kError_WrongROMForType:	return kStr_WrongROMForType;
			case kError_UnsupportedROM:		return kStr_UnsupportedROM;
			case kError_CantDownloadROM_BadBaudRate:	return kStr_GenericError;
			case kError_CantDownloadROM_SerialPortBusy:	return kStr_GenericError;
			case kError_CantDownloadROM_Generic:		return kStr_GenericError;
			case kError_UnimplementedTrap:	return kStr_UnimplementedTrap;
			default:						return kStr_GenericError;
		}
	}

	return Platform::GetIDForError (error);
}


// ---------------------------------------------------------------------------
//		 Errors::GetIDForRecovery
// ---------------------------------------------------------------------------

int Errors::GetIDForRecovery (ErrCode error)
{
	if (::IsPalmError (error))
	{
		switch (::ConvertToPalmError (error))
		{
			default:
				return 0;
		}
	}
	else if (::IsEmuError (error))
	{
		switch (error)
		{
			default:
				return 0;
		}
	}

	return Platform::GetIDForRecovery (error);
}


// ---------------------------------------------------------------------------
//		 Errors::GetAppName
// ---------------------------------------------------------------------------
// Get the current application's name.  If the name is not known, return
// "Unknown application" and "unknown application".

void Errors::GetAppName (string& appNameUC, string& appNameLC)
{
	if (strlen (Patches::fgCurAppInfo.fName) > 0)
	{
		appNameUC = Patches::fgCurAppInfo.fName;
		appNameLC = Patches::fgCurAppInfo.fName;
	}
	else
	{
		appNameUC = Platform::GetString (kStr_CurrentAppUC);
		appNameLC = Platform::GetString (kStr_CurrentAppLC);
	}
}


// ---------------------------------------------------------------------------
//		 Errors::GetAppVersion
// ---------------------------------------------------------------------------
// Get the current application's version.  If the version cannot be
// determined, return "(unknown version)".

void Errors::GetAppVersion (string& appVersion)
{
	if (strlen (Patches::fgCurAppInfo.fVersion) > 0)
	{
		appVersion = Patches::fgCurAppInfo.fVersion;
	}
	else
	{
		appVersion = Platform::GetString (kStr_UnknownVersion);
	}
}


// ---------------------------------------------------------------------------
//		 Errors::AlreadyWarned
// ---------------------------------------------------------------------------

Bool Errors::AlreadyWarned (EAccessType kind)
{
	// Always warn if Gremlins is on.  The point of warning only once is
	// to reduce the annoyance to the person trying to interactively use
	// a slightly bent program.  But when running Gremlins, we aren't
	// annoying the user by showing dialog boxes, so always warn.

	if (Gremlins_IsOn ())
	{
		return false;
	}

	long	kindBit = 1 << ((long) kind);

	// See if we've already issued this warning for this application.
	// If so, just leave.

	if ((fgWarningFlags & kindBit) != 0)
	{
		return true;
	}

	// Remember that we've issued a warning for this application.
	// Remember only certain kinds of warnings.  All the other kinds
	// are considered serious enough to warn EVERY time.

	const unsigned long	kMask = (1 << (long) kRegisterAccess)		|
								(1 << (long) kLowMemAccess)			|
								(1 << (long) kGlobalVarAccess)		|
								(1 << (long) kScreenAccess)			|
								(1 << (long) kStorageHeapAccess)	|
								(1 << (long) kStackOverflowWarning)	|
								(1 << (long) kSizelessFormObject)	|
								(1 << (long) kOffscreenFormObject)	|
								0;

	fgWarningFlags |= kMask & kindBit;

	return false;
}
