/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#include "EmulatorCommon.h"
#include "Gremlins.h"

#include "CGremlins.h"
#include "CGremlinsStubs.h"

#include "AppPreferences.h"		// gPrefs
#include "Byteswapping.h"		// Canonical
#include "ErrorHandling.h"		// Errors::ReportMessage
#include "Logging.h"			// OpenLogFile, CloseLogFile
#include "Miscellaneous.h"		// ResetClibrationInfo
#include "Platform.h"			// Platform::GetMilliseconds
#include "Strings.r.h"			// kStr_ values
#include "UAE_Utils.h"			// uae_memcpy



// -----------------------------
   #pragma mark *** Globals
// -----------------------------

static Gremlins			gGremlins;
static Bool				gGremlinsOn;
static uae_u32			gLastTick;

static unsigned long	gGremlinStartTime;
static unsigned long	gGremlinStopTime;


// -----------------------------
   #pragma mark -
   #pragma mark *** Private Functions
// -----------------------------

static bool		AppCompareDataBaseNames (const AppInfo& a, const AppInfo& b);
static Err		AppGetExtraInfo (AppInfo* infoP);


// -----------------------------
   #pragma mark -
   #pragma mark *** Public Functions
// -----------------------------


// ---------------------------------------------------------------------------
//		 Gremlins_Initialize
// ---------------------------------------------------------------------------

void Gremlins_Initialize (void)
{
}


// ---------------------------------------------------------------------------
//		 Gremlins_Initialize
// ---------------------------------------------------------------------------

void Gremlins_Reset (void)
{
	Gremlins_Stop ();
}


// ---------------------------------------------------------------------------
//		 Gremlins_IsOn
// ---------------------------------------------------------------------------

Bool Gremlins_IsOn (void)
{
	return gGremlinsOn;
}


// ---------------------------------------------------------------------------
//		 Gremlins_CanSwitchToApp
// ---------------------------------------------------------------------------

Bool Gremlins_CanSwitchToApp (UInt cardNo, LocalID dbID)
{
	AppPreferences::GremlinInfo&	info = gPrefs->GetGremlinInfo ();

	if (info.fAppList.size () == 0)
	{
		return true;
	}

	AppInfoList::iterator	iter = info.fAppList.begin ();
	
	while (iter != info.fAppList.end ())
	{
		AppInfo	appInfo = *iter++;
		
		if (appInfo.cardNo == cardNo && appInfo.dbID == dbID)
		{
			return true;
		}
	}

	return false;
}


// ---------------------------------------------------------------------------
//		 Gremlins_PostFakeEvent
// ---------------------------------------------------------------------------

Bool Gremlins_PostFakeEvent (void)
{
	return gGremlins.GetFakeEvent() != 0;
}


// ---------------------------------------------------------------------------
//		 Gremlins_PostFakePenEvent
// ---------------------------------------------------------------------------

void Gremlins_PostFakePenEvent (void)
{
	// Only post every once and a while, so as to not overflow
	// the Palm OS's pen queue. (cf. StubEnqueuePenPosition)

	// (Note: taking out the pause.  It lead to indeterminism (the
	//  same Gremlin didn't do exactly the same thing every time),
	//  and doesn't seem necessary anyway, since we only post fake
	//  events when the event queue is empty anyway.)

//	uae_u32	now = Platform::GetMilliseconds() / 20;

//	if (now != gLastTick)
	{
		gGremlins.GetPenMovement ();

//		gLastTick = now;
	}
}


// ---------------------------------------------------------------------------
//		 Gremlins_GremlinNumber
// ---------------------------------------------------------------------------

uae_u32 Gremlins_GremlinNumber (void)
{
	unsigned short	number;
	unsigned long	step;
	unsigned long	until;
	gGremlins.Status (&number, &step, &until);

	return number;
}


// ---------------------------------------------------------------------------
//		 Gremlins_EventCounter
// ---------------------------------------------------------------------------

uae_u32 Gremlins_EventCounter (void)
{
	unsigned short	number;
	unsigned long	step;
	unsigned long	until;
	gGremlins.Status (&number, &step, &until);

	return step;
}


// ---------------------------------------------------------------------------
//		 Gremlins_EventLimit
// ---------------------------------------------------------------------------

uae_u32 Gremlins_EventLimit (void)
{
	unsigned short	number;
	unsigned long	step;
	unsigned long	until;
	gGremlins.Status (&number, &step, &until);

	return until;
}


// ---------------------------------------------------------------------------
//		 Gremlins_New
// ---------------------------------------------------------------------------

void Gremlins_New (void)
{
	AppPreferences::GremlinInfo&	info = gPrefs->GetGremlinInfo ();

	StartNewLog ();

	if (LogGremlins ())
	{
		string	templ = Platform::GetString (kStr_GremlinStarted);
		AppendMsg (templ.c_str (), (int) info.fNumber, info.fSteps);
	}

	// If needed, switch to an "approved" application.
	// This code roughly follows that in AppsViewSwitchApp in Launcher.

	if (info.fAppList.size () > 0)
	{
		// Switch to the first on the list.
		
		AppInfo	appInfo = *(info.fAppList.begin ());

		//---------------------------------------------------------------------
		// If this is an executable, call SysUIAppSwitch
		//---------------------------------------------------------------------
		if (IsExecutable (appInfo.type, appInfo.creator, appInfo.dbAttrs))
		{
			Err err = SysUIAppSwitch (appInfo.cardNo, appInfo.dbID,
							sysAppLaunchCmdNormalLaunch, NULL);
			Errors::ThrowIfPalmError (err);
		}

		//---------------------------------------------------------------------
		// else, this must be a launchable data database. Find it's owner app
		//  and launch it with a pointer to the data database name.
		//---------------------------------------------------------------------
		else
		{
			DmSearchStateType	searchState;
			UInt				cardNo;
			LocalID				dbID;
			Err err = DmGetNextDatabaseByTypeCreator (true, &searchState, 
						sysFileTApplication, appInfo.creator, 
						true, &cardNo, &dbID);
			Errors::ThrowIfPalmError (err);

			// Create the param block
			uaecptr	cmdPBP = (uaecptr) MemPtrNew (sizeof (SysAppLaunchCmdOpenDBType));
			Errors::ThrowIfNULL ((void*) cmdPBP);

			// Fill it in
			MemPtrSetOwner ((VoidPtr) cmdPBP, 0);
			put_word (cmdPBP + offsetof (SysAppLaunchCmdOpenDBType, cardNo), appInfo.cardNo);
			put_long (cmdPBP + offsetof (SysAppLaunchCmdOpenDBType, dbID), appInfo.dbID);

			// Switch now
			err = SysUIAppSwitch (cardNo, dbID, sysAppLaunchCmdOpenDB, (Ptr) cmdPBP);
			Errors::ThrowIfPalmError (err);
		}
	}

	gGremlins.Initialize (info.fNumber, info.fSteps);
	gGremlinsOn = true;

	// Make sure we're all on the same page...
	::ResetCalibrationInfo ();

	gPrefs->UseLoggingOptions (AppPreferences::kGremlinLogging);

	gGremlinStartTime = Platform::GetMilliseconds ();

	// Make sure the app's awake.  Normally, we post events on a patch to
	// SysEvGroupWait.  However, if the Palm device is already waiting,
	// then that trap will never get called.  By calling EvtWakeup now,
	// we'll wake up the Palm device from its nap.

	Errors::ThrowIfPalmError (EvtWakeup ());

	Platform::GCW_Open ();
	Platform::GCW_SetGremlinNumber (info.fNumber);
	Platform::GCW_SetEventNumber (0);
	Platform::GCW_SetMaxEventNumber (info.fSteps);
	Platform::GCW_SetElapsedTime (0);
	Platform::GCW_SetRunning ();
}


// ---------------------------------------------------------------------------
//		 Gremlins_Step
// ---------------------------------------------------------------------------

void Gremlins_Step (void)
{
	if (!gGremlinsOn)
	{
		gGremlinsOn = true;
		gGremlins.Step ();

		gPrefs->UseLoggingOptions (AppPreferences::kGremlinLogging);

		// Make sure the app's awake.  Normally, we post events on a patch to
		// SysEvGroupWait.  However, if the Palm device is already waiting,
		// then that trap will never get called.  By calling EvtWakeup now,
		// we'll wake up the Palm device from its nap.

		Errors::ThrowIfPalmError (EvtWakeup ());

		Platform::GCW_SetRunning ();
	}
}


// ---------------------------------------------------------------------------
//		 Gremlins_Resume
// ---------------------------------------------------------------------------

void Gremlins_Resume (void)
{
	if (!gGremlinsOn)
	{
		gGremlinsOn = true;
//		gGremlins.SetUntil (-1);		// in case a step action set the until value

		// Make sure we're all on the same page...
		::ResetCalibrationInfo ();

		gPrefs->UseLoggingOptions (AppPreferences::kGremlinLogging);

		// Make sure the app's awake.  Normally, we post events on a patch to
		// SysEvGroupWait.  However, if the Palm device is already waiting,
		// then that trap will never get called.  By calling EvtWakeup now,
		// we'll wake up the Palm device from its nap.

		Errors::ThrowIfPalmError (EvtWakeup ());

//		Platform::GCW_SetMaxEventNumber (-1);
		Platform::GCW_SetRunning ();
	}
}


// ---------------------------------------------------------------------------
//		 Gremlins_Stop
// ---------------------------------------------------------------------------

void Gremlins_Stop (void)
{
	if (gGremlinsOn)
	{
		gGremlinsOn = false;
		gGremlinStopTime = Platform::GetMilliseconds ();

		gPrefs->UseLoggingOptions (AppPreferences::kNormalLogging);

		Platform::GCW_SetElapsedTime (gGremlinStopTime - gGremlinStartTime);
		Platform::GCW_SetStopped ();

		unsigned short	number;
		unsigned long	step;
		unsigned long	until;
		gGremlins.Status (&number, &step, &until);

		if (LogGremlins ())
		{
			string	templ = Platform::GetString (kStr_GremlinEnded);
			AppendMsg (templ.c_str (),
				(int) number, step, until, (gGremlinStopTime - gGremlinStartTime));
		}

		DumpLog ();
	}
}


// ---------------------------------------------------------------------------
//		 Gremlins_CanNew
// ---------------------------------------------------------------------------

Bool Gremlins_CanNew (void)
{
	return true;
}


// ---------------------------------------------------------------------------
//		 Gremlins_CanStep
// ---------------------------------------------------------------------------

Bool Gremlins_CanStep (void)
{
	return gGremlins.IsInitialized () != 0;
}


// ---------------------------------------------------------------------------
//		 Gremlins_CanResume
// ---------------------------------------------------------------------------

Bool Gremlins_CanResume (void)
{
	return gGremlins.IsInitialized () && !gGremlinsOn;
}


// ---------------------------------------------------------------------------
//		 Gremlins_CanStop
// ---------------------------------------------------------------------------

Bool Gremlins_CanStop (void)
{
	return gGremlinsOn;
}


// -----------------------------
   #pragma mark -
   #pragma mark *** Stub Functions
// -----------------------------



// ---------------------------------------------------------------------------
//		 StubEnqueuePt
// ---------------------------------------------------------------------------
// This is a stub routine that called the application object's method that
// enqueues a point.

void	StubAppEnqueuePt(PointPtr pen)
{
	PointType	pt;
	pt.x = pen->h;
	pt.y = pen->v;
	::EvtEnqueuePenPoint(&pt);

	Platform::GCW_SetElapsedTime (Platform::GetMilliseconds () - gGremlinStartTime);
	Platform::GCW_SetEventNumber (Gremlins_EventCounter ());
}


// ---------------------------------------------------------------------------
//		 StubEnqueuePt
// ---------------------------------------------------------------------------
// This is a stub routine that called the Application object's method that
// enqueues a key.

void StubAppEnqueueKey (UInt ascii, UInt keycode, UInt modifiers)
{
	switch (ascii)
	{
		// This translates some control characters into system chars
		case 0x01:	// control-A
			ascii = menuChr;
			modifiers = commandKeyMask;
			break;
		case 0x03:	// control-C
			ascii = commandChr;
			modifiers = commandKeyMask;
			break;
		case 0x04:	// control-D
			ascii = confirmChr;
			modifiers = commandKeyMask;
			break;
		case 0x05:	// control-E
			ascii = launchChr;
			modifiers = commandKeyMask;
			break;
		case 0x06:	// control-F
			ascii = keyboardChr;
			modifiers = commandKeyMask;
			break;
		case 0x0D:	// control-D
			ascii = linefeedChr;
			break;
		case 0x0E:	// control-N
			ascii = nextFieldChr;
			modifiers = commandKeyMask;
			break;			
	}


	// If this is one of the hard keys, send it through the Key Manager so that
	//  it can do the right things concering double-taps, etc.
	if ((modifiers & commandKeyMask) &&
			(ascii > hardKeyMin || (ascii == pageUpChr) || (ascii == pageDownChr))) {

		DWord		status;

		switch (ascii) {
			case hard1Chr: 		status = keyBitHard1; break;
			case hard2Chr: 		status = keyBitHard2; break;
			case hard3Chr: 		status = keyBitHard3; break;
			case hard4Chr: 		status = keyBitHard4; break;
			case hardPowerChr: 	status = keyBitPower; break;
			case pageUpChr: 	status = keyBitPageUp; break;
			case pageDownChr: 	status = keyBitPageDown; break;
			}

		// I'm dubious that this is the correct thing to do in the emulator.
		// This is an interrupt routine that we're calling directly.  But
		// what if a "real" interrupt is triggered from some other source
		// while this call is executing (for instance, the user clicks on
		// one of the "keyboard buttons", thus generating a keyboard
		// interrupt)?  It's possible that the two will stomp on each other.
		// The best thing to do is to post events the emulator way.
		// However, I'm hard-pressed to do that.  That would require calling
		// Hardware_ButtonEvent twice (once to signal the keyboard
		// interrupt, and another to clear it), but I don't know how to give
		// the processor time in between to respond to the first call.

		::KeyHandleInterrupt (false, status);
		}

	// For other keys, enqueue them directly.
	else 
		::EvtEnqueueKey (ascii, keycode, modifiers);

	Platform::GCW_SetElapsedTime (Platform::GetMilliseconds () - gGremlinStartTime);
	Platform::GCW_SetEventNumber (Gremlins_EventCounter ());
}


// ---------------------------------------------------------------------------
//		 StubAppGremlinsOn
// ---------------------------------------------------------------------------
// Stub routine that update the Gremlins menu an the global variable the
// keeps track of the ????

void StubAppGremlinsOn (void)
{
	// Called from Gremlins::Initialize.
}


// ---------------------------------------------------------------------------
//		 StubAppGremlinsOff
// ---------------------------------------------------------------------------
// Stub routine that update the Gremlins menu an the global variable the
// keeps track of the ????

void StubAppGremlinsOff (void)
{
	// Called by Gremlins when counter > until.

	Gremlins_Stop ();
}


// ---------------------------------------------------------------------------
//		 StubViewDrawLine
// ---------------------------------------------------------------------------
// This is a stub routine that called the View object's method that draw
// a line.

void StubViewDrawLine (int xStart, int yStart, int xEnd, int yEnd)
{
	Platform::ViewDrawLine (xStart, yStart, xEnd, yEnd);
}


// ---------------------------------------------------------------------------
//		 StubViewDrawPixel
// ---------------------------------------------------------------------------
// This is a stub routine that called the View object's method that draw
// a pixel.

void StubViewDrawPixel (int xPos, int yPos)
{
	Platform::ViewDrawPixel (xPos, yPos);
}
