/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#ifndef _PLATFORM_H_
#define _PLATFORM_H_

#include "ErrorHandling.h"		// ErrCode
#include "UART.h"				// UART::Parity

#include "LoadApplication.h"	// LoadableList

// Public interface to our byte queue.  This queue is used to collect bytes
// going in an out of the serial ports, as well as characters to be passed
// from the keyboard to the emulated ROM.
//
// Most of the implementation of this class is provided by a platform-
// dependent, private class.  The reason for splitting the class up into
// two parts is because the multi-threaded Windows version of the emulator
// needs some synchronization when bytes are added and removed.

class TByteQueuePrivate;

class TByteQueue
{
	public:
								TByteQueue	();
								~TByteQueue	();

		void					Put			(uae_u8 b);
		uae_u8					Get			(void);
		int						GetUsed		(void);
		int						GetFree		(void);
		Bool					WaitForDataAvailable (int timeoutms);

	private:
		TByteQueuePrivate*		fImplementation;
};


// Globals

extern unsigned long	gPseudoTickcount;
extern Bool				gEnableSounds;		// accessed by CPU_setsounds

int _stricmp(const char *s1, const char *s2);
int _strnicmp(const char *s1, const char *s2, int n);
char* _strdup (const char* s);

// Function prototypes.

namespace Platform
{
	void		Initialize				(void);
	void		Reset					(void);

	// Dialog-related functions

	int			CommonDialog			(const char*, const char*, int flags);

	// Resource-related functions

	string		GetString				(int id);
	int			GetIDForError			(ErrCode error);
	int			GetIDForRecovery		(ErrCode error);
	
	string		GetShortVersionString	(void);

	// Time-related functions

	uae_u32		GetMilliseconds			(void);

	// External debugger-related functions

	ErrCode		ConnectToDebugger		(void);
	ErrCode		SendPacket				(const void* pkt, long size);
	void		DisconnectFromDebugger	(void);
	void		ExitDebugger			(void);

	// Graphics-related functions

	void		ViewDrawLine			(int xStart, int yStart, int xEnd, int yEnd);
	void		ViewDrawPixel			(int xPos, int yPos);

	// Serial-related functions

	ErrCode		OpenSerialPort			(void);
	ErrCode		SetSerialPortSettings	(UART::Parity	parity,
										 int			stopBits,
										 int			dataBits,
										 uae_u32		baud,
										 Bool			hwHandshaking);
	void		GetSerialPortSettings	(UART::Parity&	parity,
										 int&			stopBits,
										 int&			dataBits,
										 uae_u32&		baud,
										 Bool&			hwHandshaking);
	Bool		SerialPortOpen			(void);
	void		CloseSerialPort			(void);
	void		TransmitTxFIFO			(void);
	void		ReceiveRxFIFO			(void);

	// "Load File Progress Window"

	void		LFPW_Open				(void);
	void		LFPW_SetNumFiles		(int);
	void		LFPW_StartNewFile		(const char*);
	void		LFPW_SetFileMax			(int);
	void		LFPW_SetFileValue		(int);
	void		LFPW_Close				(void);

	// "Gremlin Control Window"

	void		GCW_Open				(void);
	void		GCW_SetGremlinNumber	(int);
	void		GCW_SetEventNumber		(int);
	void		GCW_SetMaxEventNumber	(int);
	void		GCW_SetElapsedTime		(uae_u32 elapsedMilliseconds);
	void		GCW_SetRunning			(void);
	void		GCW_SetStopped			(void);
	void		GCW_Close				(void);

	// Whatever....

	void		GetLoadableFileList		(string directoryName,
										 LoadableList& fileList);

	void		Delay					(void);
	Bool		TimeForCycle			(Bool sleeping);
	void		Cycle					(void);

	void*		AllocateMemory			(size_t size, Bool clear = false);
	void*		ReallocMemory			(void* p, size_t size);
	template <class T>
	void		DisposeMemory			(T*& p)
				{
					// Normally, I don't put functions inline like
					// this.  However, if I try defining this function
					// in a separate body, VC++ complains that DisposeMemory
					// doesn't take one parameter.  Go figure...

					if (p)
					{
						RealDisposeMemory ((void*) p);
						p = NULL;
					}
				}
	void		RealDisposeMemory		(void* p);

	Bool		StopOnResetKeyDown		(void);

	void		Debugger				(void);
}


namespace Platform
{
#if defined (__MACOS__)

	extern unsigned long	gLastCycleTime;

	inline Bool Platform::TimeForCycle (Bool sleeping)
	{
		// Calling LMGetTicks every time through the CPU loop is excessive.
		// On a 250MHz PowerMac, we can execute about 30K opcodes/tick, so
		// let's check in with LMGetTicks only that often.  Checking in every
		// 32768 opcodes, we get about a 9% performance increase (1850 msecs
		// vs. 1690 msecs in the tests I ran).

		static instructionCount;
		instructionCount++;
		if (!sleeping && (instructionCount & 0x7FFF) != 0)
			return false;

		return ((LMGetTicks () - gLastCycleTime) >= EVENT_THRESHHOLD);
	}

	inline void Platform::Debugger (void)
	{
		::Debugger ();
	}

#endif


#if defined (_WINDOWS) || defined (UNIX)
	inline Bool TimeForCycle (Bool)
	{
		static unsigned long	lastPseudoTickcount;

		if (lastPseudoTickcount != gPseudoTickcount)
		{
			lastPseudoTickcount = gPseudoTickcount;
			return true;
		}

		return false;
	}
#endif

#if defined (_WINDOWS)
	inline void Debugger (void)
	{
		__asm int 3;
	}

#endif

#if defined(UNIX)
	inline void Platform::Debugger(void)
	{
	}
	
#endif
}	// end namespace Platform


#endif /* _PLATFORM_H_ */
