/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#include "EmulatorCommon.h"
#include "Application.h"

#include "About.h"				// DoAbout
#include "Byteswapping.h"		// Canonical
#include "DebugMgr.h"			// Debug::HandlePacket
#include "CPU_MT.h"				// CPUStopper
#include "Document.h"			// QueryNewDocument, QueryOpenDocument, gCurrentRAMPath
#include "DownloadROM.h"		// DownloadROM
#include "Emulator.h"			// gInstance
#include "Gremlins.h"			// Gremlins_New, etc.
#include "LCD.h"				// CenterWindow, LCD_xxx functions
#include "LoadApplication.h"	// LoadPalmFile
#include "MMFMessaging.h"		// MMFOpenFiles
#include "Platform.h"			// Platform::LFPW_Open
#include "Profiling.h"			// ProfileInit, etc.
#include "Settings.h"			// gSettings
#include "SocketMessaging.h"
#include "Strings.r.h"			// kStr_ values
#include "TrapPatches.h"		// Patches::fgUIInitialized

#include "resource.h"
#include "AfxRes.h"				// Additional #defines (ID_NEW, etc.)

#include <CommCtrl.h>			// InitCommonControls, etc.
#include <windowsx.h>			// Button_SetCheck, etc.

#include <malloc.h>				// _alloca


// ======================================================================
//	Globals and constants
// ======================================================================

typedef void	(*MenuHandler) (HWND hwnd, WPARAM command);
typedef Bool	(*UpdateHandler) (HWND hwnd, WPARAM command);

WORD			gWinSockVersion;
QuitParams*		gQuitParams;

static DWORD	gHotSyncDownStart;


// ======================================================================
//	Private functions
// ======================================================================

static void DoAbout				(HWND hwnd, WPARAM command);
static void DoNew				(HWND hwnd, WPARAM command);
static void DoOpenOther			(HWND hwnd, WPARAM command);
static void DoOpenRecent		(HWND hwnd, WPARAM command);
static void DoClose				(HWND hwnd, WPARAM command);
static void DoSave				(HWND hwnd, WPARAM command);
static void DoSaveAs			(HWND hwnd, WPARAM command);
static void DoSaveScreen		(HWND hwnd, WPARAM command);
static void DoInstallOther		(HWND hwnd, WPARAM command);
static void DoInstallRecent		(HWND hwnd, WPARAM command);
static void DoHotSync			(HWND hwnd, WPARAM command);
static void DoReset				(HWND hwnd, WPARAM command);
static void DoTransferROM		(HWND hwnd, WPARAM command);
static void DoGremlinsNew		(HWND hwnd, WPARAM command);
	   void DoGremlinsStep		(HWND hwnd, WPARAM command);
	   void DoGremlinsResume	(HWND hwnd, WPARAM command);
	   void DoGremlinsStop		(HWND hwnd, WPARAM command);
static void DoProfileInit		(HWND hwnd, WPARAM command);
static void DoProfileStart		(HWND hwnd, WPARAM command);
static void DoProfileStop		(HWND hwnd, WPARAM command);
static void DoProfileDump		(HWND hwnd, WPARAM command);
static void DoProperties		(HWND hwnd, WPARAM command);
static void DoLoggingOptions	(HWND hwnd, WPARAM command);
static void DoDebugOptions		(HWND hwnd, WPARAM command);
static void DoBreakpoints		(HWND hwnd, WPARAM command);
static void DoExit				(HWND hwnd, WPARAM command);


/***********************************************************************
 *
 * FUNCTION:	PrvValidateNumber
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

static int PrvValidateNumber(	const char* text,
								long low,
								long high,
								const char* label,
								long* result)
{
	// !!! Localize calls to this function.

	const char *nptr = text;
	int c;

	while ( isspace((int)(unsigned char)*nptr) )
		++nptr;

	c = (int)(unsigned char) *nptr++;
	int	sign = c;
	if (c == '-' || c == '+')
	{
		c = (int)(unsigned char) *nptr++;
	}

	*result = 0;

	while (isdigit(c))
	{
		*result = 10 * *result + (c - '0');
		c = (int)(unsigned char) *nptr++;
	}

	if (sign == '-')
	{
		*result *= -1;
	}

	if (*result < low || *result > high)
	{
		char	buffer[200];
		// !!! Put in resource; use Errors::CommonDialog
		sprintf (buffer, "The value entered for %s must be in the range %ld to %ld.",
			label, low, high);
		::MessageBox (NULL, buffer, "Error", MB_OK);
		
		return false;
	}

	if (nptr <= text + strlen (text))
	{
		char	buffer[200];
		// !!! Put in resource; use Errors::CommonDialog
		sprintf (buffer, "A non-numeric character was entered for %s.", label);
		::MessageBox (NULL, buffer, "Error", MB_OK);

		return false;
	}

	return true;
}


/***********************************************************************
 *
 * FUNCTION:	NewGremlinProc
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

BOOL CALLBACK NewGremlinProc (	HWND hwnd,
								UINT msg,
								WPARAM wparam,
								LPARAM lparam)
{
	static AppInfoList				appList;
	AppPreferences::GremlinInfo&	info = gPrefs->GetGremlinInfo ();

	switch (msg)
	{
		default:
			return FALSE;	// The dialog box procedure should return nonzero if it processes the message, and zero if it does not.

		case WM_DESTROY:
			appList.clear ();
			break;

		case WM_COMMAND:
			switch (LOWORD (wparam))
			{
				default:
					return FALSE;	// The dialog box procedure should return nonzero if it processes the message, and zero if it does not.

				case IDOK:
				case IDCANCEL:
				{
					if (LOWORD (wparam) == IDOK)
					{
						char text[256];

						// Get the Gremlin number (as text) and convert it to a scaler.

						::GetDlgItemText (hwnd, IDC_GREMLIN_NUMBER, text, sizeof (text));
						if (!::PrvValidateNumber (text, 0, 999, "Gremlin Number", &info.fNumber))
							break;

						// Get the # of events (as text) and convert it to a scaler.

						::GetDlgItemText (hwnd, IDC_GREMLIN_STEPS, text, sizeof (text));
						if (!::PrvValidateNumber (text, -1, 0x7FFFFFFF, "Number of Steps", &info.fSteps))
							break;

						HWND	checkbox;
						int		checkState;

						// Get the checkbox setting for trap logging.

						checkbox = ::GetDlgItem (hwnd, IDC_CONTINUE_PAST_WARNINGS);
						info.fContinuePastWarnings = ::SendMessage (checkbox, BM_GETCHECK, 0, 0) == BST_CHECKED;

						// Get the number of selected items.

						HWND	list = GetDlgItem (hwnd, IDC_APPLICATION_LIST);
						LRESULT	numSelItems = ::SendMessage (list, LB_GETSELCOUNT, 0, 0);

						// Allocate a buffer for the result and get the list of selected items.

						LPINT	selItems = (LPINT) _alloca (numSelItems * sizeof (INT));
						::SendMessage (list, LB_GETSELITEMS, numSelItems, (LPARAM) selItems);

						// Fill gGremlinSelectedAppList based on the returned list of
						// selected items.

						info.fAppList.clear ();

						for (LRESULT ii = 0; ii < numSelItems; ++ii)
						{
							info.fAppList.push_back (appList [selItems[ii]]);
						}

						// Startup up the gremlin sub-system.

						::Gremlins_New ();
					}

					::EndDialog (hwnd, LOWORD (wparam));
					break;
				}

				case IDC_LOGGING_OPTIONS:
				{
					gHostPrefs->EditLoggingOptions (hwnd, 2);
					break;
				}
			}
			break;

		case WM_INITDIALOG:
		{
			::CenterWindow (hwnd);

			SetDlgItemInt (hwnd, IDC_GREMLIN_NUMBER, info.fNumber, true);
			SetDlgItemInt (hwnd, IDC_GREMLIN_STEPS, info.fSteps, true);

			HWND	checkbox;

			checkbox = ::GetDlgItem (hwnd, IDC_CONTINUE_PAST_WARNINGS);
			Button_SetCheck (checkbox, info.fContinuePastWarnings ? BST_CHECKED : BST_UNCHECKED);

			HWND	list = GetDlgItem (hwnd, IDC_APPLICATION_LIST);
			Bool	selected = false;

			// Get the list of installed applications.
			// (I tried doing this in WM_CREATE, but that's not passed to a DlgProc).

			::GetApplicationNames (appList);

			for (int ii = 0; ii < appList.size (); ++ii)
			{
				::SendMessage (list, LB_ADDSTRING, 0, (LPARAM) appList[ii].name);

				// If the item we just added is in our "selected" list,
				// then select it.
				
				AppInfoList::iterator	iter = info.fAppList.begin ();
				while (iter != info.fAppList.end ())
				{
					if (appList[ii] == *iter++)
					{
						ListBox_SetSel (list, true, ii);
						selected = true;
					}
				}
			}

			// If nothing's selected, then select something (the first item).

			if (!selected)
			{
				ListBox_SetSel (list, true, 0);
			}

			/*
				In response to a WM_INITDIALOG message, the dialog box procedure should
				return zero if it calls the SetFocus function to set the focus to one of
				the controls in the dialog box. Otherwise, it should return nonzero, in
				which case the system sets the focus to the first control in the dialog
				box that can be given the focus.
			*/

			break;
		}
	}

	return TRUE;	// The dialog box procedure should return nonzero if it processes the message, and zero if it does not.
}



// Parse an integer, which may be either decimal ("12345") or hex ("0x12345").
// Returns true if OK, false on parse error.
BOOL parseInt(char *s, DWord *i)
{
	if (sscanf(s, " %i", i) == 1)
		return TRUE;
	return FALSE;
}

static BOOL CALLBACK CodeBreakpointProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	switch (msg)
	{
		case WM_COMMAND:
			switch (LOWORD (wparam))
			{
				case IDOK:
				case IDCANCEL:
					if (LOWORD (wparam) == IDOK)
					{
						char text[256];
						DWord address;

						GetDlgItemText (hwnd, IDC_ADDRESS, text, sizeof (text));
						if (!parseInt(text, &address)) {
							MessageBox(0, "bad address", 0, MB_ICONEXCLAMATION);
							break;
						}

						GetDlgItemText(hwnd, IDC_CONDITION, text, sizeof text);
						BreakpointCondition *c = 0;
						if (*text) {
							c = Debug::NewBreakpointCondition(text);
							if (!c) {
								MessageBox(0, "can't parse condition", 0, MB_ICONEXCLAMATION);
								break;
							}
						}

						// install
						int i = GetWindowLong(hwnd, DWL_USER);
						Debug::SetBreakpoint(i, address, c);
					}
					EndDialog (hwnd, LOWORD (wparam));
					break;
			}
			break;

		case WM_INITDIALOG:
		{
			char text[256];
			int i = lparam;		// index of code breakpoint to edit
			CenterWindow (hwnd);
			if (Debug::fgDebuggerGlobals.bp[i].enabled) {
				sprintf(text, "0x%x", Debug::fgDebuggerGlobals.bp[i].addr);
				SetDlgItemText (hwnd, IDC_ADDRESS, text);
				BreakpointCondition *c = Debug::fgDebuggerGlobals.bpCondition[i];
				if (c)
					SetDlgItemText(hwnd, IDC_CONDITION, c->source);
			}
			SetWindowLong(hwnd, DWL_USER, i);
			break;
		}

		default:
			return FALSE;
	}

	return TRUE;
}

static void CodeBreakpointDialog(HWND hwnd, int i)
// Edit the code breakpoint at index i
{
	DialogBoxParam(gInstance, MAKEINTRESOURCE (IDD_CODEBREAKPOINT), hwnd,
		CodeBreakpointProc, i);
}

static void enableControl(HWND hwnd, int control, bool enable)
{
	HWND w = GetDlgItem(hwnd, control);
	EnableWindow(w, enable);
}

static void EnableCodeBreakpointControls(HWND hwnd, bool enable)
{
	enableControl(hwnd, IDC_EDIT, enable);
	enableControl(hwnd, IDC_CLEAR, enable);
}

static void EnableDataBreakpointControls(HWND hwnd, bool enable)
{
	enableControl(hwnd, IDC_STARTADDRESS, enable);
	enableControl(hwnd, IDC_ADDRESS, enable);
	enableControl(hwnd, IDC_NUMBEROFBYTES, enable);
	enableControl(hwnd, IDC_NBYTES, enable);
}

// return the index of the breakpoint which is currently selected, or LB_ERR if none
static int selectedBreakpoint(HWND hwnd)
{
	HWND list = GetDlgItem(hwnd, IDC_LIST);
	return ListBox_GetCurSel(list);
}

static void RefreshBreakpointList(HWND hwnd)
{
	HWND list = GetDlgItem (hwnd, IDC_LIST);
	SetWindowRedraw(list, false);
	int selected = ListBox_GetCurSel(list);
	ListBox_ResetContent(list);
	for (int i = 0 ; i < dbgTotalBreakpoints ; ++i)
	{
		char text[256];
		if (Debug::fgDebuggerGlobals.bp[i].enabled) {
			sprintf(text, "at 0x%x", Debug::fgDebuggerGlobals.bp[i].addr);
			BreakpointCondition *c = Debug::fgDebuggerGlobals.bpCondition[i];
			if (c) {
				strcat(text, " when ");
				strcat(text, c->source);
			}
		} else sprintf(text, "(#%d - not set)", i);
		ListBox_AddString(list, text);
	}
	if (selected != LB_ERR)
		ListBox_SetCurSel(list, selected);
	SetWindowRedraw(list, true);
	InvalidateRect(list, NULL, true);
}

static bool intToBool(int i)
// convert an int to a bool without generating a compiler warning
{
	return i ? true : false;
}

static BOOL CALLBACK BreakpointsProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	switch (msg)
	{
		case WM_COMMAND:
			switch (LOWORD (wparam))
			{
				case IDOK:
				// even though we don't have a Cancel button, the dialog has an X box which
				// sends an IDCANCEL notification
				case IDCANCEL:	
				{
					// set data breakpoint
					char text[256];
					bool watchEnabled;
					DWord watchAddr, watchBytes;

					HWND check = GetDlgItem(hwnd, IDC_ENABLED);
					if (watchEnabled = intToBool(Button_GetCheck(check))) {
					
						GetDlgItemText (hwnd, IDC_ADDRESS, text, sizeof (text));
						if (!parseInt(text, &watchAddr)) {
							MessageBox(0, "bad address", 0, MB_ICONEXCLAMATION);
							break;
						}

						GetDlgItemText (hwnd, IDC_NBYTES, text, sizeof (text));
						if (!parseInt(text, &watchBytes)) {
							MessageBox(0, "bad number of bytes", 0, MB_ICONEXCLAMATION);
							break;
						}

					}

					if (Debug::fgDebuggerGlobals.watchEnabled = watchEnabled) {
						Debug::fgDebuggerGlobals.watchAddr = watchAddr;
						Debug::fgDebuggerGlobals.watchBytes = watchBytes;
					}
					EndDialog (hwnd, LOWORD (wparam));
					break;
				}

				case IDC_LIST:
					{
					WORD code = HIWORD(wparam);
					if (code == LBN_SELCHANGE) { // a list item was selected
						int index = selectedBreakpoint(hwnd);
						EnableCodeBreakpointControls(hwnd, index != LB_ERR);
					}
					if (code != LBN_DBLCLK)
						break;
					}
					/* FALL THROUGH */
				case IDC_EDIT:
					{
					int index = selectedBreakpoint(hwnd);
					if (index != LB_ERR) {
						CodeBreakpointDialog(hwnd, index);
						RefreshBreakpointList(hwnd);
					}
					break;
					}

				case IDC_CLEAR:
					{
					int index = selectedBreakpoint(hwnd);
					if (index != LB_ERR) {
						Debug::ClearBreakpoint(index);
						RefreshBreakpointList(hwnd);
					}
					break;
					}

				case IDC_ENABLED:
					{
					HWND check = GetDlgItem(hwnd, IDC_ENABLED);
					bool c = intToBool(Button_GetCheck(check));
					EnableDataBreakpointControls(hwnd, c);
					break;
					}
			}
			break;

		case WM_INITDIALOG:
		{
			CenterWindow (hwnd);

			char text[256];

			// code breakpoints
			RefreshBreakpointList(hwnd);

			// Initially no breakpoint is selected in the list, so the "edit" and
			// "clear" buttons should be disabled.
			EnableCodeBreakpointControls(hwnd, false);

			// data breakpoint

			HWND check = GetDlgItem(hwnd, IDC_ENABLED);
			Button_SetCheck(check, Debug::fgDebuggerGlobals.watchEnabled);

			HWND address = GetDlgItem(hwnd, IDC_ADDRESS);
			sprintf(text, "0x%x", Debug::fgDebuggerGlobals.watchAddr);
			Edit_SetText(address, text);

			HWND bytes = GetDlgItem(hwnd, IDC_NBYTES);
			sprintf(text, "0x%x", Debug::fgDebuggerGlobals.watchBytes);
			Edit_SetText(bytes, text);

			EnableDataBreakpointControls(hwnd, Debug::fgDebuggerGlobals.watchEnabled);

			break;
		}

		default:
			return FALSE;
	}

	return TRUE;
}

void Breakpoints(HWND hwnd)
{
	DialogBox (gInstance, MAKEINTRESOURCE (IDD_BREAKPOINTS), hwnd, BreakpointsProc);
}


/***********************************************************************
 *
 * FUNCTION:	DoLoadPalmFiles
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoLoadPalmFiles (LoadableList& fileList)
{
	CPUStopper	stopper (::LCD_GetWindow (), "load an application");
	if (stopper.Stopped ())
	{
		::LoadPalmFileList (fileList);
	}
}



/***********************************************************************
 *
 * FUNCTION:	DoAbout
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoAbout (HWND hwnd, WPARAM command)
{
	::DoAbout (hwnd);
}


/***********************************************************************
 *
 * FUNCTION:	DoNew
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoNew (HWND hwnd, WPARAM command)
{
	gQuitParams->fCfg = gPrefs->GetLastConfiguration ();

	if (::QueryNewDocument (hwnd, gQuitParams->fCfg))
	{
		gQuitParams->fCommand = QuitParams::kUnspecified;
		::DoClose (hwnd, command);

		if (gQuitParams->fCommand == QuitParams::kNone)
		{
			gQuitParams->fCommand = QuitParams::kNew;
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	DoOpenOther
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoOpenOther (HWND hwnd, WPARAM command)
{
	if (::QueryOpenDocument (hwnd, gQuitParams->fRAMFileRef))
	{
		gQuitParams->fCommand = QuitParams::kUnspecified;
		::DoClose (hwnd, command);

		if (gQuitParams->fCommand == QuitParams::kNone)
		{
			gQuitParams->fCommand = QuitParams::kOpen;
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	DoOpenRecent
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoOpenRecent (HWND hwnd, WPARAM command)
{
	if (command >= ID_OPEN_RECENT &&
		command < ID_OPEN_RECENT + TSettings::MRU_COUNT)
	{
		gQuitParams->fCommand = QuitParams::kUnspecified;
		::DoClose (hwnd, command);

		if (gQuitParams->fCommand == QuitParams::kNone)
		{
			gQuitParams->fCommand = QuitParams::kOpen;
			gQuitParams->fRAMFileRef = gPrefs->GetRAMMRU (command - ID_OPEN_RECENT);
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	DoClose
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoClose (HWND hwnd, WPARAM command)
{
	int	action = gPrefs->GetCloseAction ();

	if (action == AppPreferences::kSaveAsk)
	{
		char	buffer[300];
		string	saveChanges (Platform::GetString (IDS_SAVECHANGES));
		string	untitled (Platform::GetString (kStr_Untitled));
		string	app (Platform::GetString (kStr_AppName));

		sprintf (buffer, saveChanges.c_str (), gCurrentRAMRef.IsSpecified ()
										? gCurrentRAMRef.GetFileName ().c_str ()
										: untitled.c_str ());
		int	result = ::MessageBox (hwnd, buffer, app.c_str (),
						MB_APPLMODAL | MB_ICONQUESTION | MB_YESNOCANCEL);

		if (result == IDYES)
			action = AppPreferences::kSaveAlways;

		else if (result == IDNO)
			action = AppPreferences::kSaveNever;

		else		// result == IDCANCEL
			return;
	}

	::CPU_stop ();

	if (action == AppPreferences::kSaveAlways)
	{
		::SaveRAMImage (hwnd, gCurrentRAMRef, false);
	}

	::DestroyWindow (hwnd);

	gQuitParams->fCommand = QuitParams::kNone;
}


/***********************************************************************
 *
 * FUNCTION:	DoSave
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoSave (HWND hwnd, WPARAM command)
{
	CPUStopper	stopper (::LCD_GetWindow (), "be saved");

	if (stopper.Stopped ())
	{
		::SaveRAMImage (hwnd, gCurrentRAMRef, /* askUser = */ false);
	}
}


/***********************************************************************
 *
 * FUNCTION:	DoSaveAs
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoSaveAs (HWND hwnd, WPARAM command)
{
	CPUStopper	stopper (::LCD_GetWindow (), "be saved");

	if (stopper.Stopped ())
	{
		::SaveRAMImage (hwnd, gCurrentRAMRef, /* askUser = */ true);
	}
}


/***********************************************************************
 *
 * FUNCTION:	DoSaveScreen
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoSaveScreen (HWND hwnd, WPARAM command)
{
	CPUStopper	stopper (::LCD_GetWindow (), "have its picture taken");

	if (stopper.Stopped ())
	{
		::LCD_ScreenShot ();
	}
}


/***********************************************************************
 *
 * FUNCTION:	DoInstallOther
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoInstallOther (HWND hwnd, WPARAM command)
{
	static const char	filters[] =
	{
		"Palm OS Files (*.prc, *.pdb, *.pqa)\0"	"*.prc;*.pdb;*.pqa\0"
		"Palm OS Applications (*.prc)\0"		"*.prc\0"
		"Palm OS Databases (*.pdb)\0"			"*.pdb\0"
		"Palm OS Clippings (*.pqa)\0"			"*.pqa\0"
		"All Files (*.*)\0"						"*.*\0"
		""
	};

	char fn[MAX_PATH];
	fn[0] = 0;

	OPENFILENAME	reply;
	memset (&reply, 0, sizeof(reply));

	reply.lStructSize		= sizeof (OPENFILENAME);
	reply.hInstance			= gInstance;
	reply.lpstrFilter		= filters;
	reply.nFilterIndex		= 1;
	reply.lpstrFile			= fn;
	reply.nMaxFile			= sizeof (fn);
	reply.lpstrInitialDir	= gHostPrefs->GetLastPRCDir ().GetFilePath ();
	reply.Flags				= OFN_ENABLEHOOK | OFN_EXPLORER | OFN_FILEMUSTEXIST | OFN_HIDEREADONLY | OFN_NOCHANGEDIR;
	reply.lpstrDefExt		= "prc";
	reply.lpfnHook			= &::CenterDlgProc;

	if (::GetOpenFileName (&reply))
	{
		LoadableList	fileList;
		fileList.push_back (LoadFilePB (fn));
		::DoLoadPalmFiles (fileList);

		fn[reply.nFileOffset] = 0;
		gHostPrefs->SetLastPRCDir (FileReference (fn));
	}
}


/***********************************************************************
 *
 * FUNCTION:	DoInstallRecent
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoInstallRecent (HWND hwnd, WPARAM command)
{
	if (command >= ID_INSTALL_RECENT &&
		command < ID_INSTALL_RECENT + TSettings::MRU_COUNT)
	{
		FileReference	theRef = gPrefs->GetPRCMRU (command - ID_INSTALL_RECENT);

		if (theRef.IsSpecified ())
		{
			LoadableList	fileList;
			fileList.push_back (LoadFilePB (theRef));
			::DoLoadPalmFiles (fileList);
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	DoHotSync
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoHotSync (HWND hwnd, WPARAM command)
{
	::CPU_hotsync (true);
	gHotSyncDownStart = ::GetTickCount ();
}


/***********************************************************************
 *
 * FUNCTION:	DoReset
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoReset (HWND hwnd, WPARAM command)
{
	::CPU_reset ();
}


/***********************************************************************
 *
 * FUNCTION:	DoTransferROM
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoTransferROM (HWND hwnd, WPARAM command)
{
	::DownloadROM (hwnd, "Palm.ROM");
}


/***********************************************************************
 *
 * FUNCTION:	DoGremlinsNew
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoGremlinsNew (HWND hwnd, WPARAM command)
{
	CPUStopper	stopper (::LCD_GetWindow (), "run a gremlin");

	if (stopper.Stopped ())
	{
		::DialogBox (	gInstance,
						MAKEINTRESOURCE (IDD_NEWGREMLIN),
						hwnd,
						&::NewGremlinProc);
	}
}


/***********************************************************************
 *
 * FUNCTION:	DoGremlinsStep
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoGremlinsStep (HWND hwnd, WPARAM command)
{
	CPUStopper	stopper (::LCD_GetWindow (), "single-step a gremlin");

	if (stopper.Stopped ())
		::Gremlins_Step ();
}


/***********************************************************************
 *
 * FUNCTION:	DoGremlinsResume
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoGremlinsResume (HWND hwnd, WPARAM command)
{
	CPUStopper	stopper (::LCD_GetWindow (), "resume a gremlin");

	if (stopper.Stopped ())
		::Gremlins_Resume ();
}


/***********************************************************************
 *
 * FUNCTION:	DoGremlinsStop
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoGremlinsStop (HWND hwnd, WPARAM command)
{
	CPUStopper	stopper (::LCD_GetWindow (), "resume a gremlin");

	if (stopper.Stopped ())
		::Gremlins_Stop ();
}


/***********************************************************************
 *
 * FUNCTION:	DoProfileInit
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoProfileInit (HWND hwnd, WPARAM command)
{
//	::ProfileInit (MAXFNCALLS, MAXDEPTH);
	::ProfileInit (0x0C000, MAXDEPTH);
}


/***********************************************************************
 *
 * FUNCTION:	DoProfileStart
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoProfileStart (HWND hwnd, WPARAM command)
{
	::ProfileStart ();
}


/***********************************************************************
 *
 * FUNCTION:	DoProfileStop
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoProfileStop (HWND hwnd, WPARAM command)
{
	::ProfileStop ();
}


/***********************************************************************
 *
 * FUNCTION:	DoProfileDump
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoProfileDump (HWND hwnd, WPARAM command)
{
	::ProfileDump (NULL);
	::ProfileCleanup ();
}


/***********************************************************************
 *
 * FUNCTION:	DoProperties
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoProperties (HWND hwnd, WPARAM command)
{
	Bool	needNewWindow = gHostPrefs->EditProperties (hwnd);

	if (needNewWindow)
	{
		::LCD_ResetWindow ();
	}
}


/***********************************************************************
 *
 * FUNCTION:	DoLoggingOptions
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoLoggingOptions (HWND hwnd, WPARAM command)
{
	gHostPrefs->EditLoggingOptions (hwnd, 1);
}


/***********************************************************************
 *
 * FUNCTION:	DoDebugOptions
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoDebugOptions (HWND hwnd, WPARAM command)
{
	gHostPrefs->EditDebugOptions (hwnd);
}


/***********************************************************************
 *
 * FUNCTION:	DoBreakpoints
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoBreakpoints (HWND hwnd, WPARAM command)
{
	CPUStopper	stopper (::LCD_GetWindow (), "change breakpoints");

	if (stopper.Stopped ())
		::DialogBox (gInstance, MAKEINTRESOURCE (IDD_BREAKPOINTS), hwnd, &::BreakpointsProc);
}


/***********************************************************************
 *
 * FUNCTION:	DoExit
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoExit (HWND hwnd, WPARAM command)
{
	::DoClose (hwnd, command);

	gQuitParams->fCommand = QuitParams::kQuit;
}


/***********************************************************************
 *
 * FUNCTION:	PopupMenu
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void PopupMenu (HWND hwnd, int x, int y)
{
	#define	MENU_OPEN			3
	#define MENU_INSTALL		10
	#define MENU_PROFILE		16

	HMENU	rootMenu = ::LoadMenu (gInstance, MAKEINTRESOURCE (IDR_MENU));
	assert (rootMenu);

	HMENU	menu = ::GetSubMenu (rootMenu, 0);
	assert (menu);

	HMENU	openMenu = ::GetSubMenu (menu, MENU_OPEN);
	assert (openMenu);

	HMENU	installMenu = ::GetSubMenu (menu, MENU_INSTALL);
	assert (installMenu);

	HMENU	profileMenu = ::GetSubMenu (menu, MENU_PROFILE);
	assert (profileMenu);

	string	emptyString (Platform::GetString (IDS_EMPTY));

	// Add recent .ram names.

	for (int ii = 0; ; ++ii)
	{
		FileReference	fileRef = gPrefs->GetRAMMRU (ii);
		if (!fileRef.IsSpecified ())
			break;

		char	buf[MAX_PATH];
		sprintf (buf, "&%d %s", ii + 1, fileRef.GetFilePath ());
		::AppendMenu (openMenu, MF_STRING, ID_OPEN_RECENT + ii, buf);
	}

	if (ii == 0)
	{
		::AppendMenu (openMenu, MF_STRING | MF_GRAYED, 0, emptyString.c_str ());
	}

	// Add recent .prc names.

	for (ii = 0; ; ++ii)
	{
		FileReference	fileRef = gPrefs->GetPRCMRU (ii);
		if (!fileRef.IsSpecified ())
			break;

		char	buf[MAX_PATH];
		sprintf (buf, "&%d %s", ii + 1, fileRef.GetFilePath ());
		::AppendMenu (installMenu, MF_STRING, ID_INSTALL_RECENT + ii, buf);
	}

	if (ii == 0)
	{
		::AppendMenu (installMenu, MF_STRING | MF_GRAYED, 0, emptyString.c_str ());
	}

#if !HAS_PROFILING
	// Enable/disable menu items.

	::EnableMenuItem (profileMenu, ID_PROFILE_INITIALIZE, MF_GRAYED);
	::EnableMenuItem (profileMenu, ID_PROFILE_START, MF_GRAYED);
	::EnableMenuItem (profileMenu, ID_PROFILE_STOP, MF_GRAYED);
	::EnableMenuItem (profileMenu, ID_PROFILE_DUMP, MF_GRAYED);
#endif

	// Show the menu.

	POINT	p = { x,  y };
	::ClientToScreen (hwnd, &p);

	::TrackPopupMenu (menu, TPM_LEFTALIGN | TPM_RIGHTBUTTON, p.x, p.y, 0, hwnd, NULL);

	::DestroyMenu (rootMenu);
}


/***********************************************************************
 *
 * FUNCTION:	HandlePacket
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

// Handle a debugger packet from an external debugger.  There are three
// scenarios under which we could receive a message:
//
//	(a) The CPU thread encountered an exception and sent off a state packet
//		to the debugger.  In that case, the CPU thread may or may not have
//		halted itself by now.  If it hasn't, we want to wait for it to stop
//		before handling the packet.  If it has, we can go ahead and handle
//		the packet.  In both cases, after the packet has been handled, we
//		need to leave the CPU thread halted.
//
//	(b) The packet may be part of a sequence of packets we're dealing with
//		while already being in debug mode.  In that case, the CPU thread
//		should already be halted, and we can process the packet.  If the
//		packet is anything other than a sysContinueCmd packet, we need to
//		leave the CPU thread halted.  Otherwise, we want to start it up
//		again.
//
//	(c) The packet could be sent to us from out of the blue.  For instance,
//		the user could be trying to set or clear a breakpoint.  In that case,
//		we want to halt the CPU thread, handle the packet, and restart the
//		thread.
//
// We can stop the CPU thread with either CPU_stop or CPU_wait.  We'd use
// CPU_wait only if we were expecting the thread to stop on its own, so
// we use CPU_stop, which will ask the thread to stop and then wait for it
// to do so. It's OK if the thread was already coming to a halt (as in
// scenario (a)).
//
// When deciding to restart the thread, we just look at the result of
// Debug::InDebugger().  In case (a), it should be true, so we leave the
// thread halted.  In case (b), it will be true unless we exited debug
// mode with a sysContinueCmd packet.  In case (c), it should always be
// false, as we never entered debug mode.

void HandlePacket (const void* data, long size, LPARAM userData)
{
	::CPU_stop ();

	SlkPktHeaderType&	header = *(SlkPktHeaderType*) data;
	SysPktBodyType&		body = *(SysPktBodyType*) ((&header) + 1);

	Debug::HandleNewPacket (header, body);

	Bool	inDebugger = Debug::InDebugger ();

	if (!inDebugger)
	{
		::CPU_run ();
	}
}


/***********************************************************************
 *
 * FUNCTION:	DisplayWndProc
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

LRESULT CALLBACK DisplayWndProc (HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	switch (msg)
	{
		case WM_CHAR:
			switch (wparam)
			{
				case VK_RETURN:	wparam = linefeedChr;		break;
			}

			::CPU_putkey (wparam);
			return 0;	// An application should return zero if it processes this message.

		case WM_COMMAND:
		{
			static struct
			{
				UINT			fCommand;
				int				fOperation;
				MenuHandler		fHandler;
				UpdateHandler	fUpdater;
			} commandHandlers[] =
			{
				ID_ABOUT,				kStr_CmdAbout,			DoAbout,			NULL,

				ID_NEW,					kStr_CmdNew,			DoNew,				NULL,
				ID_OPEN_OTHER,			kStr_CmdOpen,			DoOpenOther,		NULL,
				ID_OPEN_RECENT + 0,		kStr_CmdOpen,			DoOpenRecent,		NULL,
				ID_OPEN_RECENT + 1,		kStr_CmdOpen,			DoOpenRecent,		NULL,
				ID_OPEN_RECENT + 2,		kStr_CmdOpen,			DoOpenRecent,		NULL,
				ID_OPEN_RECENT + 3,		kStr_CmdOpen,			DoOpenRecent,		NULL,
				ID_OPEN_RECENT + 4,		kStr_CmdOpen,			DoOpenRecent,		NULL,
				ID_OPEN_RECENT + 5,		kStr_CmdOpen,			DoOpenRecent,		NULL,
				ID_OPEN_RECENT + 6,		kStr_CmdOpen,			DoOpenRecent,		NULL,
				ID_OPEN_RECENT + 7,		kStr_CmdOpen,			DoOpenRecent,		NULL,
				ID_OPEN_RECENT + 8,		kStr_CmdOpen,			DoOpenRecent,		NULL,
				ID_OPEN_RECENT + 9,		kStr_CmdOpen,			DoOpenRecent,		NULL,
				ID_CLOSE,				kStr_CmdClose,			DoClose,			NULL,

				ID_SAVE,				kStr_CmdSave,			DoSave,				NULL,
				ID_SAVEAS,				kStr_CmdSave,			DoSaveAs,			NULL,
				ID_SAVESCREEN,			kStr_CmdSaveScreen,		DoSaveScreen,		NULL,

				ID_INSTALL_OTHER,		kStr_CmdInstall,		DoInstallOther,		NULL,
				ID_INSTALL_RECENT + 0,	kStr_CmdInstall,		DoInstallRecent,	NULL,
				ID_INSTALL_RECENT + 1,	kStr_CmdInstall,		DoInstallRecent,	NULL,
				ID_INSTALL_RECENT + 2,	kStr_CmdInstall,		DoInstallRecent,	NULL,
				ID_INSTALL_RECENT + 3,	kStr_CmdInstall,		DoInstallRecent,	NULL,
				ID_INSTALL_RECENT + 4,	kStr_CmdInstall,		DoInstallRecent,	NULL,
				ID_INSTALL_RECENT + 5,	kStr_CmdInstall,		DoInstallRecent,	NULL,
				ID_INSTALL_RECENT + 6,	kStr_CmdInstall,		DoInstallRecent,	NULL,
				ID_INSTALL_RECENT + 7,	kStr_CmdInstall,		DoInstallRecent,	NULL,
				ID_INSTALL_RECENT + 8,	kStr_CmdInstall,		DoInstallRecent,	NULL,
				ID_INSTALL_RECENT + 9,	kStr_CmdInstall,		DoInstallRecent,	NULL,
				ID_HOTSYNC,				kStr_CmdHotSync,		DoHotSync,			NULL,
				ID_RESET,				kStr_CmdReset,			DoReset,			NULL,
				ID_TRANSFERROM,			kStr_CmdTransferROM,	DoTransferROM,		NULL,

				ID_GREMLINS_NEW,		kStr_CmdGremlinNew,		DoGremlinsNew,		NULL,
				ID_GREMLINS_STEP,		kStr_CmdGremlinStep,	DoGremlinsStep,		NULL,
				ID_GREMLINS_RESUME,		kStr_CmdGremlinResume,	DoGremlinsResume,	NULL,
				ID_GREMLINS_STOP,		kStr_CmdGremlinStop,	DoGremlinsStop,	NULL,

#if HAS_PROFILING
				ID_PROFILE_INITIALIZE,	kStr_CmdProfileInit,	DoProfileInit,		NULL,
				ID_PROFILE_START,		kStr_CmdProfileStart,	DoProfileStart,		NULL,
				ID_PROFILE_STOP,		kStr_CmdProfileStop,	DoProfileStop,		NULL,
				ID_PROFILE_DUMP,		kStr_CmdProfileDump,	DoProfileDump,		NULL,
#endif

				ID_PROPERTIES,			kStr_CmdPreferences,	DoProperties,		NULL,
				ID_LOGGINGOPTIONS,		kStr_CmdLoggingOptions,	DoLoggingOptions,	NULL,
				ID_DEBUGOPTIONS,		kStr_CmdDebugOptions,	DoDebugOptions,		NULL,
				ID_BREAKPOINTS,			kStr_CmdBreakpoints,	DoBreakpoints,		NULL,

				ID_EXIT,				kStr_CmdQuit,			DoExit,				NULL,

				NULL,					kStr_GenericOperation,	NULL,				NULL
			};

			int	index = 0;
			while (commandHandlers[index].fCommand)
			{
				if (commandHandlers[index].fCommand == wparam)
				{
					try
					{
						commandHandlers[index].fHandler (hwnd, wparam);
					}
					catch (ErrCode errCode)
					{
						Errors::ReportIfError (commandHandlers[index].fOperation, errCode, 0, false);
					}

					break;
				}

				index++;
			}

			return 0;	// If an application processes this message, it should return zero.
		}

		case WM_CREATE:
			::LCD_ResetWindow ();
			::DragAcceptFiles (hwnd, TRUE);
			return 0;	// If an application processes this message, it should return 0 to continue creation of the window.

		case WM_DESTROY:
			::LCD_SetWindow (NULL);
			::CPU_stop ();
			::DragAcceptFiles (hwnd, FALSE);

			{
				RECT	wr;
				::GetWindowRect (hwnd, &wr);

				Point	pt = { wr.top, wr.left };
				gPrefs->SetWindowLocation (pt);
			}

			Platform::GCW_Close ();

			::KillTimer (hwnd, 1);
			::PostQuitMessage (0);

			return 0;	// If an application processes this message, it should return zero.

		case WM_DROPFILES:
		{
			// A file is being dropped.

			HDROP	hDropInfo = (HDROP) wparam;

			// Get the number of files.

			int	numFiles = ::DragQueryFile (hDropInfo, (DWORD)(-1), NULL, 0);

			// This would get the drop point, if we needed it.

		//	POINT	pt;
		//	::DragQueryPoint (hDropInfo, &pt);

			int		prcCount = 0;
			int		psfCount = 0;
			int		romCount = 0;
			int		otherCount = 0;

			LoadableList	fileList;

			for (int ii = 0; ii < numFiles; ++ii)
			{
				char	filePath [MAX_PATH];
				::DragQueryFile (hDropInfo, ii, filePath, sizeof (filePath));

				fileList.push_back (LoadFilePB (FileReference (filePath)));

				char	fileName [_MAX_PATH], ext[_MAX_EXT];
				_splitpath (filePath, NULL, NULL, fileName, ext);
				strcat (fileName, ext);

				if (::EndsWith (fileName, ".prc"))	prcCount++;
				else if (::EndsWith (fileName, ".pdb"))	prcCount++;
				else if (::EndsWith (fileName, ".pqa"))	prcCount++;
				else if (::EndsWith (fileName, ".psf"))	psfCount++;
				else if (::EndsWith (fileName, ".rom"))	romCount++;
				else if (::StartsWith (fileName, "rom."))	romCount++;
				else otherCount++;
			}

			// Signal that the drag-and-drop operation is over.

			::DragFinish (hDropInfo);

			if ((prcCount > 0) + (psfCount > 0) + (romCount > 0) + (otherCount > 0) > 1)
			{
				string	app (Platform::GetString (kStr_AppName));
				string	msg (Platform::GetString (kStr_OnlySameType));
				::MessageBox (hwnd, msg.c_str (), app.c_str (), MB_OK);
			}
			else if (prcCount > 0)
			{
				::DoLoadPalmFiles (fileList);
			}
			else if (psfCount > 0)
			{
				if (psfCount > 1)
				{
					string	app (Platform::GetString (kStr_AppName));
					string	msg (Platform::GetString (kStr_OnlyOnePSF));
					::MessageBox (hwnd, msg.c_str (), app.c_str (), MB_OK);
				}
				else
				{
					gPrefs->UpdateRAMMRU ((*fileList.begin ()).appFile);
					::DoOpenRecent (hwnd, ID_INSTALL_RECENT);
				}
			}
			else if (romCount > 0)
			{
				if (romCount > 1)
				{
					string	app (Platform::GetString (kStr_AppName));
					string	msg (Platform::GetString (kStr_OnlyOneROM));
					::MessageBox (hwnd, msg.c_str (), app.c_str (), MB_OK);
				}
				else
				{
					AppPreferences::Configuration	cfg;
					cfg = gPrefs->GetLastConfiguration ();
					cfg.fROMFile = (*fileList.begin ()).appFile;
					gPrefs->SetLastConfiguration (cfg);
					::DoNew (hwnd, ID_NEW);
				}
			}
			else
			{
				string	app (Platform::GetString (kStr_AppName));
				string	msg (Platform::GetString (kStr_UnknownType));
				::MessageBox (hwnd, msg.c_str (), app.c_str (), MB_OK);
			}

			return 0;	// An application should return zero if it processes this message.
		}

		case WM_KEYDOWN:
			switch (wparam)
			{
				case VK_UP:		::CPU_putkey (upArrowChr);		return 0;
				case VK_DOWN:	::CPU_putkey (downArrowChr);		return 0;
				case VK_LEFT:	::CPU_putkey (leftArrowChr);		return 0;
				case VK_RIGHT:	::CPU_putkey (rightArrowChr);		return 0;
			}
			/***** FALL THROUGH *****/

		case WM_KEYUP:
		{
			int		whichButton = kButton_None;
			Bool	buttonDown = msg == WM_KEYDOWN;

			switch (wparam)
			{
				case VK_ESCAPE:	whichButton = kButton_Power;		break;
				case VK_PRIOR:	whichButton = kButton_UpArrow;		break;
				case VK_NEXT:	whichButton = kButton_DownArrow;	break;
				case VK_F1:		whichButton = kButton_DateBook;		break;
				case VK_F2:		whichButton = kButton_Phone;		break;
				case VK_F3:		whichButton = kButton_ToDo;			break;
				case VK_F4:		whichButton = kButton_Memo;			break;
			}

			if (whichButton != kButton_None)
			{
				::CPU_button (buttonDown, whichButton);
			}
			
			return 0;	// An application should return zero if it processes this message.
		}

		case WM_LBUTTONDOWN:
			::LCD_PenEvent (hwnd, lparam, true);
			return 0;	// If an application processes this message, it should return zero.

		case WM_LBUTTONUP:
			::LCD_PenEvent (hwnd, lparam, false);
			return 0;	// If an application processes this message, it should return zero.

		case WM_MOUSEMOVE:
			if (GetCapture ())
			{
				::LCD_PenEvent (hwnd, lparam, TRUE);
			}
			break;	// (No comment in the docs on what to return)

		case WM_PAINT:
		{
			::LCD_Paint (hwnd);
			return 0;	// An application returns zero if it processes this message.
		}

		case WM_PALETTECHANGED:
			if (HWND(wparam) == hwnd)
			{
				break;	// (No comment in the docs on what to return)
			}
			/***** FALL THROUGH *****/

		case WM_QUERYNEWPALETTE:
		{
			int n = ::LCD_SelectPalette (hwnd);

			if (n != -1)
				return TRUE;	// If the window realizes its logical palette, it must return TRUE; otherwise, it must return FALSE.

			return FALSE;
		}

		case WM_RBUTTONDOWN:

			// When the user clicks with the right mouse button, popup our menu.

			::PopupMenu (hwnd, LOWORD (lparam), HIWORD (lparam));
			return 0;	// If an application processes this message, it should return zero.

		case WM_SYSKEYDOWN:
			switch (wparam)
			{
				// When the user presses F10, popup our menu.

				case VK_F10:
					::PopupMenu (hwnd, 0, 0);
					return 0;	// An application should return zero if it processes this message.
			}
			break;

		case WM_TIMER:
		{
			// Perform periodic tasks.

			// Update our window with the current contents of the emulated LCD buffer.

			::LCD_Refresh (hwnd);

			// See if any external debugger is trying to connect to us.

			if (Socket::ConnectPending ())
			{
				Socket::AcceptConnection ();

				// If we're listening on a socket, install the 'gdbS' feature.  The
				// existance of this feature causes programs written the the prc tools
				// to enter the debugger  when they're launched.

				if (Patches::fgUIInitialized)
				{
					CPUStopper	stopper (LCD_GetWindow (), "set the gdb feature");
					if (stopper.Stopped ())
					{
						::FtrSet ('gdbS', 0, 0x12BEEF34);
					}
				}
			}

			// If we're connected to an external debugger and it has
			// sent us data, read that data and process it.

			if (Socket::IsConnected () && Socket::HasUnreadData (0))
			{
				::CPU_stop ();

				do {
					SlkPktHeaderType	header;
					SysPktBodyType		body;
					SlkPktFooterType	footer;

					if (Socket::Read (&header, sizeof (header)) == sizeof (header))
					{

						Word	bodySize = header.bodySize;
						Canonical (bodySize);

						Socket::Read (&body, bodySize);
						Socket::Read (&footer, sizeof (footer));

						Debug::HandleNewPacket (header, body);
					}
				} while (Socket::IsConnected () && Socket::HasUnreadData (50));

				Bool	inDebugger = Debug::InDebugger ();

				if (!inDebugger)
				{
					::CPU_run ();
				}
			}

			// If we're emulating the user pressing the hotsync button, emulate
			// releasing the button after 1 second.

			if (gHotSyncDownStart && ::GetTickCount () - gHotSyncDownStart > 1000)
			{
				::CPU_hotsync (false);
				gHotSyncDownStart = 0;
			}

			// Bump this guy so that periodic tasks can be handled in
			// Hardware::Cycle.

			gPseudoTickcount++;

			if (Patches::fgQuitStage == kTimeToQuit)
			{
				// Reset fgQuitStage so that if DoExit brings up a "Save this?"
				// dialog, we don't re-enter this code at the dialogs idle time.

				Patches::fgQuitStage = kNoQuit;
				::DoExit (hwnd, ID_EXIT);
			}

			return 0;	// An application should return zero if it processes this message.
		}
	}

	return ::DefWindowProc (hwnd, msg, wparam, lparam);
}

Bool	gKillTimerThread;

VOID CALLBACK TimerProc (
	HWND hwnd,		// handle of window for timer messages
	UINT uMsg,		// WM_TIMER message
	UINT idEvent,	// timer identifier
	DWORD dwTime)	// current system time)
{
	HWRegisters::SetCurrentMilliseconds (dwTime /* ::GetTickCount ()*/);

	if (gKillTimerThread)
	{
		::PostQuitMessage (0);
	}
}


DWORD CALLBACK TimerThread (void*)
{
	UINT	timerID = ::SetTimer (NULL, 0, 10, TimerProc);

	MSG	msg;
	while (::GetMessage (&msg, NULL, 0, 0))
	{
		::TranslateMessage (&msg);
		::DispatchMessage (&msg);
	}

	::KillTimer (NULL, timerID);

	return 0;
}

/***********************************************************************
 *
 * FUNCTION:	DoUIThread
 *
 * DESCRIPTION:	Thread proc for the UI thread.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

DWORD CALLBACK DoUIThread (void* quitParams)
{
	gQuitParams = (QuitParams*) quitParams;

	::InitCommonControls ();

	::LCD_Init ();

	// Open the connection to the external debugger.  We used to do
	// this from Platform::ConnectToDebugger, but that would get
	// called from a different thread.  Because the window used for
	// communications would belong to a different thread, the message
	// loop below would not process messages for it.  So we do it here.

	::MMFOpenFiles (NULL, HandlePacket, (LPARAM) LCD_GetWindow (),
					kFromEmulatorFileName, kToEmulatorFileName);

	/*
		Start up WinSock.

		FROM THE WINDOWS SOCKETS DOCUMENTATION:

			Upon entry to WSAStartup, the WS2_32.DLL examines the version requested
			by the application. If this version is equal to or higher than the lowest
			version supported by the DLL, the call succeeds and the DLL returns in
			wHighVersion the highest version it supports and in wVersion the minimum
			of its high version and wVersionRequested. The WS2_32.DLL then assumes
			that the application will use wVersion.

		WHAT THIS MEANS:

						requested_version
								|
				+---------------+---------------+
				| (a)			| (b)			| (c)
				V				V				V
				   +-------------------------+
				   | Supported version range |
				   +-------------------------+

			(a) We've requested a version older than that supported by the DLL. The docs
				don't say what happens here. I assume WSAStartup returns WSAVERNOTSUPPORTED.
				Outcome: we don't have a chance, so error out.

			(b) We've requested a version number in the range supported by the DLL.
				wHighVersion holds the right end of the range shown above.
				wVersion holds the requested version.
				Outcome: use the version in wVersion.

			(c) We've requested a version newer than that supported by the DLL.
				wHighVersion holds the right end of the range shown above.
				wVersion holds the right end of the range shown above.
				Outcome: use the version in wVersion
	*/

	WORD		versionRequested = MAKEWORD (2, 2);
	WSADATA		wsaData;

	gWinSockVersion = 0;
	int	result = ::WSAStartup (versionRequested, &wsaData);

	if (result == NO_ERROR)
	{
		gWinSockVersion = wsaData.wVersion;
	}
	else
	{
		::MessageBox (NULL, "Unable to open the Windows Sockets library.\n"
			"Networking services will be unavailable.", "Error", MB_OK);
	}

	// Don't make this check.  Windows '95 returns version 1.1 from WSAStartup.

//	if (LOBYTE (wsaData.wVersion) != 2 || HIBYTE (wsaData.wVersion) != 2)
//	{
//		gSocketLog.Printf ("wsaData.wVersion = %08X\n", (long) wsaData.wVersion);
//		PrvErrorOccurred ();
//		return;
//	}

	// Open a TCP socket for any TCP-based debuggers out there.

	if (gWinSockVersion > 0)
	{
		Socket::Start ();
	}

	// Kickstart the CPU emulator (running in its own thread).

	::CPU_run ();

	// Start up a little timer thread used to keep the emulated tickcount
	// in sync with the host's

	gKillTimerThread = false;
	DWORD	tid;
	HANDLE	timer = ::CreateThread (	NULL,			// Security attributes
										0,				// Stack size
										TimerThread,	// Start address
										NULL,			// Thread-func parameter
										0,				// Creation flags
										&tid);			// Thread ID


	// Our main event loop.  Retrieve and handle events until the application quits.

	MSG	msg;
	while (::GetMessage (&msg, NULL, 0, 0))
	{
		::TranslateMessage (&msg);
		::DispatchMessage (&msg);
	}

	// Kill the timer thread.

	gKillTimerThread = true;
	::WaitForSingleObject (timer, INFINITE);	// Wait for task to die.
	::CloseHandle (timer);

	Socket::Stop ();

	::WSACleanup ();

	::LCD_Quit ();

	// Exit the thread.

	return 0;
}


/***********************************************************************
 *
 * FUNCTION:	StartUIThread
 *
 * DESCRIPTION:	Create and return the handle to the UI thread.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

HANDLE StartUIThread (QuitParams* quitParams)
{
	DWORD	tid;
	HANDLE	thread = ::CreateThread (	NULL,			// Security attributes
										0,				// Stack size
										DoUIThread,		// Start address
										quitParams,		// Thread-func parameter
										0,				// Creation flags
										&tid);			// Thread ID
	return thread;
}
