/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#include "EmulatorCommon.h"
#include "Platform.h"

#include <COMMCTRL.H>			// PBM_SETRANGE, etc.

#include "Application.h"		// EndsWith
#include "AppPreferences.h"		// gPrefs
#include "Emulator.h"			// gInstance
#include "Gremlins.h"			// Gremlins_CanStop, etc.
#include "LCD.h"				// LCD_GetWindow, CenterWindow
#include "Logging.h"			// gLogSerial, AppendMsg
#include "MMFMessaging.h"		// MMFSendData
#include "TrapPatches.h"		// Patches::SetSwitchApp, Patches::QuitOnAppExit
#include "Strings.r.h"			// kStr_ values

#include "resource.h"

#include <locale.h>				// localconv, lconv
#include <malloc.h>				// _alloca


// ======================================================================
//	Types
// ======================================================================

// Queue used for keyboard and serial input/output.  This is the
// private, platform-dependent implementation used by TByteQueue.

// This class does not need a mutex because it assumes there is only
// one producer (who calls Put) and one consumer (who calls Get).

class TByteQueuePrivate
{
	public:
								TByteQueuePrivate	();
								~TByteQueuePrivate	();

		void					Put			(uae_u8 b);
		uae_u8					Get			(void);
		int						GetUsed		(void) { return fUsedCount; }
		int						GetFree		(void) { return fFreeCount; }
		Bool					WaitForDataAvailable (int timeoutms);

	private:

		enum { QUEUE_SIZE = 8 };

		long					fUsedCount;
		long					fFreeCount;
		HANDLE					fDataAvailable;
		uae_u8					fBuffer[QUEUE_SIZE];
		int						fHead;
		int						fTail;
};


// ======================================================================
//	Globals and constants
// ======================================================================

extern HANDLE		gCPUDelayEvent;		// In CPU_MT.cpp

Bool				gEnableSounds;
unsigned long		gPseudoTickcount;

HANDLE				gCommHandle;


// ======================================================================
//	Private functions
// ======================================================================

static void	PrvInsertString (char* dest, int insertBefore, const char* src);
static void	PrvFormatInteger (char* dest, uae_u32 integer);

static DWORD __stdcall CommRead (void*);
static DWORD __stdcall CommWrite (void*);


/***********************************************************************
 *
 * FUNCTION:	Platform::Initialize
 *
 * DESCRIPTION:	Initializes platform-dependent stuff
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::Initialize (void)
{
}


/***********************************************************************
 *
 * FUNCTION:	Platform::Reset
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::Reset (void)
{
}


/***********************************************************************
 *
 * FUNCTION:	I_wish_I_had_lamda_functions
 *
 * DESCRIPTION:	DlgProc for CommonDialog.  Sets up the dialog
 *				according to the specifications of the caller and
 *				handles clicks on the buttons
 *
 * PARAMETERS:	Standard DlgProc parameters
 *
 * RETURNED:	Standard DlgProc results.
 *
 ***********************************************************************/

static const char*	gTitle;
static const char*	gMessage;
static int			gFlags;
static int			gDefaultID;
static int			gCancelID;
static int			gExtraID;

static BOOL CALLBACK I_wish_I_had_lamda_functions (HWND hwnd, UINT msg,
												   WPARAM wparam, LPARAM lparam)
{
	switch (msg)
	{
		case WM_INITDIALOG:
		{
			HWND	defaultButton = ::GetDlgItem (hwnd, IDOK);
			HWND	cancelButton = ::GetDlgItem (hwnd, IDCANCEL);
			HWND	extraButton = ::GetDlgItem (hwnd, IDC_EXTRA);
			HWND	messageBox = ::GetDlgItem (hwnd, IDC_MESSAGE);

			assert (defaultButton);
			assert (cancelButton);
			assert (extraButton);
			assert (messageBox);

			// Sort out the button situation.

			if (gFlags & Errors::kOKCancelMask)
			{
				string	okString = Platform::GetString (kStr_OK);
				string	cancelString = Platform::GetString (kStr_Cancel);

				if ((gFlags & Errors::kOKCancelMask) == Errors::kOK)
				{
					::SetWindowText (defaultButton, okString.c_str ());
					::ShowWindow (cancelButton, SW_HIDE);
					::ShowWindow (extraButton, SW_HIDE);

					gDefaultID = Errors::kOK;
					gCancelID = 0;
					gExtraID = 0;
				}
				else if ((gFlags & Errors::kOKCancelMask) == Errors::kCancel)
				{
					::SetWindowText (defaultButton, cancelString.c_str ());
					::ShowWindow (cancelButton, SW_HIDE);
					::ShowWindow (extraButton, SW_HIDE);

					gDefaultID = Errors::kCancel;
					gCancelID = 0;
					gExtraID = 0;
				}
				else
				{
					::SetWindowText (defaultButton, okString.c_str ());
					::SetWindowText (cancelButton, cancelString.c_str ());
					::ShowWindow (extraButton, SW_HIDE);

					gDefaultID = Errors::kOK;
					gCancelID = Errors::kCancel;
					gExtraID = 0;
				}
			}
			else if (gFlags & Errors::kYesNoMask)
			{
				string	yesString = Platform::GetString (kStr_Yes);
				string	noString = Platform::GetString (kStr_No);

				if ((gFlags & Errors::kYesNoMask) == Errors::kYes)
				{
					::SetWindowText (defaultButton, yesString.c_str ());
					::ShowWindow (cancelButton, SW_HIDE);
					::ShowWindow (extraButton, SW_HIDE);

					gDefaultID = Errors::kYes;
					gCancelID = 0;
					gExtraID = 0;
				}
				else if ((gFlags & Errors::kYesNoMask) == Errors::kNo)
				{
					::SetWindowText (defaultButton, noString.c_str ());
					::ShowWindow (cancelButton, SW_HIDE);
					::ShowWindow (extraButton, SW_HIDE);

					gDefaultID = Errors::kNo;
					gCancelID = 0;
					gExtraID = 0;
				}
				else
				{
					::SetWindowText (defaultButton, yesString.c_str ());
					::SetWindowText (cancelButton, noString.c_str ());
					::ShowWindow (extraButton, SW_HIDE);

					gDefaultID = Errors::kYes;
					gCancelID = Errors::kNo;
					gExtraID = 0;
				}
			}
			else if (gFlags & Errors::kContinueDebugResetMask)
			{
				string	continueString = Platform::GetString (kStr_Continue);
				string	debugString = Platform::GetString (kStr_Debug);
				string	resetString = Platform::GetString (kStr_Reset);

				::SetWindowText (defaultButton, continueString.c_str ());
				::SetWindowText (cancelButton, debugString.c_str ());
				::SetWindowText (extraButton, resetString.c_str ());

				if ((gFlags & Errors::kContinue) == 0)
				{
				//	::ShowWindow (defaultButton, SW_HIDE);
					::EnableWindow (defaultButton, false);
				}

				if ((gFlags & Errors::kDebug) == 0)
				{
				//	::ShowWindow (cancelButton, SW_HIDE);
					::EnableWindow (cancelButton, false);
				}

				if ((gFlags & Errors::kReset) == 0)
				{
				//	::ShowWindow (extraButton, SW_HIDE);
					::EnableWindow (extraButton, false);
				}

				gDefaultID = Errors::kContinue;
				gCancelID = Errors::kDebug;
				gExtraID = Errors::kReset;
			}

			// Determine the height the text will need.

			RECT	extraRect;
			::GetWindowRect (extraButton, &extraRect);
			::MapWindowPoints (NULL, hwnd, (LPPOINT) &extraRect, 2);

			RECT	mbRect;
			::GetWindowRect (messageBox, &mbRect);
			::MapWindowPoints (NULL, hwnd, (LPPOINT) &mbRect, 2);

			HDC		mbDC = ::GetDC (hwnd);
			HGDIOBJ	oldFont = ::SelectObject (mbDC, ::GetStockObject (DEFAULT_GUI_FONT));

			RECT	newMBRect = mbRect;
			int		height = ::DrawText (mbDC, gMessage, -1, &newMBRect,
								DT_LEFT | DT_WORDBREAK | DT_CALCRECT);

			::SelectObject (mbDC, oldFont);
			::ReleaseDC (hwnd, mbDC);


			// Don't let the height of the window get so small that
			// the bottom button is clipped off.

			if (newMBRect.bottom < extraRect.bottom)
			{
				newMBRect.bottom = extraRect.bottom;
			}

			int	delta = newMBRect.bottom - mbRect.bottom;	// amt to grow

			// Resize the dialog.

			RECT	frameRect;
			::GetWindowRect (hwnd, &frameRect);

			::MoveWindow (	hwnd, frameRect.left, frameRect.top,
							frameRect.right - frameRect.left,
							frameRect.bottom - frameRect.top + delta, true);

			// Resize the message control

			::MoveWindow (	messageBox, newMBRect.left, newMBRect.top,
							newMBRect.right - newMBRect.left,
							newMBRect.bottom - newMBRect.top, true);

			// Center the window and set the text.

			::CenterWindow (hwnd);
			::SetDlgItemText (hwnd, IDC_MESSAGE, gMessage);
			break;
		}

		case WM_COMMAND:
		{
			switch (LOWORD (wparam))
			{
				case IDOK:
				case IDCANCEL:
				case IDC_EXTRA:
					EndDialog (hwnd, LOWORD (wparam));
					break;
			}
			break;
		}

		default:
			return FALSE;
	}

	return TRUE;
}


/***********************************************************************
 *
 * FUNCTION:	Platform::CommonDialog
 *
 * DESCRIPTION:	Display a dialog box to handle a number of needs.
 *				Based on "flags" the dialog can display OK/Cancel
 *				buttons, Yes/No buttons, or Continue/Debug/Reset
 *				buttons.
 *
 * PARAMETERS:	title - text to appear in dialog box title.
 *
 *				msg - text to appear in the body of the dialog box.
 *					The dialog box is resized to accomodate any
 *					amount of text.
 *
 *				flags - determines buttons to be displayed.
 *
 * RETURNED:	Value indicating what button was selected. Example
 *				values are Errors::kOK, Errors::kCancel, etc.
 *
 ***********************************************************************/

int Platform::CommonDialog (const char* title, const char* msg, int flags)
{
	// Store our parameters where our callback function can get to it.

	gTitle = title;
	gMessage = msg;
	gFlags = flags;

	int	button = DialogBox (gInstance, MAKEINTRESOURCE (IDD_COMMON),
							LCD_GetWindow (), I_wish_I_had_lamda_functions);

	return	(button == IDOK) ? gDefaultID :
			(button == IDCANCEL) ? gCancelID :
			gExtraID;
}


/***********************************************************************
 *
 * FUNCTION:	Platform::GetString
 *
 * DESCRIPTION:	Return the resource-based string based on the given ID.
 *
 * PARAMETERS:	id - id of the string to return.  IDs are defined
 *					in Strings.r.h.
 *
 * RETURNED:	String object containing the requested text.  If the
 *				resource could not be locationed, the returned string
 *				object contains the error message "<missing string #>".
 *				Clients can test the result for this pattern to see if
 *				the request succeeded or not.  In general, its a
 *				programming error if the request fails.
 *
 ***********************************************************************/

string Platform::GetString (int id)
{
	int		amtRead;
	int		size = 256;		// Arbitrary starting size
	string	result;

	// Allocate a buffer; load the string; loop if the buffer wasn't
	// big enough.

	while (1)
	{
		result.resize (size);
		amtRead = ::LoadString(gInstance, id, &result[0], size);

		if (amtRead == 0)
		{
			char	buffer[20];
			sprintf (buffer, "%d", id);
			return string ("<missing string ") + buffer + ">";
		}

		if (amtRead < (size - 1))
			break;
		size += 256;	// Arbitrary amount by which to increment
	};

	result.resize (amtRead);

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	Platform::GetIDForError
 *
 * DESCRIPTION:	Map a platform-specific error number into an id for
 *				an error string.
 *
 * PARAMETERS:	error - platform-specific error code.
 *
 * RETURNED:	String ID for corresponding error message.  If a
 *				string specific to the given error cannot be found,
 *				kStr_GenericError is returned.
 *
 ***********************************************************************/

int Platform::GetIDForError (ErrCode error)
{
	switch (error)
	{
		case ERROR_FILE_NOT_FOUND:		return kStr_FileNotFound;		// The system cannot find the file specified
		case ERROR_PATH_NOT_FOUND:		return kStr_FileNotFound;		// The system cannot find the path specified
		case ERROR_TOO_MANY_OPEN_FILES:	return kStr_TooManyFilesOpen;	// The system cannot open the file
		case ERROR_ACCESS_DENIED:		return kStr_SerialPortBusy;		// Access is denied
		case ERROR_NOT_ENOUGH_MEMORY:	return kStr_MemFull;			// Not enough storage is available to process this command
		case ERROR_OUTOFMEMORY:			return kStr_MemFull;			// Not enough storage is available to complete this operation
		case ERROR_INVALID_DRIVE:		return kStr_DiskMissing;		// The system cannot find the drive specified
		case ERROR_WRITE_PROTECT:		return kStr_DiskLocked;			// The media is write protected
		case ERROR_NOT_READY:			return kStr_GenericError;		// The device is not ready
		case ERROR_SHARING_VIOLATION:	return kStr_FileBusy;			// The process cannot access the file because it is being used by another process
		case ERROR_LOCK_VIOLATION:		return kStr_GenericError;		// The process cannot access the file because another process has locked a portion of the file
		case ERROR_HANDLE_DISK_FULL:	return kStr_DiskFull;			// The disk is full
		case ERROR_DUP_NAME:			return kStr_GenericError;		// A duplicate name exists on the network
		case ERROR_BAD_NETPATH:			return kStr_GenericError;		// The network path was not found
		case ERROR_NETWORK_BUSY:		return kStr_GenericError;		// The network is busy
		case ERROR_DEV_NOT_EXIST:		return kStr_DiskMissing;		// The specified network resource or device is no longer available
		case ERROR_FILE_EXISTS:			return kStr_GenericError;		// The file exists
		case ERROR_CANNOT_MAKE:			return kStr_GenericError;		// The directory or file cannot be created
		case ERROR_DRIVE_LOCKED:		return kStr_DiskLocked;			// The disk is in use or locked by another process
		case ERROR_OPEN_FAILED:			return kStr_GenericError;		// The system cannot open the device or file specified
		case ERROR_DISK_FULL:			return kStr_DiskFull;			// There is not enough space on the disk
		case ERROR_INVALID_NAME:		return kStr_BadFileName;		// The filename, directory name, or volume label syntax is incorrect
		case ERROR_BAD_PATHNAME:		return kStr_BadFileName;		// The specified path is invalid
		case ERROR_ALREADY_EXISTS:		return kStr_DuplicateFileName;	// Cannot create a file when that file already exists
		case ERROR_FILENAME_EXCED_RANGE:return kStr_BadFileName;		// The filename or extension is too long
		case ERROR_DIRECTORY:			return kStr_BadFileName;		// The directory name is invalid

		case ERROR_BADDB:				return kStr_GenericError;		// The configuration registry database is corrupt
		case ERROR_BADKEY:				return kStr_GenericError;		// The configuration registry key is invalid
		case ERROR_CANTOPEN:			return kStr_GenericError;		// The configuration registry key could not be opened
		case ERROR_CANTREAD:			return kStr_GenericError;		// The configuration registry key could not be read
		case ERROR_CANTWRITE:			return kStr_GenericError;		// The configuration registry key could not be written
		case ERROR_REGISTRY_CORRUPT:	return kStr_GenericError;		// The Registry is corrupt...
		case ERROR_REGISTRY_IO_FAILED:	return kStr_GenericError;		// An I/O operation initiated by the Registry failed...
		case ERROR_NO_LOG_SPACE:		return kStr_GenericError;		// System could not allocate the required space in a Registry log

		case ERROR_IO_DEVICE:			return kStr_GenericError;		// The request could not be performed because of an I/O device error.
		case ERROR_SERIAL_NO_DEVICE:	return kStr_GenericError;		// No serial device was successfully initialized.  The serial driver will unload.
		case ERROR_IRQ_BUSY:			return kStr_SerialPortBusy;		// Unable to open a device that was sharing an interrupt request (IRQ) with other devices. At least one other device that uses that IRQ was already opened.

		case ERROR_TIMEOUT:				return kStr_TimeoutError;		// This operation returned because the timeout period expired
	}

	return kStr_GenericError;
}


/***********************************************************************
 *
 * FUNCTION:	Platform::GetIDForRecovery
 *
 * DESCRIPTION:	Map the given platform-specific error code into and ID
 *				for an error recovery string.
 *
 * PARAMETERS:	error - platform-specific error code.
 *
 * RETURNED:	String ID for corresponding recovery message.  If a
 *				string specific to the given error cannot be found,
 *				zero is returned.
 *
 ***********************************************************************/

int Platform::GetIDForRecovery (ErrCode error)
{
	return 0;
}


/***********************************************************************
 *
 * FUNCTION:	Platform::GetShortVersionString
 *
 * DESCRIPTION:	Returns a short version string.  The format of the
 *				string is:
 *
 *					#.# (.#) ([dab]#)
 *
 *					# = one or more numeric digits
 *					. = literal '.'
 *					Patterns in parentheses are optional
 *					Patterns in brackets mean "one of these characters"
 *					Spaces are shown above for clarity; they do not
 *						appear in the string
 *
 *				Valid strings would be: 2.1d7, 2.1.1b14, 2.99, 2.1.1.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	Formatted version string.
 *
 ***********************************************************************/

string Platform::GetShortVersionString (void)
{
	string	result;
	char	szFullPath[MAX_PATH];

	::GetModuleFileName (gInstance, szFullPath, sizeof(szFullPath));

 	DWORD	dummy;
 	DWORD	infoSize = ::GetFileVersionInfoSize (szFullPath, &dummy);

	if (infoSize > 0)
	{
        // Allocate memory to hold the verinfo block

		LPVOID	versionPtr = _alloca(infoSize);

        if (::GetFileVersionInfo (szFullPath, dummy, infoSize, versionPtr))
		{
			LPVOID	versString = NULL;
			UINT	versSize = 0;

			::VerQueryValue (	versionPtr, 
								"\\StringFileInfo\\040904b0\\FileVersion", 
								&versString, &versSize);

			if ((versString != NULL) && (versSize > 0))
			{
				result = (char*) versString;
			}
		}
	}

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	Platform::GetLoadableFileList
 *
 * DESCRIPTION:	Scan the given directory for files that can be loaded
 *				into the Palm OS environment and add them to the
 *				given collection object.  Loadable files are PRCs,
 *				PDBs, and PQAs.
 *
 * PARAMETERS:	directoryName - name of the directory to search.
 *					This directory is assumed to be in the emulator's
 *					containing directory.
 *
 *				fileList - collection to receive the found files.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::GetLoadableFileList (string directoryName, LoadableList& fileList)
{
	// Get the application's directory.

	char	path[_MAX_PATH], drive[_MAX_DRIVE], dir[_MAX_DIR];

	::GetModuleFileName (gInstance, path, _MAX_PATH);
	_splitpath (path, drive, dir, NULL, NULL);
	_makepath (path, drive, dir, NULL, NULL);

	string	pathName (path + directoryName + "\\");
	string	searchPattern (pathName + "*");

	// Start iterating over it.  We'll iterate over all the contents
	// and filter by filetype by hand.

	WIN32_FIND_DATA	fileData;
	HANDLE			hSearch = ::FindFirstFile (searchPattern.c_str (), &fileData);
	if (hSearch == INVALID_HANDLE_VALUE)
	{
		// If an error occurs (probably no such directory), just return.

		return;
	}


	BOOL	finished = FALSE;
	while (!finished)
	{
		if (::EndsWith (fileData.cFileName, ".prc") ||
			::EndsWith (fileData.cFileName, ".pdb") ||
			::EndsWith (fileData.cFileName, ".pqa"))
		{
			string	fullPath (pathName + fileData.cFileName);
			fileList.push_back (LoadFilePB (FileReference (fullPath)));
		}

		if (!::FindNextFile (hSearch, &fileData))
		{
			DWORD	lastError = GetLastError ();
			if (lastError != ERROR_SUCCESS)
			{
				// Blow out of here on *any* error, not just ERROR_NO_MORE_FILES.
				finished = TRUE;
			}
		}
	}

	::FindClose (hSearch);
}


/***********************************************************************
 *
 * FUNCTION:	Platform::Delay
 *
 * DESCRIPTION:	Delay for a time period appropriate for a sleeping 68328.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::Delay (void)
{
	// Delay 10 msecs.  Delaying by this amount pauses us 1/100 of a second,
	// which is the rate at which the device's tickcount counter increments.
	//
	// Wait on an event instead of just calling Sleep(10) so that another
	// thread can wake us up before our time.

	::WaitForSingleObject (gCPUDelayEvent, 10);
}


/***********************************************************************
 *
 * FUNCTION:	Platform::Cycle
 *
 * DESCRIPTION:	Perform platform-specific actions required after every
 *				opcode is executed.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::Cycle (void)
{
	// Nothing to do in Windows.
}


/***********************************************************************
 *
 * FUNCTION:	Platform::AllocateMemory
 *
 * DESCRIPTION:	Allocate a buffer of the given size, optionally clearing
 *				the contents to zero.
 *
 * PARAMETERS:	size - size of the buffer to create.
 *
 *				clear - true if the buffer should be cleared.
 *
 * RETURNED:	Pointer to the allocated buffer.
 *
 ***********************************************************************/

void* Platform::AllocateMemory (size_t size, Bool clear)
{
	void*	result;

	if (clear)
		result = calloc (size, 1);
	else
		result = malloc (size);

	Errors::ThrowIfNULL (result);

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	Platform::ReallocMemory
 *
 * DESCRIPTION:	Resize the given buffer to the new given size.
 *
 * PARAMETERS:	p - Pointer to buffer to resize
 *
 *				size - desired new size of the buffer.
 *
 * RETURNED:	The buffer is allowed to move during the resize process,
 *				so this function returns the buffer's new location.
 *
 ***********************************************************************/

void* Platform::ReallocMemory (void* p, size_t size)
{
	void*	result = realloc (p, size);

	Errors::ThrowIfNULL (result);

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	Platform::RealDisposeMemory
 *
 * DESCRIPTION:	Primitve function for releasing allocated memory.  This
 *				function is called by a templatized version of the
 *				function that deals with accepting pointers to any
 *				data type and casting it to "void*" so that this
 *				function can be called.
 *
 * PARAMETERS:	p - pointer to the chunk of memory to release.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::RealDisposeMemory (void* p)
{
	if (p)
	{
		free (p);
	}
}


/***********************************************************************
 *
 * FUNCTION:	Platform::StopOnResetKeyDown
 *
 * DESCRIPTION:	Return whether or not the user wants to set a break-
 *				point at some early place in the ROM.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	True if an early breakpoint should be set, false otherwise.
 *
 ***********************************************************************/

Bool Platform::StopOnResetKeyDown (void)
{
	// Comment this out for now.  It seems that Windows doesn't always return
	// the expected result.  That is, this function often returns TRUE even
	// though the Control is not down.
	//
	// Since this functionality is really only required by Palm OS engineers,
	// and since they're working on Macs, we don't really need this feature
	// in the Windows version now anyway.

//	return ::GetAsyncKeyState (VK_CONTROL) != 0;
	return false;
}


/***********************************************************************
 *
 * FUNCTION:	Platform::GetMilliseconds
 *
 * DESCRIPTION:	Return a current "millisecond count".  This value
 *				correspond to any current time; it's just used to
 *				measure elapsed time.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	The current millisecond count.
 *
 ***********************************************************************/

uae_u32 Platform::GetMilliseconds (void)
{
	return (uae_u32) ::GetTickCount();
}


/***********************************************************************
 *
 * FUNCTION:	Platform::HeadPatch
 *
 * DESCRIPTION:	Perform any platform-specific patching actions.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

#if 0
int Platform::HeadPatch (uae_u16 iATrap)
{
	switch (iATrap)
	{
		case sysTrapSndDoCmd:
		{
			SndCommandType*	sc = (SndCommandType*) get_real_address(get_long(m68k_areg(regs, 7) + 4));
			if (do_get_mem_word(&sc->cmd) == 1)
			{
				if (gEnableSounds)
				{
					Beep(do_get_mem_word(&sc->param1lo), do_get_mem_word(&sc->param2));
				}

				return true;
			}

			break;
		}
	}

	return false;
}
#endif


/***********************************************************************
 *
 * FUNCTION:	Platform::ConnectToDebugger
 *
 * DESCRIPTION:	Called from Debug::Initialize to connect to an
 *				external debugger.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	Error code for any errors that occur.
 *
 ***********************************************************************/

ErrCode Platform::ConnectToDebugger (void)
{
	// Don't call MMFOpenFiles here.  Doing so would create the
	// hidden window used for message passing in a thread different
	// from the one that processes the messages.

//	::MMFOpenFiles (NULL, HandlePacket, (LPARAM) LCD_GetWindow (),
//					kFromEmulatorFileName, kToEmulatorFileName);

	return errNone;
}


/***********************************************************************
 *
 * FUNCTION:	Platform::SendPacket
 *
 * DESCRIPTION:	Send a packet of data to the debugger.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	Error code for any errors that occur.
 *
 ***********************************************************************/

ErrCode Platform::SendPacket (const void* pkt, long size)
{
	return (ErrCode) ::MMFSendData (pkt, size);
}


/***********************************************************************
 *
 * FUNCTION:	Platform::DisconnectFromDebugger
 *
 * DESCRIPTION:	Break the connection with the external debugger.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::DisconnectFromDebugger (void)
{
	::MMFCloseFiles (gInstance);
}


/***********************************************************************
 *
 * FUNCTION:	Platform::ExitDebugger
 *
 * DESCRIPTION:	Perform platform-specific operations when debug mode is
 *				exited.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::ExitDebugger (void)
{
}


/***********************************************************************
 *
 * FUNCTION:	Platform::ViewDrawLine
 *
 * DESCRIPTION:	Draw a line between the given points on the LCD.
 *
 * PARAMETERS:	xStart, yStart, xEnd, yEnd - coordinates of line to draw.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::ViewDrawLine (int xStart, int yStart, int xEnd, int yEnd)
{
	// !!! Do nothing for now, as any drawing we do gets overwritten
	// by the next blit of the LCD buffer to the screen.
}


/***********************************************************************
 *
 * FUNCTION:	Platform::ViewDrawPixel
 *
 * DESCRIPTION:	Draw a pixel at the given location on the LCD.
 *
 * PARAMETERS:	xPos, yPos - location of pixel to plot.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::ViewDrawPixel (int xPos, int yPos)
{
	// !!! Do nothing for now, as any drawing we do gets overwritten
	// by the next blit of the LCD buffer to the screen.
}


// ---------------------------------------------------------------------------
//		 Serial port emulation
// ---------------------------------------------------------------------------

#define PRINTF	if (!LogSerial ()) ; else AppendMsg

/***********************************************************************
 *
 * FUNCTION:	Platform::OpenSerialPort
 *
 * DESCRIPTION:	Open the serial port with default settings.  A call to
 *				this function  should be followed by a call to
 *				SetSerialPortSettings.
 *
 *				This function will close the serial port before trying
 *				to open it.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	Error code for any errors that occur.
 *
 ***********************************************************************/

ErrCode Platform::OpenSerialPort (void)
{
	PRINTF ("Serial Emulation (Windows): Attempting to open serial port...");

	ErrCode	err = kError_NoError;

	// Make sure the serial port is closed.

	CloseSerialPort ();

	// Create the comm handle used for serial communications.

	string	commPort = gPrefs->GetCommPort ();

	if (commPort.length () > 0)
	{
		PRINTF ("Serial Emulation (Windows): Opening serial port...");

		gCommHandle = ::CreateFile (commPort.c_str (), GENERIC_READ | GENERIC_WRITE, 0,
								NULL, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, NULL);

		if (gCommHandle == INVALID_HANDLE_VALUE)
		{
			err = (ErrCode) ::GetLastError ();
			gCommHandle = NULL;
			gPrefs->SetCommPort ("");

			PRINTF ("Serial Emulation (Windows): Error %ld serial port...", err);

			Errors::ReportIfError (kStr_OpenTheSerialPort, err, 0, false);

			return err;
		}
	}
	else
	{
		PRINTF ("Serial Emulation (Windows): No port selected in the Properties dialog box...");
	}

	if (gCommHandle)
	{
		PRINTF ("Serial Emulation (Windows): Creating serial port handler threads...");

		// !!! Re-examine these settings.  Understand what they are,
		// and compare them to the MSDN samples.

		COMMTIMEOUTS	ctmo;

		ctmo.ReadIntervalTimeout			= 50;
		ctmo.ReadTotalTimeoutMultiplier		= 0;
		ctmo.ReadTotalTimeoutConstant		= 100;
		ctmo.WriteTotalTimeoutMultiplier	= 0;
		ctmo.WriteTotalTimeoutConstant		= 0;

		::SetCommTimeouts (gCommHandle, &ctmo);

		// Create the threads that will interact with the port.  Make
		// them high priority threads.

		DWORD tid;
		HANDLE commRead = ::CreateThread (
								NULL,		// security descriptor
								0,			// stack size
								CommRead,	// thread function
								NULL,		// parameter to thread function
								0,			// creation flags
								&tid);		// address of thread ID

		HANDLE commWrite = ::CreateThread (
								NULL,		// security descriptor
								0,			// stack size
								CommWrite,	// thread function
								NULL,		// parameter to thread function
								0,			// creation flags
								&tid);		// address of thread ID

		::SetThreadPriority (commRead, THREAD_PRIORITY_HIGHEST);
		::SetThreadPriority (commWrite, THREAD_PRIORITY_HIGHEST);

		::CloseHandle (commRead);
		::CloseHandle (commWrite);
	}

	return err;
}


/***********************************************************************
 *
 * FUNCTION:	Platform::SetSerialPortSettings
 *
 * DESCRIPTION:	Establish the serial port settings.  This function
 *				should be called after OpenSerialPort.
 *
 * PARAMETERS:	parity - desired parity
 *
 *				stopBits - number of stop bits: 1, or 2
 *
 *				dataBits - number of data bits: 7 or 8
 *
 *				baud - desired baud (e.g., 57600)
 *
 *				hwHandshaking - true for hardware hardshaking
 *
 * RETURNED:	Error code for any errors that occur.
 *
 ***********************************************************************/
// .

ErrCode Platform::SetSerialPortSettings (	UART::Parity	parity,
											int				stopBits,
											int				dataBits,
											uae_u32			baud,
											Bool			hwHandshaking)
{
	PRINTF ("Serial Emulation (Windows): Setting settings...");

	if (Platform::SerialPortOpen ())
	{
		PRINTF ("Serial Emulation (Windows): Serial port open, so plowing ahead...");
		PRINTF ("	parity			= %ld", (long) parity);
		PRINTF ("	stopBits		= %ld", (long) stopBits);
		PRINTF ("	dataBits		= %ld", (long) dataBits);
		PRINTF ("	baud			= %ld", (long) baud);
		PRINTF ("	hwHandshaking	= %ld", (long) hwHandshaking);

		DCB	dcb;
		dcb.DCBlength		= sizeof (dcb);
		::GetCommState (gCommHandle, &dcb);

		dcb.BaudRate		= baud;
		dcb.fBinary			= TRUE;
		dcb.fParity			= TRUE;
		dcb.fOutxCtsFlow	= hwHandshaking;
		dcb.fOutxDsrFlow	= FALSE;
		dcb.fDtrControl		= DTR_CONTROL_ENABLE;
		dcb.fDsrSensitivity	= FALSE;
		dcb.fOutX			= FALSE;
		dcb.fInX			= FALSE;
		dcb.fErrorChar		= FALSE;
		dcb.fNull			= FALSE;
		dcb.fRtsControl		= hwHandshaking ? RTS_CONTROL_HANDSHAKE : RTS_CONTROL_ENABLE;
		dcb.fAbortOnError	= TRUE;
		dcb.ByteSize		= dataBits;
		dcb.Parity			= parity == UART::kNoParity ? NOPARITY :
							  parity == UART::kOddParity ? ODDPARITY : EVENPARITY;
		dcb.StopBits		= stopBits == 1 ? ONESTOPBIT : TWOSTOPBITS;

		PRINTF ("	dcb.DCBlength			= %ld", (long) dcb.DCBlength);
		PRINTF ("	dcb.BaudRate			= %ld", (long) dcb.BaudRate);
		PRINTF ("	dcb.fBinary				= %ld", (long) dcb.fBinary);
		PRINTF ("	dcb.fParity				= %ld", (long) dcb.fParity);
		PRINTF ("	dcb.fOutxCtsFlow		= %ld", (long) dcb.fOutxCtsFlow);
		PRINTF ("	dcb.fOutxDsrFlow		= %ld", (long) dcb.fOutxDsrFlow);
		PRINTF ("	dcb.fDtrControl			= %ld", (long) dcb.fDtrControl);
		PRINTF ("	dcb.fDsrSensitivity		= %ld", (long) dcb.fDsrSensitivity);
		PRINTF ("	dcb.fTXContinueOnXoff	= %ld", (long) dcb.fTXContinueOnXoff);
		PRINTF ("	dcb.fOutX				= %ld", (long) dcb.fOutX);
		PRINTF ("	dcb.fInX				= %ld", (long) dcb.fInX);
		PRINTF ("	dcb.fErrorChar			= %ld", (long) dcb.fErrorChar);
		PRINTF ("	dcb.fNull				= %ld", (long) dcb.fNull);
		PRINTF ("	dcb.fRtsControl			= %ld", (long) dcb.fRtsControl);
		PRINTF ("	dcb.fAbortOnError		= %ld", (long) dcb.fAbortOnError);
		PRINTF ("	dcb.fDummy2				= %ld", (long) dcb.fDummy2);
		PRINTF ("	dcb.wReserved			= %ld", (long) dcb.wReserved);
		PRINTF ("	dcb.XonLim				= %ld", (long) dcb.XonLim);
		PRINTF ("	dcb.XoffLim				= %ld", (long) dcb.XoffLim);
		PRINTF ("	dcb.ByteSize			= %ld", (long) dcb.ByteSize);
		PRINTF ("	dcb.Parity				= %ld", (long) dcb.Parity);
		PRINTF ("	dcb.StopBits			= %ld", (long) dcb.StopBits);
		PRINTF ("	dcb.XonChar				= %ld", (long) dcb.XonChar);
		PRINTF ("	dcb.XoffChar			= %ld", (long) dcb.XoffChar);
		PRINTF ("	dcb.ErrorChar			= %ld", (long) dcb.ErrorChar);
		PRINTF ("	dcb.EofChar				= %ld", (long) dcb.EofChar);
		PRINTF ("	dcb.EvtChar				= %ld", (long) dcb.EvtChar);
		PRINTF ("	dcb.wReserved1			= %ld", (long) dcb.wReserved1);

		BOOL	result = ::SetCommState (gCommHandle, &dcb);

		if (result)
		{
			PRINTF ("Serial Emulation (Windows): Setting settings...succeeded.");

			return errNone;
		}
		else
		{
			DWORD	err = ::GetLastError ();
			PRINTF ("Serial Emulation (Windows): Setting settings...failed (%ld).", err);
			return err;
		}
	}
	else
	{
		PRINTF ("Serial Emulation (Windows): Serial port closed, so not setting settings...");
	}

	return kError_CommNotOpen;
}

/***********************************************************************
 *
 * FUNCTION:	Platform::GetSerialPortSettings
 *
 * DESCRIPTION:	Return the serial port settings.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::GetSerialPortSettings (	UART::Parity&	parity,
										int&			stopBits,
										int&			dataBits,
										uae_u32&		baud,
										Bool&			hwHandshaking)
{
	parity			= UART::gLastParity;
	stopBits		= UART::gLastStopBits;
	dataBits		= UART::gLastDataBits;
	baud			= UART::gLastBaud;
	hwHandshaking	= UART::gLastHwHandshaking;
}


/***********************************************************************
 *
 * FUNCTION:	Platform::SerialPortOpen
 *
 * DESCRIPTION:	Returns whether or not the serial port is open for
 *				business.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

Bool Platform::SerialPortOpen (void)
{
	return gCommHandle != NULL;
}


/***********************************************************************
 *
 * FUNCTION:	Platform::CloseSerialPort
 *
 * DESCRIPTION:	Close the serial port.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::CloseSerialPort (void)
{
	PRINTF ("Serial Emulation (Windows): Attempting to close serial port...");

	if (Platform::SerialPortOpen ())
	{
		PRINTF ("Serial Emulation (Windows): Closing serial port...");

		// Set gCommHandle to NULL now. That will cause the CommRead
		// thread to quit after we close the handle.

		HANDLE	commHandle = gCommHandle;
		gCommHandle = NULL;

		::CloseHandle (commHandle);
	}
	else
	{
		PRINTF ("Serial Emulation (Windows): Serial port already closed...");
	}
}


/***********************************************************************
 *
 * FUNCTION:	Platform::TransmitTxFIFO
 *
 * DESCRIPTION:	Transmit any bytes in the Tx FIFO out the serial port.
 *				Assumes that the serial port is open.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::TransmitTxFIFO (void)
{
	// Nothing to do on Windows; the Comm threads will pick up the new
	// data automatically.
}


/***********************************************************************
 *
 * FUNCTION:	Platform::ReceiveRxFIFO
 *
 * DESCRIPTION:	Fills up the Rx FIFO with as many bytes as it can from
 *				the host serial port.  Assumes that the serial port is
 *				open.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::ReceiveRxFIFO (void)
{
	// Nothing to do on Windows; the Comm threads will pick up the new
	// data automatically.
}


/***********************************************************************
 *
 * FUNCTION:	CommRead
 *
 * DESCRIPTION:	This function sits in its own thread, waiting for data
 *				to show up in the serial port.  If data arrives, this
 *				function plucks it out and stores it in a thread-safe
 *				queue.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

DWORD __stdcall CommRead (void*)
{
	HANDLE	hOverlappedEvent = ::CreateEvent (NULL, TRUE, FALSE, NULL);

	while (gCommHandle)
	{
		DWORD		error;
		OVERLAPPED	ov;
		memset (&ov, 0, sizeof(ov));
		ov.hEvent = hOverlappedEvent;

		BYTE	buf[16];
		DWORD	n = UART::GetRxQueue ().GetFree ();

		if (n == 0)
		{
			// No space for new data, so don't read any.

			::Sleep(10);
			continue;
		}

		if (::ReadFile (gCommHandle, buf, n, &n, &ov) ||
			::GetLastError () == ERROR_IO_PENDING && ::GetOverlappedResult (gCommHandle, &ov, &n, TRUE))
		{
			if (n > 0)
			{
				if (LogSerialData ())
					AppendData (buf, n, "Serial Emulation (Windows): Received data:");
				else
					PRINTF ("Serial Emulation (Windows): Received %ld serial bytes.", n);

				for (DWORD ii = 0; ii < n; ++ii)
				{
					UART::GetRxQueue ().Put (buf[ii]);
				}
			}
		}
	}

	::CloseHandle (hOverlappedEvent);

	return 0;
}


/***********************************************************************
 *
 * FUNCTION:	CommWrite
 *
 * DESCRIPTION:	This function sits in its own thread, waiting for data
 *				to show up in the write queue.  If data arrives, this
 *				function plucks it out and sends it out to the port.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

DWORD __stdcall CommWrite (void*)
{
	HANDLE	hOverlappedEvent = ::CreateEvent (NULL, TRUE, FALSE, NULL);

	while (gCommHandle)
	{
		if (!UART::GetTxQueue ().WaitForDataAvailable (1000))
		{
			continue;
		}

		BYTE	buf[8];
		DWORD	n = 0;

		while (n < sizeof (buf) && UART::GetTxQueue ().GetUsed () > 0)
		{
			buf[n++] = UART::GetTxQueue ().Get ();
		}

		if (n == 0)
		{
			// Never mind, there really wasn't any data in the buffer
			// (see note in TByteQueuePrivate::Put).

			continue;
		}

		if (LogSerialData ())
			AppendData (buf, n, "Serial Emulation (Windows): Transmitted data:");
		else
			PRINTF ("Serial Emulation (Windows): Transmitted %ld serial bytes.", n);

		OVERLAPPED ov;
		memset (&ov, 0, sizeof (ov));
		ov.hEvent = hOverlappedEvent;

		if (!::WriteFile (gCommHandle, buf, n, &n, &ov) && ::GetLastError () == ERROR_IO_PENDING)
		{
			::GetOverlappedResult (gCommHandle, &ov, &n, TRUE);
		}
	}

	::CloseHandle (hOverlappedEvent);

	return 0;
}


// ---------------------------------------------------------------------------
//		 Load File Progress Window
// ---------------------------------------------------------------------------

static HWND		gLFPW_Dlg;
static int		gLFPW_MaxFile;
static int		gLFPW_CurFile;
static char		gLFPW_CurFileName[MAX_PATH];
static int		gLFPW_MaxValue;
static int		gLFPW_CurValue;


/***********************************************************************
 *
 * FUNCTION:	LFPW_DlgProc
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

static BOOL CALLBACK LFPW_DlgProc (	HWND hwnd,
									UINT msg,
									WPARAM wParam,
									LPARAM lParam)
{
	switch (msg)
	{
		case WM_INITDIALOG:
		{
			::CenterWindow (hwnd);
			break;
		}

		case WM_COMMAND:
		{
			switch (wParam)
			{
				case IDCANCEL:
				{
					::EndDialog (hwnd, 0);
					return TRUE;
				}
			}

			return FALSE;
		}

		default:
			return FALSE;
	}

	return TRUE;
}


/***********************************************************************
 *
 * FUNCTION:	Platform::LFPW_Open
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::LFPW_Open (void)
{
	gLFPW_Dlg = ::CreateDialogParam (	gInstance,
										MAKEINTRESOURCE (IDD_LOADFILEPROGRESS),
										NULL,
										LFPW_DlgProc,
										0);

	gLFPW_MaxFile = 0;
	gLFPW_CurFile = 0;
	gLFPW_CurFileName[0] = 0;
	gLFPW_MaxValue = 0;
	gLFPW_CurValue = 0;
}


/***********************************************************************
 *
 * FUNCTION:	Platform::LFPW_SetNumFiles
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::LFPW_SetNumFiles (int numFiles)
{
	if (!gLFPW_Dlg)
		return;

	gLFPW_MaxFile = numFiles;

	// If there's only one file, hide the "## of ## files" section.

	if (gLFPW_MaxFile == 1)
	{
		HWND	dlgItem;

		// Hide the "Installing:" text.

		dlgItem = ::GetDlgItem (gLFPW_Dlg, IDC_INSTALLING);
		::ShowWindow (dlgItem, SW_HIDE);

		// Hide the "## of ## files" text.

		dlgItem = ::GetDlgItem (gLFPW_Dlg, IDC_CURRENT_FILENUMBER);
		::ShowWindow (dlgItem, SW_HIDE);

		// Get the bounds of that text while we're here.

		RECT	topRect;
		::GetWindowRect (dlgItem, &topRect);

		// Hide the associated progress bar.

		dlgItem = ::GetDlgItem (gLFPW_Dlg, IDC_PROGRESS_WHICHFILE);
		::ShowWindow (dlgItem, SW_HIDE);

		// Get the Cancel button and get its bounds.

		dlgItem = ::GetDlgItem (gLFPW_Dlg, IDCANCEL);

		RECT	bottomRect;
		::GetWindowRect (dlgItem, &bottomRect);

		// Figure out how much to move it.

		int	delta = bottomRect.top - topRect.top;

		// Move the Cancel button.

		::MapWindowPoints (NULL, gLFPW_Dlg, (LPPOINT) &bottomRect, 2);

		::MoveWindow (	dlgItem, bottomRect.left, bottomRect.top - delta,
						bottomRect.right - bottomRect.left,
						bottomRect.bottom - bottomRect.top, true);

		// Get the dialog frame bounds and resize it.

		RECT	frameRect;
		::GetWindowRect (gLFPW_Dlg, &frameRect);

		::MoveWindow (	gLFPW_Dlg, frameRect.left, frameRect.top,
						frameRect.right - frameRect.left,
						frameRect.bottom - frameRect.top - delta, true);
	}
}


/***********************************************************************
 *
 * FUNCTION:	Platform::LFPS_StartNewFile
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::LFPW_StartNewFile (const char* path)
{
	if (!gLFPW_Dlg)
		return;

	// Remember the new name (stripping of path information).

	char	fName[_MAX_FNAME], ext[_MAX_EXT];
	_splitpath (path, NULL, NULL, fName, ext);

	strcpy (gLFPW_CurFileName, fName);
	strcat (gLFPW_CurFileName, ext);

	gLFPW_CurFile++;
	gLFPW_MaxValue = 0;
	gLFPW_CurValue = 0;

	HWND	dlgItem;

	// Show the new name.

	dlgItem = ::GetDlgItem (gLFPW_Dlg, IDC_CURRENT_FILENAME);
	::SetWindowText (dlgItem, gLFPW_CurFileName);

	// Update the "progress for this file" indicator.

	dlgItem = ::GetDlgItem (gLFPW_Dlg, IDC_PROGRESS_WHICHRECORD);
	::SendMessage (dlgItem, PBM_SETRANGE, 0, MAKELPARAM(0, gLFPW_MaxValue));
	::SendMessage (dlgItem, PBM_SETPOS, gLFPW_CurValue, 0);

	// Update the "## of ## files" message.

	string	progress (Platform::GetString (kStr_XofYFiles));
	char	buffer[200];
	sprintf(buffer, progress.c_str (), gLFPW_CurFile, gLFPW_MaxFile);

	dlgItem = ::GetDlgItem (gLFPW_Dlg, IDC_CURRENT_FILENUMBER);
	::SetWindowText (dlgItem, buffer);

	// Update the "progress for all files" indicator.

	dlgItem = ::GetDlgItem (gLFPW_Dlg, IDC_PROGRESS_WHICHFILE);
	::SendMessage (dlgItem, PBM_SETRANGE, 0, MAKELPARAM(0, gLFPW_MaxFile));
	::SendMessage (dlgItem, PBM_SETPOS, gLFPW_CurFile, 0);

	::UpdateWindow (gLFPW_Dlg);
}


/***********************************************************************
 *
 * FUNCTION:	Platform::LFPW_SetFileMax
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::LFPW_SetFileMax (int max)
{
	if (!gLFPW_Dlg)
		return;

	gLFPW_MaxValue = max;
	gLFPW_CurValue = 0;

	// Update the "progress for this file" indicator.

	HWND	dlgItem = ::GetDlgItem (gLFPW_Dlg, IDC_PROGRESS_WHICHRECORD);
	::SendMessage (dlgItem, PBM_SETRANGE, 0, MAKELPARAM(0, gLFPW_MaxValue));
	::SendMessage (dlgItem, PBM_SETPOS, gLFPW_CurValue, 0);

	::UpdateWindow (gLFPW_Dlg);
}


/***********************************************************************
 *
 * FUNCTION:	Platform::LFPW_SetFileValue
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::LFPW_SetFileValue (int value)
{
	if (!gLFPW_Dlg)
		return;

	gLFPW_CurValue = value;

	// Update the "progress for this file" indicator.

	HWND	dlgItem = ::GetDlgItem (gLFPW_Dlg, IDC_PROGRESS_WHICHRECORD);
	::SendMessage (dlgItem, PBM_SETPOS, gLFPW_CurValue, 0);

	::UpdateWindow (gLFPW_Dlg);
}


/***********************************************************************
 *
 * FUNCTION:	Platform::LFPW_Close
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::LFPW_Close (void)
{
	if (!gLFPW_Dlg)
		return;

	::DestroyWindow (gLFPW_Dlg);
	gLFPW_Dlg = NULL;
}



// ---------------------------------------------------------------------------
//		 Gremlin Control Window
// ---------------------------------------------------------------------------

static HWND	gGCW_Dlg;
static long	gGCW_GremlinNumber;
static long	gGCW_EventNumber;
static long	gGCW_MaxEventNumber;
static long	gGCW_ElapsedMilliseconds;
static Bool	gGCW_NeedUpdate;

// From Application.cpp

void DoGremlinsStep		(HWND hwnd, WPARAM command);
void DoGremlinsResume	(HWND hwnd, WPARAM command);
void DoGremlinsStop		(HWND hwnd, WPARAM command);


/***********************************************************************
 *
 * FUNCTION:	UpdateEventNumber
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

static void UpdateEventNumber (void)
{
	if (!gGCW_Dlg)
		return;

	// Update the "Event #: xxx of yyy" message.

	char	xxx[20];
	char	yyy[20];
	char	buffer[200];

	::PrvFormatInteger (xxx, gGCW_EventNumber);
	::PrvFormatInteger (yyy, gGCW_MaxEventNumber);

	// If there is a max event number (indicated by a non -1 value
	// for gGCW_MaxEventNumber), display a string including that
	// value.  Otherwise, the number of events to post is unlimited.
	// In that case, we just show the current event number.

	if (gGCW_MaxEventNumber != -1)
	{
		static string	fmt (Platform::GetString (kStr_XofYEvents));
		sprintf(buffer, fmt.c_str (), xxx, yyy);
	}
	else
	{
		static string	fmt (Platform::GetString (kStr_XEvents));
		sprintf(buffer, fmt.c_str (), xxx);
	}

	HWND	dlgItem = ::GetDlgItem (gGCW_Dlg, IDC_EVENT_NUMBER);
	::SetWindowText (dlgItem, buffer);
}


/***********************************************************************
 *
 * FUNCTION:	UpdateElapsedTime
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

static void UpdateElapsedTime (void)
{
	if (!gGCW_Dlg)
		return;

	// Get hours, minutes, and seconds.

	const long	kMillisecondsPerSecond	= 1000;
	const long	kSecondsPerMinute		= 60;
	const long	kMinutesPerHour			= 60;

	const long	kMillisecondsPerMinute	= kMillisecondsPerSecond * kSecondsPerMinute;
	const long	kMillisecondsPerHour	= kMillisecondsPerMinute * kMinutesPerHour;

	long	msecs = gGCW_ElapsedMilliseconds;
	long	hours = msecs / kMillisecondsPerHour;		msecs -= hours * kMillisecondsPerHour;
	long	minutes = msecs / kMillisecondsPerMinute;	msecs -= minutes * kMillisecondsPerMinute;
	long	seconds = msecs / kMillisecondsPerSecond;	msecs -= seconds * kMillisecondsPerSecond;

	// Format them into a string.

	char	formattedTime[20];
	sprintf (formattedTime, "%ld:%02ld:%02ld", hours, minutes, seconds);

	// Combine the formatted time with the rest of the string.

	char			msg[200];
	static string	fmt (Platform::GetString (kStr_ElapsedTime));
	sprintf (msg, fmt.c_str (), formattedTime);

	// Display the string.

	HWND	dlgItem = ::GetDlgItem (gGCW_Dlg, IDC_ELAPSED_TIME);
	::SetWindowText (dlgItem, msg);
}


/***********************************************************************
 *
 * FUNCTION:	GCW_DlgProc
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

static BOOL CALLBACK GCW_DlgProc (	HWND hwnd,
									UINT msg,
									WPARAM wParam,
									LPARAM lParam)
{
	switch (msg)
	{
		case WM_INITDIALOG:
		{
			Point	pos = gPrefs->GetGCWLocation ();
			if (pos.h == -1 && pos.v == -1)
			{
				::CenterWindow (hwnd);
			}
			else
			{
				::SetWindowPos (hwnd, NULL, pos.h, pos.v,
							-1, -1, SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE);
			}
			
			::SetTimer (hwnd, 2, 100, NULL);

			return TRUE;	// The dialog box procedure should return TRUE to direct
							// the system to set the keyboard focus to the control
							// given by hwndFocus. Otherwise, it should return FALSE
							// to prevent the system from setting the default keyboard focus.
		}

		case WM_COMMAND:
		{
			switch (wParam)
			{
				case IDC_STOP:
				{
					::DoGremlinsStop (LCD_GetWindow (), ID_GREMLINS_STOP);
					break;
				}

				case IDC_RESUME:
				{
					::DoGremlinsResume (LCD_GetWindow (), ID_GREMLINS_RESUME);
					break;
				}

				case IDC_STEP:
				{
					::DoGremlinsStep (LCD_GetWindow (), ID_GREMLINS_STEP);
					break;
				}

				default:
					return 0;
			}

			return 1;
		}

		case WM_CLOSE:
		{
			Platform::GCW_Close ();
			return 1;
		}

		case WM_DESTROY:
		{
			RECT	wr;
			::GetWindowRect (hwnd, &wr);

			Point	pt = { wr.top, wr.left };
			gPrefs->SetGCWLocation (pt);

			::KillTimer (hwnd, 2);
			return 1;
		}

		case WM_TIMER:
		{
			// See comments in GCW_SetEventNumber.

			if (gGCW_NeedUpdate)
			{
				gGCW_NeedUpdate = false;

				::UpdateEventNumber ();
				::UpdateElapsedTime ();
			}

			return 1;
		}
	}

	return 0;	// Except in response to the WM_INITDIALOG message, the dialog box
				// procedure should return nonzero if it processes the message, and
				// zero if it does not.
}


/***********************************************************************
 *
 * FUNCTION:	Platform::GCW_Open
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::GCW_Open (void)
{
	if (!gGCW_Dlg)
	{
		gGCW_Dlg = ::CreateDialogParam (	gInstance,
											MAKEINTRESOURCE (IDD_GREMLIN_CONTROL),
											NULL,
											GCW_DlgProc,
											0);
	}

	gGCW_GremlinNumber			= -1;
	gGCW_EventNumber			= -1;
	gGCW_MaxEventNumber			= -1;
	gGCW_ElapsedMilliseconds	= 0;
	gGCW_NeedUpdate				= true;

	HWND	dlgItem;

	dlgItem = ::GetDlgItem (gGCW_Dlg, IDC_EVENT_NUMBER);
	::SetWindowText (dlgItem, "");

	dlgItem = ::GetDlgItem (gGCW_Dlg, IDC_GREMLIN_NUMBER);
	::SetWindowText (dlgItem, "");

	dlgItem = ::GetDlgItem (gGCW_Dlg, IDC_ELAPSED_TIME);
	::SetWindowText (dlgItem, "");
}


/***********************************************************************
 *
 * FUNCTION:	Platform::GCW_SetGremlinNumber
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::GCW_SetGremlinNumber (int gremlinNumber)
{
	gGCW_GremlinNumber = gremlinNumber;

	if (!gGCW_Dlg)
		return;

	// Update the "Gremlin #xxx" message.

	char	buffer[200];
	string	fmt (Platform::GetString (kStr_GremlinNumber));
	sprintf(buffer, fmt.c_str (), (long) gGCW_GremlinNumber);

	HWND	dlgItem = ::GetDlgItem (gGCW_Dlg, IDC_GREMLIN_NUMBER);
	::SetWindowText (dlgItem, buffer);
}


/***********************************************************************
 *
 * FUNCTION:	Platform::GCW_SetEventNumber
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::GCW_SetEventNumber (int eventNumber)
{
	// Just remember the event number for now.  We can't actually draw it.
	// This function is called from the CPU thread, while the window is
	// owned by the UI thread.  Most of the time, attempting to redraw
	// the event number under these circumstances works, but will pretty
	// much fail all the time when the user clicks on the "Stop" button.
	// That's because the UI thread blocks on this thread, waiting for
	// it to stop.  For some reason, while that thread is waiting, the
	// CPU thread ends up calling this function.  If we were to attempt
	// to update the text while the UI thread was blocked, *we'd* block,
	// and the UI thread would eventually timeout while waiting for the
	// CPU thread to stop.
	//
	// Therefore, we don't draw in this function.  Instead, we remember
	// the event number, and draw it later in a WM_TIMER handler.

	gGCW_EventNumber			= eventNumber;
	gGCW_NeedUpdate				= true;
}


/***********************************************************************
 *
 * FUNCTION:	Platform::GCW_SetMaxEventNumber
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::GCW_SetMaxEventNumber (int maxEventNumber)
{
	// See comments in GCW_SetEventNumber.

	gGCW_MaxEventNumber			= maxEventNumber;
	gGCW_NeedUpdate				= true;
}


/***********************************************************************
 *
 * FUNCTION:	Platform::GCW_SetElapsedTime
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::GCW_SetElapsedTime (uae_u32 elapsedMilliseconds)
{
	// See comments in GCW_SetEventNumber.

	gGCW_ElapsedMilliseconds	= elapsedMilliseconds;
	gGCW_NeedUpdate				= true;
}


/***********************************************************************
 *
 * FUNCTION:	Platform::GCW_SetRunning
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::GCW_SetRunning (void)
{
	if (!gGCW_Dlg)
		return;

	HWND	dlgItem;

	dlgItem = ::GetDlgItem (gGCW_Dlg, IDC_STOP);
	::EnableWindow (dlgItem, Gremlins_CanStop ());

	dlgItem = ::GetDlgItem (gGCW_Dlg, IDC_RESUME);
	::EnableWindow (dlgItem, Gremlins_CanResume ());

	dlgItem = ::GetDlgItem (gGCW_Dlg, IDC_STEP);
	::EnableWindow (dlgItem, Gremlins_CanStep ());
}


/***********************************************************************
 *
 * FUNCTION:	Platform::GCW_SetStopped
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::GCW_SetStopped (void)
{
	// slacker code sharing...
	GCW_SetRunning ();
}


/***********************************************************************
 *
 * FUNCTION:	Platform::GCW_Close
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Platform::GCW_Close (void)
{
	if (!gGCW_Dlg)
		return;

	::DestroyWindow (gGCW_Dlg);
	gGCW_Dlg = NULL;
}



// ===========================================================================
//		 TByteQueue
// ===========================================================================

/***********************************************************************
 *
 * FUNCTION:	TByteQueue constructor
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

TByteQueue::TByteQueue () :
	fImplementation (new TByteQueuePrivate)
{
}


/***********************************************************************
 *
 * FUNCTION:	TByteQueue destructor
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

TByteQueue::~TByteQueue ()
{
	delete fImplementation;
}


/***********************************************************************
 *
 * FUNCTION:	TByteQueue::Put
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void TByteQueue::Put (uae_u8 b)
{
	fImplementation->Put (b);
}


/***********************************************************************
 *
 * FUNCTION:	TByteQueue::Get
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

uae_u8 TByteQueue::Get (void)
{
	return fImplementation->Get ();
}


/***********************************************************************
 *
 * FUNCTION:	TByteQueue::GetUsed
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

int TByteQueue::GetUsed (void)
{
	return fImplementation->GetUsed ();
}


/***********************************************************************
 *
 * FUNCTION:	TByteQueue::GetFree
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

int TByteQueue::GetFree (void)
{
	return fImplementation->GetFree ();
}


/***********************************************************************
 *
 * FUNCTION:	TByteQueue::WaitForDataAvailable
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

Bool TByteQueue::WaitForDataAvailable (int timeoutms)
{
	return fImplementation->WaitForDataAvailable (timeoutms);
}


// ===========================================================================
//		 TByteQueuePrivate
// ===========================================================================

/***********************************************************************
 *
 * FUNCTION:	TByteQueuePrivate constructor
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

TByteQueuePrivate::TByteQueuePrivate ()
{
	fUsedCount = 0;
	fFreeCount = QUEUE_SIZE;

	fDataAvailable = ::CreateEvent (NULL, TRUE, FALSE, NULL);

	fHead = 0;
	fTail = 0;
}


/***********************************************************************
 *
 * FUNCTION:	TByteQueuePrivate destructor
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

TByteQueuePrivate::~TByteQueuePrivate ()
{
	::CloseHandle (fDataAvailable);
	fDataAvailable = NULL;
}


/***********************************************************************
 *
 * FUNCTION:	TByteQueue::Put
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void TByteQueuePrivate::Put (uae_u8 b)
{
	// If the queue is full, just discard the data byte.

	if (fUsedCount >= QUEUE_SIZE)
	{
		return;
	}

	fBuffer[fHead] = b;
	fHead = (fHead + 1) % QUEUE_SIZE;

	::InterlockedIncrement(&fUsedCount);
	::InterlockedDecrement(&fFreeCount);

	// Tell clients that there may be new data in the buffer.

	// A possible race condition exists: Suppose thread A is at this point
	// in the code, and has just put one byte into an otherwise empty
	// buffer. If thread B then executes TByteQueuePrivate::Get, the byte
	// of data will be removed from the queue. Then thread A will set the
	// fDataAvailable event, when there really is no data available. This
	// is not a problem if clients check the amount of data actually available
	// in the queue when TByteQueue::WaitForDataAvailable returns, rather than
	// assuming there definitely is data.

	::SetEvent (fDataAvailable);
}


/***********************************************************************
 *
 * FUNCTION:	TByteQueue::Get
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

uae_u8 TByteQueuePrivate::Get ()
{
	// If the queue is empty, return a null byte (this shouldn't happen,
	// because the caller should always call GetUsed before Get).

	if (fUsedCount == 0)
	{
		return 0;
	}

	// If this is the last byte in the queue, reset the
	// fDataAvailable event _before_ removing the byte.

	if (fUsedCount == 1)
	{
		::ResetEvent (fDataAvailable);
	}

	uae_u8	r = fBuffer[fTail];
	fTail = (fTail + 1) % QUEUE_SIZE;

	::InterlockedIncrement(&fFreeCount);
	::InterlockedDecrement(&fUsedCount);

	return r;
}


/***********************************************************************
 *
 * FUNCTION:	TByteQueue::WaitForDataAvailable
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

Bool TByteQueuePrivate::WaitForDataAvailable (int timeoutms)
{
	return ::WaitForSingleObject (fDataAvailable, timeoutms) == WAIT_OBJECT_0;
}



/***********************************************************************
 *
 * FUNCTION:	PrvInsertString
 *
 * DESCRIPTION:	Insert one string into the middle of another.  The
 *				buffer containing the string being modified should be
 *				large enough to hold the resulting string.
 *
 * PARAMETERS:	dest - buffer containing the string to receive the
 *					inserted text.
 *
 *				insertBefore - character index at which the inserted
 *					text should appear. Assumed to be [0, strlen(dest)].
 *
 *				src - text to be inserted.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

static void PrvInsertString (char* dest, int insertBefore, const char* src)
{
	size_t	destLen = strlen (dest);
	size_t	srcLen = strlen (src);

	// Move down the part of the string after the insertion point.  Don't
	// use strcpy or memcpy, as we're dealing with an overlapping range.
	// Add 1 to the range length to get the terminating NULL.

	memmove (dest + insertBefore + srcLen, dest + insertBefore, destLen - insertBefore + 1);

	// Insert the new string.  *Don't* copy the NULL!

	memmove (dest + insertBefore, src, srcLen);
}


/***********************************************************************
 *
 * FUNCTION:	PrvFormatInteger
 *
 * DESCRIPTION:	Convert an integral value into a formatted string,
 *				adding thousands seperators as necessary.
 *
 * PARAMETERS:	dest - buffer for the outputted string
 *
 *				integer - input value
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

static void PrvFormatInteger (char* dest, uae_u32 integer)
{
	sprintf (dest, "%ld", integer);

	// Get the thousands separator character(s).

	struct lconv*	locale_data = localeconv ();
	char*			thousands_sep = locale_data->thousands_sep;

	if (strlen (thousands_sep) == 0)
	{
		thousands_sep = ",";
	}

	// Insert the thousands separator(s).

		// Divide by three to get the number of 3 digit groupings remaining
		// (subtracting one to get the math to come out right)
		//
		//		 1 -> 0
		//		 2 -> 0
		//		 3 -> 0
		//		 4 -> 1
		//		 5 -> 1
		//		 6 -> 1
		//		 7 -> 2
		//		 8 -> 2
		//		 9 -> 2
		//		10 -> 3
		//
		//	etc...

	int numCommas = (strlen (dest) - 1) / 3;

		// Special case the stupid rule about not putting a comma
		// in a number like xxxx.

	if (strlen (dest) <= 4)
	{
		numCommas = 0;
	}

	for (int ii = 1; ii <= numCommas; ++ii)
	{
		// Back up four for every comma (skip past every ",xxx" pattern).

		::PrvInsertString (dest, strlen (dest) + 1 - (4 * ii), thousands_sep);
	}
}
