/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#include "EmulatorCommon.h"
#include "Registry.h"


// ======================================================================
//	Globals and constants
// ======================================================================

const char*	kDefaultRegistryKey	= "Software\\3Com\\Palm OS Emulator";


/***********************************************************************
 *
 * FUNCTION:	Registry::SetRegistry
 *
 * DESCRIPTION:	Bottleneck function for setting a registry entry.
 *
 * PARAMETERS:	key - registry entry path.
 *
 *				name - registry entry name.
 *
 *				type - registry entry type (REG_SZ, REG_DWORD, etc.)
 *
 *				data - pointer to entry data
 *
 *				datasize - length of data in bytes.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Registry::SetRegistry (const char* key,
							const char* name,
							DWORD type,
							const void* data,
							DWORD datasize)
{
	HKEY	k;
	DWORD	d;

	if (::RegCreateKeyEx (HKEY_CURRENT_USER, key, 0, NULL, 0, KEY_WRITE, NULL, &k, &d) == ERROR_SUCCESS)
	{
		::RegSetValueEx (k, name, 0, type, (BYTE*) data, datasize);
		::RegCloseKey (k);
	}
}


/***********************************************************************
 *
 * FUNCTION:	Registry::GetRegistry
 *
 * DESCRIPTION:	Bottleneck function for retrieving data from a
 *				registry entry.
 *
 * PARAMETERS:	key - registry entry path.
 *
 *				name - registry entry name.
 *
 *				type - registry entry type (REG_SZ, REG_DWORD, etc.)
 *
 *				data - pointer to buffer to receive the data
 *
 *				datasize - length of buffer in bytes.
 *
 * RETURNED:	True if the data was retrieved, false otherwise
 *
 ***********************************************************************/

Bool Registry::GetRegistry (const char* key,
							const char* name,
							DWORD type,
							void* data,
							DWORD datasize)
{
	Bool	ok = false;
	HKEY	k;

	if (::RegOpenKeyEx (HKEY_CURRENT_USER, key, 0, KEY_READ, &k) == ERROR_SUCCESS)
	{
		DWORD	t;
		if (::RegQueryValueEx (k, (char*) name, NULL, &t, (BYTE*) data, &datasize) == ERROR_SUCCESS)
		{
			ok = t == type;
		}

		::RegCloseKey (k);
	}

	return ok;
}


/***********************************************************************
 *
 * FUNCTION:	Registry::RemoveSlot
 *
 * DESCRIPTION:	Remove the given entry from the registry.
 *
 * PARAMETERS:	name - name of entry to remove.
 *
 *				key - path of entry to remove.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Registry::RemoveSlot (const char* name, const char* key)
{
	HKEY	k;

	if (::RegOpenKeyEx (HKEY_CURRENT_USER, key, 0, KEY_ALL_ACCESS, &k) == ERROR_SUCCESS)
	{
		::RegDeleteValue (k, name);
		::RegCloseKey (k);
	}
}


/***********************************************************************
 *
 * FUNCTION:	Registry::SetString
 *
 * DESCRIPTION:	Save a string in the registry.
 *
 * PARAMETERS:	name - name of registry entry
 *
 *				data - string to save
 *
 *				key - path of registry entry
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Registry::SetString (	const char* name,
							const string& data,
							const char* key)
{
	Registry::SetRegistry (key, name, REG_SZ, data.c_str (), data.length () + 1);
}


/***********************************************************************
 *
 * FUNCTION:	Registry::GetString
 *
 * DESCRIPTION:	Retrieve a string from the registry
 *
 * PARAMETERS:	name - name of registry entry
 *
 *				data - buffer for retrieved string
 *
 *				key - path of registry entry
 *
 * RETURNED:	True if the data was retrieved, false otherwise
 *
 ***********************************************************************/

Bool Registry::GetString (	const char* name,
							string& data,
							const char* key)
{
	char	tempBuffer [256];
	Bool	result = Registry::GetRegistry (key, name, REG_SZ, tempBuffer, sizeof (tempBuffer));

	if (result)
	{
		data.erase ();	// KB: Q172398
		data = tempBuffer;
	}
	else
		data.erase ();

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	Registry::SetInt
 *
 * DESCRIPTION:	Save an integer in the registry.
 *
 * PARAMETERS:	name - name of registry entry
 *
 *				data - value to save
 *
 *				key - path of registry entry
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Registry::SetInt (const char* name, uae_u32 data, const char* key)
{
	Registry::SetRegistry (key, name, REG_DWORD, &data, sizeof (data));
}


/***********************************************************************
 *
 * FUNCTION:	Registry::GetInt
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

Bool Registry::GetInt (const char* name, uae_u32& data, const char* key)
{
	DWORD	dw;
	Bool	result = Registry::GetRegistry (key, name, REG_DWORD, &dw, sizeof (dw));
	if (result)
		data = (uae_u32) dw;
	return result;
}


Bool Registry::GetInt (const char* name, uae_s32& data, const char* key)
{
	DWORD	dw;
	Bool	result = Registry::GetRegistry (key, name, REG_DWORD, &dw, sizeof (dw));
	if (result)
		data = (uae_s32) dw;
	return result;
}


Bool Registry::GetInt (const char* name, uae_u16& data, const char* key)
{
	DWORD	dw;
	Bool	result = Registry::GetRegistry (key, name, REG_DWORD, &dw, sizeof (dw));
	if (result)
		data = (uae_u16) dw;
	return result;
}


Bool Registry::GetInt (const char* name, uae_s16& data, const char* key)
{
	DWORD	dw;
	Bool	result = Registry::GetRegistry (key, name, REG_DWORD, &dw, sizeof (dw));
	if (result)
		data = (uae_s16) dw;
	return result;
}


Bool Registry::GetInt (const char* name, uae_u8& data, const char* key)
{
	DWORD	dw;
	Bool	result = Registry::GetRegistry (key, name, REG_DWORD, &dw, sizeof (dw));
	if (result)
		data = (uae_u8) dw;
	return result;
}


Bool Registry::GetInt (const char* name, uae_s8& data, const char* key)
{
	DWORD	dw;
	Bool	result = Registry::GetRegistry (key, name, REG_DWORD, &dw, sizeof (dw));
	if (result)
		data = (uae_s8) dw;
	return result;
}


/***********************************************************************
 *
 * FUNCTION:	Registry::SetBool
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Registry::SetBool (const char* name,
						Bool data,
						const char* key)
{
	Registry::SetRegistry (key, name, REG_DWORD, &data, sizeof (data));
}


void Registry::SetBool (const char* name,
						bool data,
						const char* key)
{
	Registry::SetBool (name, (BOOL) data, key);
}


/***********************************************************************
 *
 * FUNCTION:	Registry::GetBool
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

Bool Registry::GetBool (const char* name,
						Bool& data,
						const char* key)
{
	return Registry::GetRegistry (key, name, REG_DWORD, &data, sizeof (data));
}


Bool Registry::GetBool (const char* name,
						bool& data,
						const char* key)
{
	DWORD	dw;
	Bool	result = Registry::GetRegistry (key, name, REG_DWORD, &dw, sizeof (dw));
	if (result)
		data = dw != 0;
	return result;
}


/***********************************************************************
 *
 * FUNCTION:	Registry::SetBinary
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Registry::SetBinary (const char* name,
						  const void* data,
						  DWORD datasize,
						  const char* key)
{
	Registry::SetRegistry (key, name, REG_BINARY, data, datasize);
}


/***********************************************************************
 *
 * FUNCTION:	Registry::GetBinary
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

Bool Registry::GetBinary (const char* name,
						  void* data,
						  DWORD datasize,
						  const char* key)
{
	return Registry::GetRegistry (key, name, REG_BINARY, data, datasize);
}
